header {* An SML equivalent to poly/Recursive *}

theory ML_Recursive_dense
imports Main
begin

section {* ML code for multiplication of recursive dense representation *}

subsection {* Auxiliary stuff *}
ML {*
fun cand a b = a andalso b
fun last_elem l = Library.nth l ((length l) - 1)
fun foldr (*1*) _ [] = I
  | foldr (*2*) f (x :: xs) = f x o foldr f xs
*} ML {*
(* ------------------------- output too verbose + need poly_str + poly_rec from below -----------
fun foldr (*1*) _ [] = 
let val _ = writeln (". . . . . . . . /foldr (*1*) []");
    val vvv = I
    in vvv end
  | foldr (*2*) f (x :: xs) = 
    let val _ = writeln (". . . . . . . . /foldr (*2*) ("  ^ poly_str x ^ ")...");
        val vvv = f x o foldr f xs
        val _ = writeln (". . . . . . . . \\foldr (*2*) ("  ^ poly_str x ^ ")  \<longrightarrow> ...");
    in vvv end
------------------------------------------------------------------------------------------------*)
*} ML {*
fun indent 0 mark = mark
  | indent n mark = "   " ^ indent (n - 1) mark;

if indent 0 "###" = "###" then () else error "indent 0 ### changed";
if indent 1 "###" = "   ###" then () else error "indent 1 ### changed";
if indent 0 "#####" = "#####" then () else error "indent 0 ##### changed";
*}

ML {*
datatype 'a poly_rec = Coeff of 'a | Powers of ('a poly_rec) list

(* examples for test; all but p7 normalised *)
val p1 = Coeff 7
val p2 = Powers [Powers [Coeff 0, Coeff 2], Coeff 3]          (*=== 2.x + 3.y             ===*)
val p3 = Powers [Powers [Coeff 0, Coeff 2]]                   (*=== 2.x                   ===*)
val p4 = Powers [Coeff 0, Coeff 3]                            (*===       3.y             ===*)
val p5 as Powers ps = Powers [Coeff 1, Coeff 2, Coeff 3]      (*=== 1 + 2.x + 3.x^2       ===*)
val p6 =                                                      (*=== 10.x + 100.y + 1000.z ===*)
  Powers [Powers [Powers [Coeff 0, Coeff 10], Coeff 100], Coeff 1000];
val p7 =                                   (*... NOT normalised ===  10.x^2.y^2.z^2       ===*)
  Powers [Powers [Powers [Coeff 0]],
  (*z^1*) Powers [Powers [Coeff 0]],
  (*z^2*) Powers [Powers [Coeff 0], Powers [Coeff 0], Powers [Coeff 0, Coeff 0, Coeff 10]]];
  (*              y^0               y^1               y^2  . (x^0      x^1      x^2     )*)
(* normalised below as:
  Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Coeff 100]]]*)
*} ML {*
*} ML {*

val q1 = Powers [Powers [Coeff 0, Coeff 4], Coeff 5]          (*=== 4.x + 5.y             ===*)
val q2 = Powers [Powers [Coeff 0, Coeff 0, Coeff 1], Powers [Coeff 0, Coeff 1], Coeff 1];
                                                              (*=== x^2 + x.y + y^2       ===*)
val q3 = Powers [Powers [Coeff 0, Coeff~2], Coeff~3]          (*===-2.x +-3.y             ===*)
val NOTnorm = Powers [Powers [Coeff 0, Powers [Coeff 111], Coeff 0, Coeff 0], Powers [Coeff 0]]
val q5 as Powers qs = Powers [Coeff 4, Coeff 5]               (*=== 4 + 5.x               ===*)
val q7 =                                                      (*=== 1 + 2.x + 3.y + 4.z   ===*)
  Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4];

fun poly_str (Coeff c) = "Coeff " ^ string_of_int c
  | poly_str (Powers ps) =
      "Powers [" ^ (commas (map poly_str ps)) ^ "]"
  | poly_str _ = error "poly_str only implemented for Rec Dense";

if poly_str q1 = "Powers [Powers [Coeff 0, Coeff 4], Coeff 5]" 
then () else error "poly_str changed";

fun no_trailing_zeros (Coeff _) = true
  | no_trailing_zeros (Powers []) = false (* for cut_empty_powers *)
  | no_trailing_zeros (Powers ps) = 
    last_elem ps <> Coeff 0 andalso fold cand (map no_trailing_zeros ps) true

fun drop_trailing_zeros ps =
  let
    val ps' = rev ps
    val ps'' = if hd ps' = Coeff 0 then tl ps' else ps'
  in rev ps'' end

fun cut_trailing_zeros (Coeff a) = (Coeff a)
  | cut_trailing_zeros (Powers ps) =
    let val ps' = drop_trailing_zeros ps
    in Powers (map cut_trailing_zeros ps') end

fun cut_empty_powers (Coeff a) = Coeff a
  | cut_empty_powers (Powers []) = Coeff 0
  | cut_empty_powers (Powers [Coeff a]) = Coeff a   
  | cut_empty_powers (Powers ps) = Powers (map cut_empty_powers ps)

fun normalise p =
  let
    val p' = cut_trailing_zeros p
    val p'' = cut_empty_powers p'
  in if no_trailing_zeros p'' then p'' else normalise p'' end
;
*} ML {*
val p7 = normalise p7 (* ---------vv *)
val ps = [p1, p2, p3, p4, p5, p6, p7];
val qs = [q1, q2, q3, q5, q7];
*} ML {*
if ps = map normalise ps then () else error "a ps example is not normalised";
*} ML {*
if qs = map normalise qs then () else error "a qs example is not normalised";
*} ML {*
if normalise NOTnorm = Powers [Powers [Coeff 0, Coeff 111]]
then () else error "normalise changed"
*} ML {*
*}

subsection {* Number of variables etc*}
ML {*
fun paths_to_coeff n (Coeff _) = [n]
  | paths_to_coeff n (Powers ps) = flat (map (paths_to_coeff (n + 1)) ps)

fun vars (Coeff 0) = 0
  | vars p = fold (curry Int.max) (paths_to_coeff 0 (normalise p)) 1
;
*} ML {*
if vars p1 = 1 then () else error "vars p1 changed";
if vars p2 = 2 then () else error "vars p2 changed";
if vars p3 = 2 then () else error "vars p3 changed"; (* "2.x" in (R[x])[y] *)
if vars q7 = 3 then () else error "vars q7 changed";
*} ML {*
*}

subsection {* Addition *}
ML {*
fun zip_with _ [] [] = []
  | zip_with _ [] ys = ys
  | zip_with _ xs [] = xs
  | zip_with f (x :: xs) (y :: ys) = f x y :: zip_with f xs ys;

(* compare plus_poly_rec: does NOT detail DENSE / SPARSE *)
fun add (Coeff a) (Coeff b) = Coeff (a + b) 
  | add (Powers ps) (Powers qs) = Powers (zip_with add ps qs)
  | add (pp as Coeff a) (qq as Powers (q :: qs)) = if a = 0 then qq else Powers ((add pp q) :: qs)
  | add (p as Powers _) (q as Coeff _) = add q p
  | add (Coeff _) (Powers []) = raise ERROR "Powers [] contradicts cut_empty_powers"    

fun plus_poly_rec p q = normalise (add p q)
;
*} ML {*
if plus_poly_rec p1 q1 = Powers [Powers [Coeff 7, Coeff 4], Coeff 5]
then () else error "plus_poly_rec p1 q1 changed";
if plus_poly_rec p2 q1 = Powers [Powers [Coeff 0, Coeff 6], Coeff 8]
then () else error "plus_poly_rec p2 q1 changed";
if plus_poly_rec p3 q1 = Powers [Powers [Coeff 0, Coeff 6], Coeff 5]
then () else error "plus_poly_rec p3 q1 changed";
if plus_poly_rec p4 q1 = Powers [Powers [Coeff 0, Coeff 4], Coeff 8]
then () else error "plus_poly_rec p4 q1 changed";
if plus_poly_rec p2 q2 = 
  Powers [Powers [Coeff 0, Coeff 2, Coeff 1], Powers [Coeff 3, Coeff 1], Coeff 1]
then () else error "plus_poly_rec p2 q2 changed";
*} ML {*
"~~~~~~ check plus_poly_rec p2 q2 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
"~~~~~ fun plus_poly_rec, args:"; val (p, q) = (p2, q2);
"~~~~~ fun add, args:"; val (Powers ps, Powers qs) = (p, q);
"~~~~~ fun zip_with, args:"; val (f, (x :: xs), (y :: ys)) = (add, ps, qs);
if f x y = Powers [Coeff 0, Coeff 2, Coeff 1] then () else error "add 2.x + x^2 changed";
"~~~~~ fun zip_with, args:"; val (f, (x :: xs), (y :: ys)) = (add, xs, ys);
if f x y = Powers [Coeff 3, Coeff 1] then () else error "add 3.y + x.y changed";
if plus_poly_rec p2 q2 = 
  Powers [Powers [Coeff 0, Coeff 2, Coeff 1], Powers [Coeff 3, Coeff 1], Coeff 1]
then () else error "add (2.x + 3.y) (x^2 + x.y + y^2) changed";
*} ML {*
*}

subsection {* Increment power *}
ML {*
*} ML {*
fun is_coeff c = case c of Coeff _ => true | _ => false

fun is_coeff_list cs = fold cand (map is_coeff cs) true

*} ML {* (* original *)
fun inc_power_rec 0 (Coeff a) = if a = 0 then Coeff 0 else Powers [Coeff 0, Coeff a]
  | inc_power_rec 0 (Powers ps) = Powers ((Coeff 0) :: ps)
  | inc_power_rec lev (Coeff a) = if a = 0 then Coeff 0 else Powers [inc_power_rec (lev - 1) (Coeff a)]
  | inc_power_rec lev (Powers ps) = Powers (map (inc_power_rec (lev - 1)) ps)
*} ML {* (* tracing *)
fun inc_power_rec 0 (Coeff a) = 
    let val vvv = if a = 0 then Coeff 0 else Powers [Coeff 0, Coeff a]
      val _ = writeln (indent 7 "..." ^ "inc_power_rec (*1*) 0 (" ^ poly_str (Coeff a) ^
    ") \<longrightarrow> (" ^ poly_str vvv ^ ")");
    in vvv end
  | inc_power_rec 0 (Powers ps) = 
    let val vvv = Powers ((Coeff 0) :: ps)
      val _ = writeln (indent 7 "..." ^ "inc_power_rec (*2*) 0 (" ^ poly_str (Powers ps) ^
    ") \<longrightarrow> (" ^ poly_str vvv ^ ")");
    in vvv end
  | inc_power_rec lev (Coeff a) = 
    let val vvv = if a = 0 then Coeff 0 else Powers [inc_power_rec (lev - 1) (Coeff a)]
      val _ = writeln (indent 7 "..." ^ "inc_power_rec (*3*) " ^ string_of_int lev ^ " (" ^ poly_str (Coeff a) ^
    ") \<longrightarrow> (" ^ poly_str vvv ^ ")");
    in vvv end
  | inc_power_rec lev (Powers ps) = 
    let val vvv =   Powers (map (inc_power_rec (lev - 1)) ps)
      val _ = writeln (indent 7 "..." ^ "inc_power_rec (*4*) " ^ string_of_int lev ^ " (" ^ poly_str (Powers ps) ^
    ") \<longrightarrow> (" ^ poly_str vvv ^ ")");
    in vvv end
*} ML {*
val lev = 0                                 (* (R[x,y])[z]:  1 + 2.x + 3.y + 4.z *)
val p =               Powers [         Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4];
if                                 (* z^0 .   (y^0    (x^0 . 1 + x^1 .2) +y^1 . 3) +z^1 . 4*)
inc_power_rec lev p = Powers [Coeff 0, Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]
then () else error "inc_power_rec lev = 0 q7 changed";

val lev = 1
val p =           
  Powers [Powers [         Powers [Coeff 1, Coeff 2], Coeff 3],                  Coeff 4];
if inc_power_rec lev p = 
  Powers [Powers [Coeff 0, Powers [Coeff 1, Coeff 2], Coeff 3], Powers [Coeff 0, Coeff 4]]
then () else error "inc_power_rec lev = 1 q7 changed";

val lev = 2
val p =           
Powers [Powers [Powers [         Coeff 1, Coeff 2],                  Coeff 3],                           Coeff 4];
if inc_power_rec lev p =
Powers [Powers [Powers [Coeff 0, Coeff 1, Coeff 2], Powers [Coeff 0, Coeff 3]], Powers [Powers [Coeff 0, Coeff 4]]]
then () else error "inc_power_rec lev = 2 q7 changed";
*} ML {*
(* compare check times (2.x + 3.y) (4.x + 5.y) *)
(* "~~m0f0mf0 --------------------------------*)
val lev = 1
val p =              Powers [Powers [         Coeff 0, Coeff 8], Coeff                  10];
if inc_power_rec lev p = Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]
then () else error "inc_power_rec lev = 1 changed";

(* "~~m0f0 return ----------------------------*)
val lev = 0
val p =                  Powers [         Powers [Coeff 0, Coeff 12], Coeff 15];
if inc_power_rec lev p = Powers [Coeff 0, Powers [Coeff 0, Coeff 12], Coeff 15]
then () else error "inc_power_rec lev = 0 changed";
*} ML {* (* check error in poly 22c450c5fda7: the result below is correct, while in Try_Test.thy
  inc_power_rec (*4*) #2 (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) \<longrightarrow> 
  (Powers [Powers [Powers [Coeff 0, Coeff 2, Coeff 4], 
                   Powers [Coeff 0, Coeff 6]], 
           Powers [Coeff 0, Powers [Coeff 0, Coeff 8]]]) *)            
val lev = 2;
val p = Powers[                         (* (R[x,y])[z]        = 2   + 4.x   + 6.y  + 8.z  *)
  Powers[                               (* z^0.                            |              *)
    Powers[Coeff 2, Coeff 4],           (*     y^0.(x^0.2 + x^1.4)         |= 2 + 4.x     *)
    Coeff 6],                           (*     y^1. x^0.6                  |= 6.y  *)
  Coeff 8];                             (* z^1.y^0. x^0.8                  |= 8.z         *)
if inc_power_rec lev p = Powers [       (* (R[x,y])[z]        = 2.x + 4.x^2 + 6.xy + 8.xz *)
  Powers [                              (* z^0.                            |              *)
    Powers [Coeff 0, Coeff 2, Coeff 4], (*     y^0.(x^0.0 + x^1.2 + x^2.4) |= 2.x + 4.x^2 *)
    Powers [Coeff 0, Coeff 6]],         (*     y^1.(x^0.0 + x^1.6)         |= 6.xy        *)
  Powers [                              (* z^1.                            |              *)
    Powers [Coeff 0, Coeff 8]]]         (*     y^0.(x^0.0 + x^1.8)         |= 8.xz        *)
then () else error "inc_power_rec (lev 2)  Powers [Powers [Powers [2, 4], 6], 8] changed";
*}

subsubsection {* Increment by n \<ge> 1 *}
(*-------------------------------------------------------------------------------------------------
ML {*
fun incn_power_rec n 0 (Coeff a) =
      if a = 0 then Coeff 0 else Powers (replicate n (Coeff 0) @ [Coeff a])
  | incn_power_rec n 0 (Powers ps) = Powers (replicate n (Coeff 0) @ ps)
  | incn_power_rec n lev (Coeff a) =
      if a = 0 then Coeff 0 else Powers [incn_power_rec n (lev - 1) (Coeff a)]
  | incn_power_rec n lev (Powers ps) = Powers (map (incn_power_rec n (lev - 1)) ps)
*} ML {*
val lev = 0 (* .z *)                    (* (R[x,y])[z]:  1 + 2.x + 3.y + 4.z *)
val p =           Powers [         Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4];
if                       (*        z^0 .  (y^0  . (x^0 . 1 + x^1 .2) +y^1 . 3) +z^1 . 4*)
incn_power_rec 1 lev p = Powers [Coeff 0, Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]
              (* p.z^1 =         z^0 . 0 + z^1 . (y^0  . (x^0 . 1 +x^1 . 2) +y^1 . 3) +z^2 . 4*)
then () else error "incn_power_rec 1 lev = 0 q7 changed";
if incn_power_rec 3 lev p = 
  Powers [Coeff 0, Coeff 0, Coeff 0,      Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]
(*p.z^3 = z^0 . 0 +z^1 . 0 +z^2 . 0 +     z^3  . (y^0  . (x^0 . 1 +x^1 . 2) +y^1 . 3) +z^4 . 4*)
then () else error "incn_power_rec 3 lev = 0 q7 changed";
*} ML {*
val lev = 1 (* .y *)
val p =           
  Powers [Powers [         Powers [Coeff 1, Coeff 2], Coeff 3],                  Coeff 4];
if incn_power_rec 1 lev p = (* p.y^1 *)
  Powers [Powers [Coeff 0, Powers [Coeff 1, Coeff 2], Coeff 3], Powers [Coeff 0, Coeff 4]]
then () else error "incn_power_rec 1 lev = 1 q7 changed";
if incn_power_rec 3 lev p = (* p.y^3 *)
  Powers [Powers [Coeff 0, Coeff 0, Coeff 0, Powers [Coeff 1, Coeff 2], Coeff 3], 
  (*      z^0 .  (y^0 . 0 +y^1 . 0 +y^2 . 0 +y^3 .  (x^0 . 1 +x^1 . 2) +y^4 . 3 +*)
    Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 4]]
  (*z^1  . (y^0 . 0 +y^1 . 0 +y^2 . 0 +y^3 . 4)      = y^3 + 2xy^3 + 3y^4 + 4y^3z *)
then () else error "incn_power_rec 3 lev = 1 q7 changed";
*} ML {*
val lev = 2 (* .x *)
val p =           
Powers [Powers [Powers [         Coeff 1, Coeff 2],                  Coeff 3],                           Coeff 4];
if incn_power_rec 1 lev p =
Powers [Powers [Powers [Coeff 0, Coeff 1, Coeff 2], Powers [Coeff 0, Coeff 3]], Powers [Powers [Coeff 0, Coeff 4]]]
then () else error "inc_power_rec 1 lev = 2 q7 changed";
if incn_power_rec 3 lev p = 
  Powers [Powers [Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 1, Coeff 2],
                  Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 3]],
          Powers [Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 4]]]
then () else error "incn_power_rec 3 lev = 2 q7 changed";
*}
-------------------------------------------------------------------------------------------------*)

subsection {* Multiplication *}
ML {*
fun smult_rec s (Coeff c) = Coeff (s * (*enforces int*) c)
  | smult_rec s (Powers ps) = Powers (map (smult_rec s) ps)
*} ML {*
fun times lev (Coeff a) (Coeff b) = 
(writeln (indent lev "###" ^ " times (*1*): lev = " ^ string_of_int lev ^ 
"  (Coeff " ^ string_of_int a ^ ") * (Coeff " ^ string_of_int b ^ 
") \<longrightarrow> (" ^ poly_str (Coeff (a * b)) ^ ")");
      Coeff (a * b)
)
  | times lev (Coeff a) (Powers qs) = 
(writeln (indent lev "###" ^ " times (*2*): lev = " ^ string_of_int lev ^ 
"  (Coeff " ^ string_of_int a ^ ") * (" ^ poly_str (Powers qs) ^ 
") \<longrightarrow> (" ^ poly_str (Powers (map (smult_rec a) qs)) ^ ")");
      Powers (map (smult_rec a) qs)
)
  | times lev (Powers ps) (Coeff b) = 
(writeln (indent lev "###" ^ " times (*3*): lev = " ^ string_of_int lev ^ 
"  (" ^ poly_str (Powers ps) ^ ") * (Coeff " ^ string_of_int b ^ 
") \<longrightarrow> (" ^ poly_str (Powers (map (smult_rec b) ps)) ^ ")");
      Powers (map (smult_rec b) ps)                
)
  | times lev (Powers ps) q =
let
  val _ = writeln (indent lev "/##" ^ " times (*4*): lev = " ^ string_of_int lev ^ 
"  (" ^ poly_str (Powers ps) ^ ") * (Coeff " ^ poly_str q ^ ")...");
  val xxx = foldr (fn a => (fn p => 
  let val _ = writeln (indent lev " /####" ^ " \<lambda> (" ^ string_of_int lev ^ 
      ")  (" ^ poly_str a ^ ")  (" ^ poly_str p ^ "). ...")
      val prod = times (lev + 1) a q
      val inc = inc_power_rec lev p
      val sum = plus_poly_rec prod inc
  in
    (writeln (indent lev " \\####" ^ " \<lambda> (" ^ string_of_int lev ^ 
      ")  (" ^ poly_str a ^ ")  (" ^ poly_str p ^ "). (" ^ 
      poly_str prod ^ ") + (" ^ poly_str inc ^ ") \<longrightarrow> (" ^ poly_str sum ^ ")");
          sum)
  end
      )) ps (Coeff 0)
in
(writeln (indent lev "\\##" ^ " times (*4*): lev = " ^ string_of_int lev ^ 
"  (" ^ poly_str (Powers ps) ^ ") * (Coeff " ^ poly_str q ^ ") \<longrightarrow> (" ^ poly_str xxx ^ ")");
    xxx
)
end
*} text {* (* original *)
fun times _ (Coeff a) (Coeff b) = Coeff (a * b)
  | times _ (Coeff a) (Powers qs) = Powers (map (smult_rec a) qs)
  | times _ (Powers ps) (Coeff b) = Powers (map (smult_rec b) ps)
  | times lev (Powers ps) q =
    foldr (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))) 
      ps (Coeff 0)

*} ML {* (* original with increment by n \<ge> 1 overwrites version for tracing *)
(*-------------------------------------------------------------------------------------------------
fun times _ (Coeff a) (Coeff b) = Coeff (a * b)
  | times _ (Coeff a) (Powers qs) = Powers (map (smult_rec a) qs)
  | times _ (Powers ps) (Coeff b) = Powers (map (smult_rec b) ps)
  | times lev (Powers ps) q =
    foldr (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (incn_power_rec 1 lev p)))) 
      ps (Coeff 0)
-------------------------------------------------------------------------------------------------*)
*} ML {*
*} ML {* (* original *)
fun times_poly_rec_dense p q = normalise (times 0 p q)
*} ML {*
*} 

subsubsection {* Version similar to Isabelle *}
text {* we make foldr similar (while also looking at generated code).
  This was stopped, because test-writeln fixes signature of foldr such,
  that other uses of foldr lead to type errors. 2 possible ways out:
  (a) add specific functions for writeln
  (b) use dictionary-mechanism from generated code for using writeln for
      appropriate calls of foldr_* only.
*}
(*-------------------------------------------------------------------------------------------------
ML {*
(* Nat_Mapping.thy:
fun foldr_dense_raw :: "nat \<Rightarrow> 'a list \<Rightarrow> 'b \<Rightarrow> 'b"
where "foldr_dense_raw n []     b = b"
    | "foldr_dense_raw n (x#xs) b = (if x = 0 then id else f n x) (foldr_dense_raw (Suc n) xs b)"*)
val pp = [Coeff 0, Coeff 0, Coeff 0, Coeff 2];
val qq = q1;
*} ML {*
fun foldr_dense_raw _ [] = I
  | foldr_dense_raw f (x :: xs) = (if x = Coeff 0 then I else f x) o foldr_dense_raw f xs;
foldr_dense_raw : (int poly_rec -> 'a -> 'a) -> int poly_rec list -> 'a -> 'a;
(*                 ^^^^^^^^^^^^ would inhibit to use foldr in foldr_step *)
*} ML {*
fun foldr_dense_raw _ _ [] = I
  | foldr_dense_raw NONE f (x :: xs) = (f x) o foldr_dense_raw NONE f xs
  | foldr_dense_raw (SOME x') f (x :: xs) = (if x = x' then I else f x) o foldr_dense_raw NONE f xs;
foldr  :                        ('a ->           'b -> 'b) -> 'a list           -> 'b -> 'b;
(*foldr_dense_raw :             (int poly_rec -> 'a -> 'a) -> int poly_rec list -> 'a -> 'a*)
foldr_dense_raw : ''a option -> (''a ->          'b -> 'b) -> ''a list          -> 'b -> 'b;
*} ML {*
(* Nat_Mapping_Recursive.thy:
definition foldr :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a)         \<Rightarrow> 'b \<Rightarrow> 'b"
where "foldr f p = List.foldr (\<lambda>(k, v) c. f k v c) (items p)"

is "foldra" in generated code:*)
fun foldra NONE f xs z = foldr f xs z                         (* this was too different *)
  | foldra (SOME x) f xs z = foldr_dense_raw (SOME x) f xs z; (* this will create more similarity *)
foldra: ''a option -> (''a -> 'b -> 'b) -> ''a list -> 'b -> 'b
*} ML {*
fun times _ (Coeff a) (Coeff b) = Coeff (a * b)
  | times _ (Coeff a) (Powers qs) = Powers (map (smult_rec a) qs)
  | times _ (Powers ps) (Coeff b) = Powers (map (smult_rec b) ps)
  | times lev (Powers ps) q =
    foldra NONE (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (incn_power_rec 1 lev p))))
      ps (Coeff 0)
(*... this runs all tests below OK *)
*} text {*
fun times _ (Coeff a) (Coeff b) = Coeff (a * b)
  | times _ (Coeff a) (Powers qs) = Powers (map (smult_rec a) qs)
  | times _ (Powers ps) (Coeff b) = Powers (map (smult_rec b) ps)
  | times lev (Powers ps) q =
    foldra (SOME (Coeff 0)) (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (incn_power_rec 1 lev p))))
      ps (Coeff 0)
(*... this gives the same error in "~~mf0~~~~ fun (fn a => ..." as in Try_Examples.thy + cp below:
  times (lev + 1) a q = Powers [Powers [Coeff 0, Coeff 8], Coeff 10]
this error should be overcome by foldr_step ...*)
*} 

text {* we make foldr_step similar to Nat_Mapping_Recursive.foldr_step --------------------------*}
(*definition foldr_step :: "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 
                            (nat \<Rightarrow> nat option \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b \<Rightarrow> 'b"
where "foldr_step f g p x = 
  (let (c, n') = foldr (\<lambda>n v (c, n'). (f n v (g n n' c), Some n)) p (x, None)
   in g 0 n' c)"*)
text {* work on signature of "foldr_step" via arguments only *}
ML {*
*} text {*
fun foldr_step f g p x =
  let val (c, n') = foldra NONE f p(*[]*) (x(*Coeff 111*), NONE)
  in g 0 n' c (*Coeff 111*) end
*} text {*
(*foldr_step : 'a -> 'b -> 'c -> 'd -> int poly_rec                               in Coeff 111    *)
(*foldr_step : 'a -> (int -> 'b option -> int poly_rec -> 'c) -> 'd -> 'e -> 'c  (Coeff 111, NONE)*)
(*foldr_step : 'a -> (int -> 'b option -> 'c           -> 'd) -> 'e -> 'c -> 'd          (x, NONE)*)
(*foldr_step : ('a -> 'b * 'c option -> 'b * 'c option) ->
                     (int -> 'c option -> 'b           -> 'd) -> 'e -> 'b -> 'd         foldr f []*)
(*foldr_step : ('a -> 'b * 'c option -> 'b * 'c option) ->
                     (int -> 'c option -> 'b           -> 'd) -> 'a list -> 'b -> 'd    foldr f p *)
  foldr_step : (''a -> 'b * 'c option -> 'b * 'c option) -> 
                     (int -> 'c option -> 'b           -> 'd) -> ''a list -> 'b -> 'd
*} text {* (*^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v^=^v^v*)
  foldr_step :              ((*\nat\*)''a          -> 'b * 'c option -> 'b * 'c option) -> 
                     (int -> 'c option -> 'b           -> 'd) -> ''a list  -> 'b -> 'd
(*definition foldr_step :: "(   nat \<Rightarrow>'a :: zero    \<Rightarrow> 'b             \<Rightarrow> 'b            ) \<Rightarrow> 
                     (nat \<Rightarrow> nat option \<Rightarrow> 'b           \<Rightarrow> 'b) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b \<Rightarrow> 'b"*)
*} 
text {* work on internal function of "foldr_step" *}
ML {*
(*definition foldr_step :: 
   "(nat \<Rightarrow> 'a :: zero \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (nat \<Rightarrow> nat option \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> (\<nat> \<Rightarrow>\<^sub>0 'a) \<Rightarrow> 'b \<Rightarrow> 'b"
*)
fun f a b = b;
f:(*\nat\*)  'a        -> 'b -> 'b;
fun g (n : int) (SOME (n' : int)) b = b;
g:                                    int -> int option -> 'a -> 'a ;
*}
text {* ... stopped, because there seems no way to meet the signature of foldr(a) *} 
-------------------------------------------------------------------------------------------------*)

declare [[ML_print_depth = 999]]

subsubsection {* OK check the base case: times (Coeff a) (Powers qs) *}
ML {*
"~~~~~~ check the base case: times (Coeff a) (Powers qs) ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
*} ML {*
p1 = Coeff 7;
p4 = Powers [Coeff 0, Coeff 3];
if times_poly_rec_dense p1 p4 = Powers [Coeff 0, Coeff 21] then () else error "times p1 p4 changed";

p6 = Powers [Powers [Powers [Coeff 0, Coeff 10], Coeff 100], Coeff 1000];
if times_poly_rec_dense p1 p6 = 
  Powers [Powers [Powers [Coeff 0, Coeff 70], Coeff 700], Coeff 7000]
then () else error "times p1 p6 changed";
*} ML {*
*}

subsubsection {* OK check the base case of times (Powers ps) q *}
ML {*
"~~~~~~ check the base case of times (Powers ps) q ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
"~~m0~~~ fun Powers, args:"; val (Powers ps, q) = (p5, q5);
*} ML {*
p5 = Powers [Coeff 1, Coeff 2, Coeff 3];
q5 = Powers [Coeff 4, Coeff 5];
"~~~~~ fun times_poly_rec_dense, args:"; val (p, q) = (p5, q5);
"~~~~~ fun times, args:"; val (lev, (Powers ps), q) = (0, p, q);
"~~~~~ fun (fn a => ..., args:"; val (a, p) = (nth ps 2, Coeff 0);
val xxx1 = 
  (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))) (nth ps 2) (Coeff 0);
*} ML {*
if xxx1 = Powers [Coeff 12, Coeff 15] then () else error "base case nth 2 changed";
*} ML {*
p = Coeff 0;
inc_power_rec lev p = Coeff 0;
*} ML {*                              
"~~m0f1~~~ fun (fn a => ..., args:"; val (a, p) = (nth ps 1, xxx1);
val xxx2 = 
  (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))) (nth ps 1) xxx1;
*} ML {*
if xxx2 = Powers [Coeff 8, Coeff 22, Coeff 15] then () else error "base case nth 1 changed";
*} ML {*
p = Powers [Coeff 12, Coeff 15];
inc_power_rec lev p = Powers [Coeff 0, Coeff 12, Coeff 15];
*} ML {*
"~~~~~ fun (fn a => ..., args:"; val (a, p) = (nth ps 0, xxx2);
val xxx3 = (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))) (nth ps 0) xxx2;
*} ML {*
if xxx3 = Powers [Coeff 4, Coeff 13, Coeff 22, Coeff 15]                                                                    
then () else error "base case nth 0 changed";
*} ML {*
p = Powers [Coeff 8, Coeff 22, Coeff 15];
inc_power_rec lev p = Powers [Coeff 0, Coeff 8, Coeff 22, Coeff 15]
*} ML {*
p5 = Powers [Coeff 1, Coeff 2, Coeff 3];
q5 = Powers [Coeff 4, Coeff 5];
*} ML {*
if times lev p5 q5 = Powers [Coeff 4, Coeff 13, Coeff 22, Coeff 15]
then () else error "times p5 q5 changed"
*} 

subsubsection {* (z-z) OK check times (3.z^3) (10.x^2.y^2.z^2) *}
ML {*
val px = Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 3];                 (*=== 3.z^3          ===*)
p7 = Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Coeff 10]]];
                                                                      (*=== 10.x^2.y^2.z^2 ===*)
if times_poly_rec_dense px p7 = 
  Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 30]]]
then () else error "times (3.z^3) (10.x^2.y^2.z^2) changed"
*} ML {*
*}

subsubsection {* (z-y) OK check times (2.y^3) (10.x^2.y^2.z^2) *}
ML {*
val px = Powers [Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 2]];        (*=== 2.y^3          ===*)
if times_poly_rec_dense px p7 = 
  Powers [Coeff 0, Coeff 0,
          Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 0,
                  Powers [Coeff 0, Coeff 0, Coeff 20]]]
then () else error "times (2.y^3) (10.x^2.y^2.z^2) changed"
*} ML {*
*}

subsubsection {* (z-x) OK check times (1.x^3) (10.x^2.y^2.z^2) *}
ML {*
val px = Powers [Powers [Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 1]]];(*=== 1.x^3          ===*)
if times_poly_rec_dense px p7 = 
  Powers [Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 10]]]
then () else error "times (1.x^3) (10.x^2.y^2.z^2) changed"
*} ML {*
*}

subsubsection {* (t) OK check times (2.x + 3.y) (4.x + 5.y) *}
ML {*
p2 = Powers [Powers [Coeff 0, Coeff 2], Coeff 3];          (*=== 2.x + 3.y ===*)
q1 = Powers [Powers [Coeff 0, Coeff 4], Coeff 5];          (*=== 4.x + 5.y ===*)
if times_poly_rec_dense p2 q1 = 
  Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15]
then () else error "times (2.x + 3.y) (4.x + 5.y) changed"
*} text {*
(2.x + 3.y).(4.x + 5.y) trace created from ML-block above and copied from Output
         /## times (*4*): lev = 0  (Powers [Powers [Coeff 0, Coeff 2], Coeff 3]) * (Coeff Powers [Powers [Coeff 0, Coeff 4], Coeff 5])... 
          /#### \<lambda> (0)  (Coeff 3)  (Coeff 0). ... 
            ### times (*2*): lev = 1  (Coeff 3) * (Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 12], Coeff 15]) 
          \#### \<lambda> (0)  (Coeff 3)  (Coeff 0). (Powers [Powers [Coeff 0, Coeff 12], Coeff 15]) + (Coeff 0) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 12], Coeff 15]) 
          /#### \<lambda> (0)  (Powers [Coeff 0, Coeff 2])  (Powers [Powers [Coeff 0, Coeff 12], Coeff 15]). ... 
            /## times (*4*): lev = 1  (Powers [Coeff 0, Coeff 2]) * (Coeff Powers [Powers [Coeff 0, Coeff 4], Coeff 5])... 
             /#### \<lambda> (1)  (Coeff 2)  (Coeff 0). ... 
               ### times (*2*): lev = 2  (Coeff 2) * (Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 8], Coeff 10]) 
             \#### \<lambda> (1)  (Coeff 2)  (Coeff 0). (Powers [Powers [Coeff 0, Coeff 8], Coeff 10]) + (Coeff 0) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 8], Coeff 10]) 

             /#### \<lambda> (1)  (Coeff 0)  (Powers [Powers [Coeff 0, Coeff 8], Coeff 10]). ... 
               ### times (*2*): lev = 2  (Coeff 0) * (Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0], Coeff 0]) 
             \#### \<lambda> (1)  (Coeff 0)  (Powers [Powers [Coeff 0, Coeff 8], Coeff 10]). (Powers [Powers [Coeff 0, Coeff 0], Coeff 0]) + (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]) 

            \## times (*4*): lev = 1  (Powers [Coeff 0, Coeff 2]) * (Coeff Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]) 
          \#### \<lambda> (0)  (Powers [Coeff 0, Coeff 2])  (Powers [Powers [Coeff 0, Coeff 12], Coeff 15]). (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]) + (Powers [Coeff 0, Powers [Coeff 0, Coeff 12], Coeff 15]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15]) 
         \## times (*4*): lev = 0  (Powers [Powers [Coeff 0, Coeff 2], Coeff 3]) * (Coeff Powers [Powers [Coeff 0, Coeff 4], Coeff 5]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15]) 
*} ML {*
"~~~~~~ check the first induction step in times (Powers ps) q ~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
(*legend
    mf1_result: folded fun with 2. list element
    mf0: folded fun with 1. list element
    mf0mf1_result: recursive descent from mf0, folded fun with 2. list element
    mf0mf0: recursive descent from mf0, folded fun with 1. list element *)
"~~~~~~~~~ fun times_poly_rec_dense, args:"; val (p, q) = (p2, q1);
"~~m~~~~~~ fun times (*4*), args:"; val (lev, (Powers ps), q) = (0, p, q);
"~~~~~~~~~ fun foldr (*2*), args:"; val (f, x :: xs, b) = 
  ((fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))), ps, (Coeff 0));
"~~mf1~~~~ fun (fn a => ..., args:"; val (a, p) = (nth ps 1, b);
*} ML {*
lev = 0;                                           (* level of recursion in int poly_rec *)
a = Coeff 3;                                       (* nth ps = 1, right most list element *)
p = Coeff 0;                                       (* init 2nd arg of foldr *)
q = Powers [Powers [Coeff 0, Coeff 4], Coeff 5];   (* the multiplicand *)
*} ML {*
if times (lev + 1) a q = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf1 , times <<<<<<<<<<<*)
  Powers [Powers [Coeff 0, Coeff 12], Coeff 15]
then () else error "(2.x + 3.y).(4.x + 5.y): mf1 , times changed";
*} ML {*
if inc_power_rec lev p = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf1 , inc_power_rec <<<*)
  Coeff 0 
then () else error "(2.x + 3.y).(4.x + 5.y): mf1 , inc_power_rec changed";
*} ML {*
val mf1_result = f a p;                            (* keep for further foldr *)
(*...= Powers [Powers [Coeff 0, Coeff 12], Coeff 15]*)
*} ML {*
"~~mf0~~~~ fun (fn a => ..., args:"; val (a, p) = (nth ps 0, mf1_result);
lev = 0;                                           (* level of recursion in int poly_rec *)
a = Powers [Coeff 0, Coeff 2];                     (* = nth ps 0, next and last list element *)
p = Powers [Powers [Coeff 0, Coeff 12], Coeff 15]; (* the intermediate result of foldr *)
q = Powers [Powers [Coeff 0, Coeff 4], Coeff 5];   (* the multiplicand *)
*} ML {*
if times (lev + 1) a q = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0 , times RECURSION <*)
  Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]
then () else error "(2.x + 3.y).(4.x + 5.y): mf0 , times changed";
*} ML {*
if inc_power_rec lev p = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0 , inc_power_rec <<<*)
  Powers [Coeff 0, Powers [Coeff 0, Coeff 12], Coeff 15]
then () else error "(2.x + 3.y).(4.x + 5.y): mf0 , inc_power_rec changed";
*} ML {*
val mf0_inc = inc_power_rec lev p;                   (* keep for return at level 0 *)
*} ML {*
f a p = Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15];
(* ... the final result, but we go INTO the RECURSION of times ===============================: *)
"~~mf0m~~~ fun times (*4*), args:"; val (lev, (Powers ps), q) = (lev + 1, a, q);
*} ML {*
(*/------- foldr_dense goes 2 rounds over [Coeff 0, Coeff 2] ----------------------------------\*)
(*1st argument*)ps = [Coeff 0, Coeff 2];
(*2nd argument*)(Coeff 0);
val mf0m = 
  foldr (fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))) 
    ps (Coeff 0);
mf0m = Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]];
(*\------- foldr_dense goes 2 rounds over [Coeff 0, Coeff 2] ----------------------------------/*)
*} ML {*
"~~~~~~~~~ fun foldr (*2*), args:"; val (f, x :: xs, b) = 
  ((fn a => (fn p => (plus_poly_rec (times (lev + 1) a q) (inc_power_rec lev p)))), ps, (Coeff 0));
"~~mf0mf1~ fun (fn a => ..., args:"; val (a, p) = (nth ps 1, b);
*} ML {*
lev = 1;                                           (* level of recursion in int poly_rec *)
a = Coeff 2;                                       (* nth ps = 1, right most list element *)
p = Coeff 0;                                       (* init 2nd arg of foldr *)
q = Powers [Powers [Coeff 0, Coeff 4], Coeff 5];   (* the multiplicand *)
*} ML {*
if times (lev + 1) a q = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0mf1 , times <<<<<<<<<<<*)
  Powers [Powers [Coeff 0, Coeff 8], Coeff 10]
then () else error "(2.x + 3.y).(4.x + 5.y): mf0mf1 , times changed";
if inc_power_rec lev p = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0mf1 , inc_power_rec <<<*)
  Coeff 0  
then () else error "times (2.x + 3.y).(4.x + 5.y): mf0mf1 inc_power_rec changed";
*} ML {*
val mf0mf1_result = f a p;                         (* keep for further foldr *)
(*... = Powers [Powers [Coeff 0, Coeff 8], Coeff 10]*)
*} ML {*
"~~mf0mf0~ fun (fn a => ..., args:"; val (a, p) = (nth ps 0, mf0mf1_result);
lev = 1;                                           (* level of recursion in int poly_rec *)
a = Coeff 0;                                       (* = nth ps 0, next and last list element *)
p = Powers [Powers [Coeff 0, Coeff 8], Coeff 10];  (* the intermediate result of foldr *)
q = Powers [Powers [Coeff 0, Coeff 4], Coeff 5];   (* the multiplicand *)
*} ML {*
if times (lev + 1) a q = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0mf0 , times <<<<<<<<<<<*)
  Powers [Powers [Coeff 0, Coeff 0], Coeff 0] (* NOT normalise'd internally *)
then () else error "(2.x + 3.y).(4.x + 5.y): mf0mf0 , times changed";
if inc_power_rec lev p = (*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< mf0mf0 , inc_power_rec <<<*)
  Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]
then () else error "(2.x + 3.y).(4.x + 5.y): mf0mf0 , inc_power_rec changed";
*} ML {*
val mf0mf0_result = f a p;                         (* keep for return at level 0: *)
(*...= Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 10]]*)
*} ML {*
"~~mf~~~~~ fun foldr, return:";
if plus_poly_rec mf0mf0_result mf0_inc = 
  Powers [Powers [Coeff 0, Coeff 0, Coeff 8], Powers [Coeff 0, Coeff 22], Coeff 15]
then () else error "(2.x + 3.y).(4.x + 5.y) changed";
*} ML {*
*}

subsubsection {* (y) OK check times (1 + 2.y + 3.z) (10.x^2.y^2.z^2) *}
ML {*
"~~~~~~ check times with 3-variate monom ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
val px = Powers [Powers [Coeff 1, Coeff 2], Coeff 3]          (*===  1 + 2.y + 3.z  ===*);
p7 = Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Coeff 10]]];
                                                              (*===  10.x^2.y^2.z^2 ===*)
*} ML {*
if times_poly_rec_dense px p7 = (* NOT OK *)
  Powers [Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 10], 
                  Powers [Coeff 0, Coeff 0, Coeff 20]],
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 30]]]
then () else error "times_poly_rec_dense (1 + 2.y + 3.z) (10.x^2.y^2.z^2) changed"
*} ML {*
*}

subsubsection {* (x) OK check times (1 + 2.y + 3.x.y) (10.x^2.y^2.z^2) *}
ML {*
"~~~~~~ check times with 3-variate monom ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~";
val px = Powers [Powers [Coeff 1, Powers [Coeff 2, Coeff 3]]] (*===  1 + 2.y + 3.x.y ===*);
p7 = Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Coeff 10]]];
                                                              (*===  10.x^2.y^2.z^2  ===*)
*} ML {*
if times_poly_rec_dense px p7 = 
  Powers [Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 10],
                  Powers [Coeff 0, Coeff 0, Coeff 20, Coeff 30]]]
then () else error "times_poly_rec_dense (1 + 2.y + 3.x.y) (10.x^2.y^2.z^2) changed"
*} ML {*
*}
 
subsubsection {* (w) OK check times (1 + 2.x + 3.z) (10.x^2.y^2.z^2) *}
ML {*
val px = Powers [Powers [Powers [Coeff 1, Coeff 2]], Coeff 3]; (*===  1 + 2.x + 3.z   ===*)
*} ML {*
if times_poly_rec_dense px p7 = 
  Powers [Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 10, Coeff 20]],
          Powers [Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 30]]]
then () else error "times_poly_rec_dense (1 + 2.x + 3.z) (10.x^2.y^2.z^2) changed"
*} ML {*
*}
subsubsection {* (v) OK check times (10.x^2.y^2.z^2) (10.x^2.y^2.z^2) *}
ML {*
p7 = Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Powers [Coeff 0, Coeff 0, Coeff 10]]];
*} ML {*
if times_poly_rec_dense p7 p7 = 
  Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, 
          Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, 
                  Powers [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff 100]]]
then () else error "times_poly_rec_dense (10.x^2.y^2.z^2) (10.x^2.y^2.z^2) changed"
*} ML {*
*}

subsubsection {* (u) OK check times (1 + 2.x + 3.y + 4.z)*(1 + 2.x + 3.y + 4.z) *}
ML {*
q7 = Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]; (*=== 1 + 2.x + 3.y + 4.z ===*)
*} ML {*
if times_poly_rec_dense q7 q7 =           (* (R[x,y])[z] *)
  Powers [                                (* *)
    Powers [                              (* z^0. *)
      Powers [Coeff 1, Coeff 4, Coeff 4], (* z^0.y^0.(x^0.1 + x^1.4 + x^2.4) *)
      Powers [Coeff 6, Coeff 12],         (* z^0.y^1.(x^0.6 + x^1.12) *)
      Coeff 9],                           (* z^0.y^2. x^0.9 *)
    Powers [                              (* z^1. *)
      Powers [Coeff 8, Coeff 16],         (* z^1.y^0.(x^0.8 + x^1.16) *)
      Coeff 24],                          (* z^1.y^1. x^0.24 *)
    Coeff 16]                             (* z^2.y^0. x^0.16 *)
then () else error "times_poly_rec_dense (1 + 2.x + 3.y + 4.z)*(1 + 2.x + 3.y + 4.z) changed"
*} text {*
/## times (*4*): lev = 0  (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4])... 
 /#### \<lambda> (0)  (Coeff 4)  (Coeff 0). ... 
   ### times (*2*): lev = 1  (Coeff 4) * (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) 
                     ...inc_power_rec (*1*) 0 (Coeff 0) \<longrightarrow> (Coeff 0) 
 \#### \<lambda> (0)  (Coeff 4)  (Coeff 0). (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) + (Coeff 0) \<longrightarrow> (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) 
 /#### \<lambda> (0)  (Powers [Powers [Coeff 1, Coeff 2], Coeff 3])  (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]). ... 
   /## times (*4*): lev = 1  (Powers [Powers [Coeff 1, Coeff 2], Coeff 3]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4])... 
    /#### \<lambda> (1)  (Coeff 3)  (Coeff 0). ... 
      ### times (*2*): lev = 2  (Coeff 3) * (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]) 
                     ...inc_power_rec (*3*) 1 (Coeff 0) \<longrightarrow> (Coeff 0) 
    \#### \<lambda> (1)  (Coeff 3)  (Coeff 0). (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]) + (Coeff 0) \<longrightarrow> (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]) 
    /#### \<lambda> (1)  (Powers [Coeff 1, Coeff 2])  (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]). ... 
      /## times (*4*): lev = 2  (Powers [Coeff 1, Coeff 2]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4])... 
       /#### \<lambda> (2)  (Coeff 2)  (Coeff 0). ... 
         ### times (*2*): lev = 3  (Coeff 2) * (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) 
                     ...inc_power_rec (*3*) 2 (Coeff 0) \<longrightarrow> (Coeff 0) 
       \#### \<lambda> (2)  (Coeff 2)  (Coeff 0). (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) + (Coeff 0) \<longrightarrow> (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) 
       /#### \<lambda> (2)  (Coeff 1)  (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]). ... 
         ### times (*2*): lev = 3  (Coeff 1) * (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) 
                     ...inc_power_rec (*2*) 0 (Powers [Coeff 2, Coeff 4]) \<longrightarrow> (Powers [Coeff 0, Coeff 2, Coeff 4]) 
                     ...inc_power_rec (*1*) 0 (Coeff 6) \<longrightarrow> (Powers [Coeff 0, Coeff 6]) 
                     ...inc_power_rec (*4*) 1 (Powers [Powers [Coeff 2, Coeff 4], Coeff 6]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]]) 
                     ...inc_power_rec (*1*) 0 (Coeff 8) \<longrightarrow> (Powers [Coeff 0, Coeff 8]) 
                     ...inc_power_rec (*3*) 1 (Coeff 8) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 8]]) 
                     ...inc_power_rec (*4*) 2 (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) \<longrightarrow> (Powers [Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]], Powers [Powers [Coeff 0, Coeff 8]]]) 
       \#### \<lambda> (2)  (Coeff 1)  (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]). (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) + (Powers [Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]], Powers [Powers [Coeff 0, Coeff 8]]]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 3, Coeff 6]], Powers [Powers [Coeff 4, Coeff 8]]]) 
      \## times (*4*): lev = 2  (Powers [Coeff 1, Coeff 2]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 3, Coeff 6]], Powers [Powers [Coeff 4, Coeff 8]]]) 
                     ...inc_power_rec (*2*) 0 (Powers [Powers [Coeff 3, Coeff 6], Coeff 9]) \<longrightarrow> (Powers [Coeff 0, Powers [Coeff 3, Coeff 6], Coeff 9]) 
                     ...inc_power_rec (*1*) 0 (Coeff 12) \<longrightarrow> (Powers [Coeff 0, Coeff 12]) 
                     ...inc_power_rec (*4*) 1 (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]) \<longrightarrow> (Powers [Powers [Coeff 0, Powers [Coeff 3, Coeff 6], Coeff 9], Powers [Coeff 0, Coeff 12]]) 
    \#### \<lambda> (1)  (Powers [Coeff 1, Coeff 2])  (Powers [Powers [Powers [Coeff 3, Coeff 6], Coeff 9], Coeff 12]). (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 3, Coeff 6]], Powers [Powers [Coeff 4, Coeff 8]]]) + (Powers [Powers [Coeff 0, Powers [Coeff 3, Coeff 6], Coeff 9], Powers [Coeff 0, Coeff 12]]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 6, Coeff 12], Coeff 9], Powers [Powers [Coeff 4, Coeff 8], Coeff 12]]) 
   \## times (*4*): lev = 1  (Powers [Powers [Coeff 1, Coeff 2], Coeff 3]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 6, Coeff 12], Coeff 9], Powers [Powers [Coeff 4, Coeff 8], Coeff 12]]) 
                     ...inc_power_rec (*2*) 0 (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) \<longrightarrow> (Powers [Coeff 0, Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) 
 \#### \<lambda> (0)  (Powers [Powers [Coeff 1, Coeff 2], Coeff 3])  (Powers [Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]). (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 6, Coeff 12], Coeff 9], Powers [Powers [Coeff 4, Coeff 8], Coeff 12]]) + (Powers [Coeff 0, Powers [Powers [Coeff 4, Coeff 8], Coeff 12], Coeff 16]) \<longrightarrow> (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 6, Coeff 12], Coeff 9], Powers [Powers [Coeff 8, Coeff 16], Coeff 24], Coeff 16]) 
\## times (*4*): lev = 0  (Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) * (Coeff Powers [Powers [Powers [Coeff 1, Coeff 2], Coeff 3], Coeff 4]) = (Powers [Powers [Powers [Coeff 1, Coeff 4, Coeff 4], Powers [Coeff 6, Coeff 12], Coeff 9], Powers [Powers [Coeff 8, Coeff 16], Coeff 24], Coeff 16]) 
*} ML {*
(* inc_power_rec and the respective trace is correct here in prep.ML.
  Below are the incorrect results found in the trace in Try_Test.thy of 7c2f7716e7f0: *)
inc_power_rec 1 (Coeff 8) = (Powers [Coeff 0, Powers [Coeff 0, Coeff 8]]);
(*                          TOO MUCH ^^^^^^^^^
  ...inc_power_rec (*1*) 0 (Coeff 8) \<longrightarrow> (Powers [Coeff 0, Coeff 8]) 
  ...inc_power_rec (*3*) 1 (Coeff 8) \<longrightarrow> 
                            (Powers [Powers [Coeff 0, Coeff 8]]) 
val it = false: bool*) 
*} ML {*
inc_power_rec 2 (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) = 
  (Powers [Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]], Powers [Coeff 0, Powers [Coeff 0, Coeff 8]]]);
(*                                                                                TOO MUCH ^^^^^^^^^
  ...inc_power_rec (*2*) 0 (Powers [Coeff 2, Coeff 4]) \<longrightarrow> (Powers [Coeff 0, Coeff 2, Coeff 4]) 
  ...inc_power_rec (*1*) 0 (Coeff 6) \<longrightarrow> (Powers [Coeff 0, Coeff 6]) 
  ...inc_power_rec (*4*) 1 (Powers [Powers [Coeff 2, Coeff 4], Coeff 6]) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]]) 
  ...inc_power_rec (*1*) 0 (Coeff 8) \<longrightarrow> (Powers [Coeff 0, Coeff 8]) 
  ...inc_power_rec (*3*) 1 (Coeff 8) \<longrightarrow> (Powers [Powers [Coeff 0, Coeff 8]]) 
  ...inc_power_rec (*4*) 2 (Powers [Powers [Powers [Coeff 2, Coeff 4], Coeff 6], Coeff 8]) \<longrightarrow> 
  (Powers [Powers [Powers [Coeff 0, Coeff 2, Coeff 4], Powers [Coeff 0, Coeff 6]], Powers [Powers [Coeff 0, Coeff 8]]]) 
val it = false: bool*)
*} ML {*
*}

section {* ML code for multiplication of recursive sparse representation *}
(* SEE Try_Poly_Sparse.thy ...

    definition p014_sparse :: "int mpoly"
    where "p014_sparse = 
      monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 1) 2 +
      monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 2 4) 3"
    value p014_sparse --{* 2 * x\<^sup>1  +  3 * z\<^sup>4  = "Rec 
     (Powers (Sparse (Abs_nat_mapping_sparse [
       (0, Powers (Sparse (Abs_nat_mapping_sparse [
         (0, Powers (Sparse (Abs_nat_mapping_sparse [(Suc (Suc (Suc (Suc 0))), Coeff 3)])))]))), 
       (Suc 0, Coeff 2)])))"*}

... a 
  datatype 'a xxx = int * Coeff of 'a | int * Powers of ('a poly_rec) list

is not possible in SML *)

subsection {**}
ML {*
"~~~~~ fun , args:"; val () = ();
*} ML {*
*}

end
