theory Quotients
imports Main
begin

-- "---------------------------------------------------------------------"
-- "subtypes"
-- "---------------------------------------------------------------------"

-- "the non-zero integers"
typedef nzint = "{i::int. i\<noteq>0}"
  morphisms Rep_nzint Abs_nzint
by (metis (mono_tags) mem_Collect_eq of_bool_eq(1) of_bool_eq_iff)
print_theorems

setup_lifting type_definition_nzint
print_theorems

-- "lift_definition must prove that subtype invariant is preserved"

lift_definition multiply :: "nzint \<Rightarrow> nzint \<Rightarrow> nzint"
  is "op *"
  by auto (* \<And>nat1 nat2. 0 < nat1 \<Longrightarrow> 0 < nat2 \<Longrightarrow> 0 < nat1 * nat2  *)
print_theorems

export_code multiply in SML

-- "---------------------------------------------------------------------"
-- "total quotients"
-- "---------------------------------------------------------------------"

(*declare [[show_types]]*)

quotient_type rat = "int * nzint" / "\<lambda>(x1,x2).\<lambda>(y1,y2). x1*(Rep_nzint y2)=y1*(Rep_nzint x2)"
morphisms pair abs_rat 
proof (unfold "equivp_def")
  show "\<forall>x y. (case x of (x1, x2) \<Rightarrow> \<lambda>(y1, y2). x1 * Rep_nzint y2 = y1 * Rep_nzint x2) y =
          ((case x of (x1, x2) \<Rightarrow> \<lambda>(y1, y2). x1 * Rep_nzint y2 = y1 * Rep_nzint x2) =
           (case y of (x1, x2) \<Rightarrow> \<lambda>(y1, y2). x1 * Rep_nzint y2 = y1 * Rep_nzint x2))"
  proof (auto)
    fix a::int and b::nzint and aa::int and ba::nzint
    assume 1: "a * Rep_nzint ba = aa * Rep_nzint b"
    show "(\<lambda>(y1, y2). a * Rep_nzint y2 = y1 * Rep_nzint b) = 
          (\<lambda>(y1, y2). aa * Rep_nzint y2 = y1 * Rep_nzint ba)" 
    proof (intro ext)
      fix x::"int \<times> nzint"
      show "(case x of (y1, y2) \<Rightarrow> a * Rep_nzint y2 = y1 * Rep_nzint b) =
            (case x of (y1, y2) \<Rightarrow> aa * Rep_nzint y2 = y1 * Rep_nzint ba)" 
      proof (cases x)
        fix ab bb
        assume 2: "x = (ab, bb)"
        show ?thesis
        proof -
          from Rep_nzint have z1: "Rep_nzint b \<noteq> 0" by auto
          from Rep_nzint have z2: "Rep_nzint ba \<noteq> 0" by auto
          from Rep_nzint have z3: "Rep_nzint bb \<noteq> 0" by auto
          let ?eq = "aa * Rep_nzint bb * Rep_nzint b = ab * Rep_nzint ba * Rep_nzint b"
          have "(a * Rep_nzint bb = ab * Rep_nzint b) = 
                (aa * Rep_nzint bb = ab * Rep_nzint ba)" (is "?assumption = ?goal")
          proof (cases)
            assume 3: "?assumption"
            have 4: "?goal"
            proof -
              from 1 3 have "?eq" 
                by (metis mult.assoc mult.commute mult.left_commute)
              from this z1 show "?goal" by auto
            qed
            from this 3 4 show ?thesis by auto
          next
            assume 3: "\<not>?assumption"
            have 4: "\<not>?goal"
            proof
              assume 5: "?goal"
              from this have "?eq" by auto
              from this 1 z1 z2 z3 have "?assumption"
                by (metis mult.left_commute mult_left_cancel mult_right_cancel mult_zero_left)
              from this 3 show False by auto
            qed
            from this 3 show ?thesis by auto
          qed
          from this 2 show ?thesis by auto
        qed
      qed
    qed
  next
    fix a b aa ba
    assume "(\<lambda>(y1, y2). a * Rep_nzint y2 = y1 * Rep_nzint b) = 
            (\<lambda>(y1, y2). aa * Rep_nzint y2 = y1 * Rep_nzint ba)"
    from this show "a * Rep_nzint ba = aa * Rep_nzint b"
      by (metis (mono_tags, lifting) old.prod.case) 
  qed
qed
print_theorems

lift_definition add :: "rat \<Rightarrow> rat \<Rightarrow> rat"
is "(\<lambda>(x1::int,x2::nzint).\<lambda>(y1::int,y2::nzint). 
       let x3 = Rep_nzint x2 in 
       let y3 = Rep_nzint y2 in
       (x1*y3+y1*x3, Abs_nzint(x3*y3)))"
proof (auto)
  fix a b aa ba ab bb ac bc ad bd ae be
  assume 0: "a * Rep_nzint ba = aa * Rep_nzint b"
  assume 1: "ab * Rep_nzint bc = ac * Rep_nzint bb"
  assume 2: "(ad, bd) = (let x3 = Rep_nzint b; y3 = Rep_nzint bb in (a * y3 + ab * x3, Abs_nzint (x3 * y3)))"
  assume 3: "(let x3 = Rep_nzint ba; y3 = Rep_nzint bc in (aa * y3 + ac * x3, Abs_nzint (x3 * y3))) = (ae, be)"
  show  "ad * Rep_nzint be = ae * Rep_nzint bd"
  proof -
    from 2 have 4: "ad = a * (Rep_nzint bb) + ab * (Rep_nzint b)" by (metis prod.sel(1)) 
    from 2 have 5: "bd = Abs_nzint((Rep_nzint b)*(Rep_nzint bb))" by (metis prod.sel(2))
    from 3 have 6: "ae = aa * (Rep_nzint bc) + ac * (Rep_nzint ba)" by (metis prod.sel(1))
    from 3 have 7: "be = Abs_nzint((Rep_nzint ba)*(Rep_nzint bc))" by (metis prod.sel(2))
    from 5 have 8: "Rep_nzint bd = (Rep_nzint b)*(Rep_nzint bb)" by (metis Rep_nzint_inverse multiply.rep_eq) 
    from 7 have 9: "Rep_nzint be = (Rep_nzint ba)*(Rep_nzint bc)" by (metis Rep_nzint_inverse multiply.rep_eq) 
    from Rep_nzint have z1: "Rep_nzint b \<noteq> 0" by auto
    from Rep_nzint have z2: "Rep_nzint ba \<noteq> 0" by auto
    from Rep_nzint have z3: "Rep_nzint bb \<noteq> 0" by auto
    from Rep_nzint have z4: "Rep_nzint bc \<noteq> 0" by auto
    from Rep_nzint have z5: "Rep_nzint bd \<noteq> 0" by auto
    from Rep_nzint have z6: "Rep_nzint be \<noteq> 0" by auto
    have
    "(a * Rep_nzint bb + ab * Rep_nzint b)*(Rep_nzint ba)*(Rep_nzint bc) =
     (aa * Rep_nzint bc + ac * Rep_nzint ba)*(Rep_nzint b)*(Rep_nzint bb)" (is "?goal0")
    proof -
      have
        "a * Rep_nzint bb * Rep_nzint ba * Rep_nzint bc + 
         ab * Rep_nzint b * Rep_nzint ba * Rep_nzint bc =
         aa * Rep_nzint bc * Rep_nzint b * Rep_nzint bb + 
         ac * Rep_nzint ba * Rep_nzint b * Rep_nzint bb" (is "?goal1")
      proof -
        from 0 1 have
          "(a * Rep_nzint ba)*(Rep_nzint bb * Rep_nzint bc) +
           (ab * Rep_nzint bc)*(Rep_nzint b * Rep_nzint ba) =
           (aa * Rep_nzint b)*(Rep_nzint bc * Rep_nzint bb) +
           (ac * Rep_nzint bb)*(Rep_nzint ba * Rep_nzint b)" (is "?goal2")
        by (metis "0" "1" mult.commute)
        from this show ?goal1 by (metis (erased, hide_lams) mult.commute mult.left_commute) 
      qed
      from this show ?goal0 by (metis (erased, hide_lams) distrib_right) 
    qed
    from this 4 6 8 9 show ?thesis by (metis mult.assoc)
  qed
qed
print_theorems

-- "apparently code generation not set up by quotient_type"
export_code add in SML

end
