(* Notes on approaching pretty printing in the polynomial package 
   WN170217: works with Isabelle2016, NOT YET with Isabelle2016-1 *)

theory Towards_Pretty_Printing 
imports Main
begin
section \<open>Preliminaries\<close>
subsection \<open>Survey on Forian's milestones\<close>
text \<open>
On 2016-06-02 09:39, Florian Haftmann wrote:

Meilenstein 1: Festlegen der konkreten Syntax

Ich weiß nicht, ob hier schon etwas passiert ist; aber man kann
an sowas denken wie

[4 * x_1^2 x_2 + 7 * x_3]

mit geeigneten Klammern [ ] und Variablen x_n. Will man die Bezeichner
für die Variablen selbst wählen können, muss man irgendwo dranschreiben,
was die Reihenfolge der Variablen ist. Braucht man, denke ich, eher nicht.

Meilenstein 2: Abstrakte Darstellung in ML --> Konkrete Syntax als
Isabelle-String

Man definiert in ML einen abstrakten Datentyp, der multivariate Polynome
repräsentiert. Dann schreibt man eine ML-Funktion, die solche Werte in
druckbare (writeln) Isabelle-Strings transformiert. Dabei kann man die
ganzen witzigen Kontrollsymbole \(*WN:disturbs parser HERE*)<^isub> etc. nutzen.

Meilenstein 3: Term in ML --> Konkrete Syntax als Isabelle-String

Man schreibt eine Funktion, die einen ML-Term, der ein Polynom
darstellt, in den abstrakten Datentyp übersetzt. Mit dem Ergebnis aus
Meilenstein 3 bekommt man eine einfache Print-Funktion vom Typ term ->
string

Meilenstein 4: Term in ML --> Konkrete Syntax als ML-Datentyp

Hier wird festgelegt, wie die konkrete Syntax konzeptionell
repräsentiert werden soll; das ist wichtig, um ein Gefühl dafür zu
bekommen, wie der nächste Meilenstein funktioniert.

Die relevanten Abschnitte sind \<section>8.5. \<guillemotright>Syntax transformations\<guillemotleft> in
isar-ref und \<section>3 \<guillemotright>Concrete syntax and type-checking\<guillemotleft> in implementation.


Meilenstein 5: Konkrete Syntax in Isabelle definieren

Hier wird es jetzt sehr spezifisch. Spezielle konkrete Syntax in
Isabelle kann z. B. so funktionieren, dass man sich eigene syntaktische
Konstanten mit \<guillemotright>syntax\<guillemotleft> deklariert und diese geeignet mit den logischen
abstrakten Termen in Beziehung setzt (\<guillemotright>translations\<guillemotleft>,
\<guillemotright>print_translation\<guillemotleft>, \<guillemotright>parse_translation\<guillemotleft>). Ein einfaches Beispiel ist
die Listen-Syntax am Beginn von List.thy. Es gibt aber auch andere
Möglichkeiten.

Was man am besten macht, ist aus meiner Sicht derzeit nicht absehbar;
dazu braucht man das Ergebnis aus Meilenstein 4.

Meilenstein 6: Konkrete Syntax mit logischer Darstellung verbinden

Meilenstein 7: Optimierung: überflüssig gewordene Zwischenschichten
(abstrakte Datentypen in ML) entfernen.
\<close> 
subsection \<open>Survey: Isabelle components -- milestones\<close>
text \<open>
The milestones serve an approach for novices and go a detour towards 
Isabelle's sophisticated mechanisms.

Isabelle component      Polynomial Package      milestones
-----------------------+-----------------------+-----------------------
Isabelle/Isar           abstract poly
|                       |
v                       v
Isabelle/ML             executable poly         (1)(2)      (6)
|                       |                           |        |
v                       v                           |        |
(Lambda-)term                                       | (3)(4) |
|                       |                           |  |  |  |
v                       v                           |  |  |  |
ast                                                 |  |  v  |
syntax                                              |  | (5) |
|                       |                           |  |  |  |
v                       v                           v  v  v  v
(Isabelle-)string       human readable poly
:
^ This is how Isabelle components convert abstract data to strings, see isar-ref p.181.
For the Polynomial Package these conversions are re-used.
The milestones approach the above chain of conversions in steps (1..6).
\<close>
subsection \<open>Pointers into the documentation\<close>
text \<open>
A valuable source is ~~/src/Pure/library.ML
Some of the addressed datatypes are discussed in implementation.pdf, 0.7 Basic data types 

probably useful Isabelle documentation:

# implementation.pdf:
## 0.1 Style and orthography
## 0.2.3 Antiquotations
## 0.2.4 Printing ML values
## 0.3.1 Forward application and composition
## 0.3.2 Canonical iteration
## 0.5 Exceptions
## 0.6 Strings of symbols
## 0.7.2 Strings
## 0.7.3 Integers
## 0.7.5 Options
## 0.7.6 Lists

## 3 \<guillemotright>Concrete syntax and type-checking\<guillemotleft>               ...Florian Haftmann
## 3.1 Reading and pretty printing
## 3.2 Parsing and unparsing

# isar-ref.pdf
## 8.1.3 Alternative print modes
## 8.2 Mixfix annotations

## 8.4.1 Lexical matters
## 8.4.3 The Pure grammar
## 8.4.4 Inspecting the syntax
## 8.5 Syntax transformations: \<guillemotright>print_translation\<guillemotleft> etc ...Florian Haftmann
\<close>

section \<open>Hints for the milestones\<close>
subsection \<open>Milestone 1: definition of concrete syntax\<close>
text \<open>
Example: "{|| 4 * x_1 ^ 5 * x_2 + 6 * x_3 ||}"
Notes:                    ^
(1) this                  ^ is questionable: it separates two elements of a monomial !
(2) specific delimitiers like "[[" and "]]" will be required for parsing.
\<close>

subsection \<open>Milestone 2: abstract poly in ML --> Isabelle string\<close>
text \<open>
The "abstract poly" is implemented in sparse distributive representation.

~~/src/Pure/library.ML is used as much as possible.
\<close> 
subsubsection \<open>Isabelle strings\<close>
text \<open>see implementation.pdf:
## 0.6 Strings of symbols
## 0.7.2 Strings\<close>
 
subsubsection \<open>Abstract polynomial\<close>
ML \<open>(* this turns out to be too little specific *)
fun sort_poly _ _ = LESS
type monom = (int * int) list
type poly = ((int * monom) list * (monom -> monom -> order))
\<close> ML \<open>
val abspoly = ([(4,[(1, 5), (2, 1)]), (6,[(3, 1)])], sort_poly) : poly
\<close> ML \<open>(* make structure more explicit for print translation *)
datatype monom = Monom of (int * int) list
datatype poly = Poly of ((int * monom) list * (monom -> monom -> order))
\<close> ML \<open>
val abspoly = Poly ([(4, Monom [(1, 5), (2, 1)]), (6, Monom [(3, 1)])], sort_poly) : poly
(* probably we can drop "sort_poly" within our tasks *)
\<close> ML \<open>
Poly ([], sort_poly);
Poly ([(6, Monom [(3, 1)])], sort_poly);
Poly ([(6, Monom [])], sort_poly);
\<close> ML \<open>

\<close> ML \<open>
\<close> ML \<open>
\<close> 
subsubsection \<open>Orders\<close>
ML \<open>
EQUAL; GREATER; LESS : order; (* from src/Pure/General.xml *)
sort; (* from ~~/src/Pure/library.ML *)
\<close> ML \<open>
\<close> ML \<open>
fun comp_pairs (p1, p2) = if fst p1 < fst p2 then LESS else GREATER
fun sort_monoms ms = sort comp_pairs ms
\<close> ML \<open>
sort_monoms [(77,1), (1, 5), (2, 1)]
(* in abstract poly this is given by nat, so we will never write unsorted monoms *)
\<close> ML \<open>
val m1 = Monom ([(1,2),(2,3),(3,1)])
val m2 = Monom ([(1,3),(2,4),(3,2)])
val m2 = Monom ([(1,33)])
\<close> ML \<open>
val e1 = [2,3,1]
val e2 = [33]
\<close> ML \<open>
fun ord (m, n) =
  if m < n then LESS
  else if m = n then EQUAL
  else GREATER
\<close> ML \<open>
list_ord ord (e1, e2) (* = GREATER *)
\<close> ML \<open>
val e2 = [3,4,5];
list_ord ord (e1, e2) (* = GREATER *)
\<close> ML \<open>
\<close> ML \<open>
sort (* TODO, NOT required for pretty printing ! *)
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close> 
subsubsection \<open>Pretty printing\<close>
ML \<open>
(* TODO *)
\<close> ML \<open>
abspoly (*\<longrightarrow> "{|| 4 * x_1 ^ 5 * x_2 + 6 * x_3 ||}"*)
\<close> ML \<open>
fun var_exp (v, e) = "x_" ^ string_of_int v ^ " ^ " ^ string_of_int e
\<close> ML \<open>
fun monom [] = ""
  | monom [ve] = var_exp ve
  | monom (ve :: ves) = var_exp ve ^ "*" ^ monom ves
\<close> ML \<open>
monom [(1,5), (2, 1)]
\<close> ML \<open>
fun coeff_monom (1, Monom m) = "1 * " ^ monom m
  | coeff_monom (c, Monom m) = string_of_int c ^ " * " ^ monom m
\<close> ML \<open>
val ms = [(4, Monom [(1, 5), (2, 1)]), (6, Monom [(3, 1)])];
val pfusch = map coeff_monom ms
\<close> ML \<open>
fold;
fold (curry op ^) pfusch ""
\<close> ML \<open>
fun string_of_poly (Poly (ms, _)) =
  let
    val pfusch = map coeff_monom ms
  in fold (curry op ^) pfusch " + " end
\<close> ML \<open>
string_of_poly abspoly = "6 * x_3 ^ 14 * x_1 ^ 5*x_2 ^ 1 + "
\<close> ML \<open>
fold
\<close> ML \<open>
\<close>

subsection \<open>Milestone 3: term in ML --> Isabelle string\<close>

--\<open>this is Isar proof language ...\<close>
consts (*isar-ref.pdf p.191: Only constant atoms — constructor Ast.Constant for ASTs 
  and Const for terms can invoke translation functions*)
  Poly :: "((('a \<times> (int \<times> int) list) list) \<times> 'b) \<Rightarrow> 'c"
  Monom :: "(int \<times> int) list \<Rightarrow> 'd"
--\<open>the antiquotation "@ {term}" connects the Isar proof language with ML ...}\<close>
ML \<open>
val termpoly =
  @{term "Poly ((*[*) [(4::int, Monom [(1::int, 5::int), (2, 1)]), (6, Monom [(3, 1)]) (*]*)], sort_poly)"} : term
\<close> 
declare [[ML_print_depth=99]]
ML \<open>
writeln (@{make_string} termpoly)
\<close> ML \<open>(* we eliminate all types from the output of above and use the remainder as pattern
which is useful as model for coding specific functions ...*)
val (Const ("Towards_Pretty_Printing.Poly", _) $
     (Const ("Product_Type.Pair", _) $
       (Const ("List.list.Cons", _) $
         (Const ("Product_Type.Pair", _) $
           (Const ("Num.numeral_class.numeral", _) $
             (Const ("Num.num.Bit0", _) $
               (Const ("Num.num.Bit0", _) $ Const ("Num.num.One", _)))) $
           (Const ("Towards_Pretty_Printing.Monom", _) $
             (Const ("List.list.Cons", _) $
               (Const ("Product_Type.Pair", _) $ Const ("Groups.one_class.one", _) $
                 (Const ("Num.numeral_class.numeral", _) $
                   (Const ("Num.num.Bit1", _) $
                     (Const ("Num.num.Bit0", _) $ Const ("Num.num.One", _))))) $
               (Const ("List.list.Cons", _) $
                 (Const ("Product_Type.Pair", _) $
                   (Const ("Num.numeral_class.numeral", _) $
                     (Const ("Num.num.Bit0", _) $ Const ("Num.num.One", _))) $
                   Const ("Groups.one_class.one", _)) $
                 Const ("List.list.Nil", _))))) $
         (Const ("List.list.Cons", _) $
           (Const ("Product_Type.Pair", _) $
             (Const ("Num.numeral_class.numeral", _) $
               (Const ("Num.num.Bit0", _) $
                 (Const ("Num.num.Bit1", _) $ Const ("Num.num.One", _)))) $
             (Const ("Towards_Pretty_Printing.Monom", _) $
               (Const ("List.list.Cons", _) $
                 (Const ("Product_Type.Pair", _) $
                   (Const ("Num.numeral_class.numeral", _) $
                     (Const ("Num.num.Bit1", _) $ Const ("Num.num.One", _))) $
                   Const ("Groups.one_class.one", _)) $
                 Const ("List.list.Nil", _)))) $
           Const ("List.list.Nil", _))) $
       Free ("sort_poly", _))) = termpoly
\<close> ML \<open>
val (Const ("Towards_Pretty_Printing.Poly", _) $ _) = termpoly;
val (Const (str, _) $ _) = termpoly;

\<close> ML \<open>
@{term 6} (* = Const ("Num.numeral_class.numeral", "num \<Rightarrow> 'a") $
           (Const ("Num.num.Bit0", "num \<Rightarrow> num") $ (Const ("Num.num.Bit1", "num \<Rightarrow> num") $ Const ("Num.num.One", "num"))): term *)
\<close> ML \<open>
Const;
op  $ ;
\<close> ML \<open>
@{term "6::nat"} (* = Const ("Groups.one_class.one", "'a"): term *)
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
(* conversions already there in Isabelle *)
val one = @{term "1::int"};
val fiv = @{term "6::int"};
HOLogic.dest_number one;
HOLogic.dest_number fiv;
\<close> ML \<open>
\<close> ML \<open>
op |>;
1 |> (curry op+ 1) |> (curry op+ 333)
\<close> ML \<open>
(*"<sub>", "</sub>"*)
(* "|>" is preferred to "string_of_int (snd (HOLogic.dest_number v))" *)
(* TODO v,e with only 1 digit, sufficient for demo; otherwise \<^bsub>", "\<^esub>" ?!?*)
fun var_exp (Const ("Product_Type.Pair", _) $ v $ e) =
  let 
    val v = v |> HOLogic.dest_number |> snd
    val e = e |> HOLogic.dest_number |> snd
  in
    "x\<^sub>" ^ string_of_int v ^ 
    (if e > 1 then "\<^sup>" ^ string_of_int e else "")
  end
  (*"x\<^bsub>" ^ (v |> HOLogic.dest_number |> snd |> string_of_int) ^ "\<^esub>" ^*)
  (* "\<^bsup>" ^ (e |> HOLogic.dest_number |> snd |> string_of_int) ^ "\<^esup>"*)
  | var_exp _ = error "var_exp"
\<close> ML \<open>
(* we immediately test the function *)
val v_e = @{term "(1::int, 5::int)"};
var_exp v_e = "x\<^sub>1\<^sup>5"
\<close> ML \<open>
fun mono (Const ("List.list.Cons", _) $ pair $ Const ("List.list.Nil", _)) = var_exp pair
  | mono (Const ("List.list.Cons", _) $ pair $ tail) =  var_exp pair ^ "." ^ mono tail
  | mono _ = error "ERROR mono"
\<close> ML \<open>
fun monom (Const("Towards_Pretty_Printing.Monom", _) $ m) = mono m
  | monom _ = error "ERROR monom"
\<close> ML \<open>
(* we immediately test the function *)
val ms = @{term "Monom [(1::int, 5::int), (2, 1)]"};
\<close> ML \<open>
monom ms = "x\<^sub>1\<^sup>5.x\<^sub>2"
\<close> ML \<open>
fun Free_to_str (Free (id, _)) = id
  | Free_to_str _ = error "ERROR Free_to_str"
\<close> ML \<open>
fun coeff_monom (Const ("Product_Type.Pair", _) $ c $ m) = 
    (c |> HOLogic.dest_number |> snd |> string_of_int) ^ "." ^ monom m
  | coeff_monom e = error ("ERROR coeff_monom: " ^ PolyML.makestring e)
\<close> ML \<open>
fun poly_to_str (Const ("List.list.Cons", _) $ c_m $ Const ("List.list.Nil", _)) =
    coeff_monom c_m
  | poly_to_str (Const ("List.list.Cons", _) $ c_m $ tail) = 
    coeff_monom c_m ^ " + " ^ poly_to_str tail
  | poly_to_str _ = error "ERROR mono"
\<close> ML \<open>
fun termpoly_to_string (Const ("Towards_Pretty_Printing.Poly", _) $
     (Const ("Product_Type.Pair", _) $ p $ s)) =
       "Poly (" ^ poly_to_str p ^ ", " ^ Free_to_str s ^ ")"
  | termpoly_to_string _ = error "ERROR termpoly_to_string"
\<close> ML \<open>
termpoly_to_string termpoly = "Poly (4.x\<^sub>1\<^sup>5.x\<^sub>2 + 6.x\<^sub>3, sort_poly)"
\<close>
subsection \<open>Milestone 4: term in ML --> concrete syntax as ML datatype\<close>
text \<open>
probably useful models in these sources:
~~/src/Pure/Syntax/ast.ML          datatype ast
~~/src/Pure/Syntax/local_syntax.ML datatype T = Syntax of ...
~~/src/Pure/Syntax/syntax.ML       datatype syntax
# isar-ref.pdf       \<section>8.5. \<guillemotright>Syntax transformations\<guillemotleft> in
# implementation.pdf \<section>3 \<guillemotright>Concrete syntax and type-checking\<guillemotleft> 
\<close> 
subsubsection \<open>Conversion term -> AST by ourselves\<close>
ML \<open>
val ctxt = Proof_Context.init_global @{theory}
\<close> text \<open>copy from ~~/src/Pure/Syntax/ast.ML
  datatype ast =
    Constant of string |
    Variable of string |
    Appl of ast list
\<close> ML \<open>
(* we mimic these Isabelle functions *)
Syntax.string_of_term; (*= fn: Proof.context -> term -> string*)
Syntax.unparse_term    (*= fn: Proof.context -> term -> Pretty.T ??? HOW HANDLE Pretty.T ???*)
\<close> ML \<open>
(*v!v*) (Syntax.string_of_term ctxt termpoly); (* \<rightarrow> AST \<rightarrow> AST \<rightarrow> string *)
writeln; (* = fn: string -> unit*)
\<close> ML \<open>
writeln (Syntax.string_of_term ctxt termpoly); (* THIS WE WANT TO CHANGE:
  Poly ([(4, Monom [(1, 5), (2, 1)]), (6, Monom [(3, 1)])], sort_poly) *)
\<close> text \<open>copy from ~~/src/Pure/term.ML
  datatype term =
    Const of string * typ |
    Free of string * typ |
    Var of indexname * typ |
    Bound of int |
    Abs of string * typ * term |
    $ of term * term
\<close> ML \<open>
(* we prepare for exploring operations on ASTs *)
fun ast_of_term (Const (id, _)) = Ast.Constant id
  | ast_of_term (Free (id, _)) = Ast.Constant id
  | ast_of_term (Var ((str, i), _)) = Ast.Constant ("Var (" ^ str ^ ", " ^ string_of_int i ^ ")")
  | ast_of_term (Bound i) = Ast.Constant ("Bound " ^ string_of_int i)
  | ast_of_term (Abs (str, _, t)) = Ast.Appl (Ast.Constant ("Abs " ^ str) :: [ast_of_term t])
  | ast_of_term (t1 $ t2) = Ast.Appl [ast_of_term t1, ast_of_term t2]
\<close> ML \<open>
ast_of_term: term -> Ast.ast
\<close> ML \<open>
val astpoly = ast_of_term termpoly (* ^^^ does Isabelle have "$" pre- or infix? *)
(*val astpoly =
   ("Towards_Pretty_Printing.Poly"
     (("Product_Type.Pair"
        (("List.list.Cons"
           (("Product_Type.Pair"
              ("Num.numeral_class.numeral" ("Num.num.Bit0" ("Num.num.Bit0" "Num.num.One"))))
             ("Towards_Pretty_Printing.Monom"
               (("List.list.Cons"
                  (("Product_Type.Pair" "Groups.one_class.one")
                    ("Num.numeral_class.numeral"
                      ("Num.num.Bit1" ("Num.num.Bit0" "Num.num.One")))))
                 (("List.list.Cons"
                    (("Product_Type.Pair"
                       ("Num.numeral_class.numeral" ("Num.num.Bit0" "Num.num.One")))
                      "Groups.one_class.one"))
                   "List.list.Nil")))))
          (("List.list.Cons"
             (("Product_Type.Pair"
                ("Num.numeral_class.numeral" ("Num.num.Bit0" ("Num.num.Bit1" "Num.num.One"))))
               ("Towards_Pretty_Printing.Monom"
                 (("List.list.Cons"
                    (("Product_Type.Pair"
                       ("Num.numeral_class.numeral" ("Num.num.Bit1" "Num.num.One")))
                      "Groups.one_class.one"))
                   "List.list.Nil"))))
            "List.list.Nil")))
       "sort_poly")):
   Ast.astt*)
\<close> ML \<open>
(*isar-ref.pdf p.192 Printing of terms: *)

(*8.5.3 Syntax translation functions: ML (instead rewriting)
  print_ast_translation*)
\<close> text \<open>
"((*[*) [(4::int,[(1::int, 5::int), (2, 1)]), (6,[(3, 1)]) (*]*)], sort_poly)";
4: ("Num.numeral_class.numeral" "$" ("Num.num.Bit0" "$" ("Num.num.Bit0" "$" "Num.num.One")))
1: "Groups.one_class.one"
5: ("Num.numeral_class.numeral" "$" ("Num.num.Bit1" "$" ("Num.num.Bit0" "$" "Num.num.One")))
2: ("Num.numeral_class.numeral" "$" ("Num.num.Bit0" "$" "Num.num.One"))
6: ("Num.numeral_class.numeral" "$" ("Num.num.Bit0" "$" ("Num.num.Bit1" "$" "Num.num.One")))
3: ("Num.numeral_class.numeral" "$" ("Num.num.Bit1" "$" "Num.num.One"))
\<close> ML \<open>
val astint_1 = Ast.Constant "Groups.one_class.one"
val astint_2 = 
  Ast.Appl [Ast.Constant "Num.numeral_class.numeral",
    Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$", Ast.Constant "Num.num.One"]]
val astint_3 = 
  Ast.Appl [Ast.Constant "Num.numeral_class.numeral", Ast.Constant "$",
    Ast.Appl [Ast.Constant "Num.num.Bit1", Ast.Constant "$", Ast.Constant "Num.num.One"]]
val astint_4 = 
  Ast.Appl [Ast.Constant "Num.numeral_class.numeral", Ast.Constant "$",
    Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$",
      Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$", Ast.Constant "Num.num.One"]]]
val astint_5 = 
  Ast.Appl [Ast.Constant "Num.numeral_class.numeral", Ast.Constant "$",
    Ast.Appl [Ast.Constant "Num.num.Bit1", Ast.Constant "$",
      Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$", Ast.Constant "Num.num.One"]]]
val astint_6 = 
  Ast.Appl [Ast.Constant "Num.numeral_class.numeral", Ast.Constant "$",
    Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$",
      Ast.Appl [Ast.Constant "Num.num.Bit1", Ast.Constant "$", Ast.Constant "Num.num.One"]]]
\<close> ML \<open>
fun ast_flat (Ast.Constant a) = [a]
  | ast_flat (Ast.Appl [Ast.Constant a1, Ast.Constant a2]) = [a1, a2]
  | ast_flat (Ast.Appl [Ast.Constant a1, Ast.Constant "$", aa]) = a1 :: ast_flat aa
  | ast_flat ast = raise ERROR ("ERROR int_of_ast " ^ PolyML.makestring ast)
\<close> ML \<open>
fun flat_int ["Num.num.One"] = 1
  | flat_int ("Num.num.Bit0" :: bs) = 2 * flat_int bs
  | flat_int ("Num.num.Bit1" :: bs) = 2 * flat_int bs + 1
  | flat_int bs = raise ERROR ("ERROR flat_int " ^ PolyML.makestring bs)
\<close> ML \<open>
val bs_6 = Ast.Appl [Ast.Constant "Num.num.Bit0", Ast.Constant "$",
      Ast.Appl [Ast.Constant "Num.num.Bit1", Ast.Constant "$", Ast.Constant "Num.num.One"]]
\<close> ML \<open>
(bs_6 |> ast_flat |> flat_int) = 6;
\<close> ML \<open>
fun ast_num (Ast.Constant "Groups.one_class.one") = SOME (Ast.Constant "1")
  | ast_num (Ast.Appl [Ast.Constant "Num.numeral_class.numeral", Ast.Constant "$", bs]) = 
    SOME (bs |> ast_flat |> flat_int |> string_of_int |> Ast.Constant)
  | ast_num _ = NONE
(*val int_of_ast = fn: Ast.ast -> string option*)
\<close> ML \<open>
SOME (* = fn: 'a -> 'a option *);
SOME 123;
NONE: int option
\<close> ML \<open>
ast_num astint_1 = SOME (Ast.Constant "1");
ast_num astint_6 = SOME (Ast.Constant "6");
\<close> ML \<open>
(* rewriting top down only *)
fun ast_to_ast rew (ast as Ast.Constant _) =
    (case rew ast of SOME r => r | NONE => ast)
  | ast_to_ast rew (ast as Ast.Appl (s :: ss)) =
    (case rew ast of 
      SOME r => r
    | NONE => Ast.Appl (ast_to_ast rew s :: (asts_to_asts rew ss)))
  | ast_to_ast _ a = raise ERROR ("ERROR ast_to_ast " ^ PolyML.makestring a)
and asts_to_asts _ [] = []
  | asts_to_asts rew (s :: ss) =
    ast_to_ast rew s :: asts_to_asts rew ss
\<close> ML \<open>
val myast = ast_to_ast ast_num astpoly (* Bit0/1 replaced by int TODO !not all! *)
\<close> 
subsubsection \<open>Conversion term -> AST by Isabelle\<close>
text \<open>from code reading:
# ~~/src/Pure/Isar/isar_cmd.ML
fun string_of_term ctxt s =
  let
    val t = Syntax.read_term ctxt s;
    val T = Term.type_of t;
    val ctxt' = Variable.auto_fixes t ctxt;
  in
    Pretty.string_of
      (Pretty.block [Pretty.quote (Syntax.pretty_term ctxt' t), Pretty.fbrk,
        Pretty.str "::", Pretty.brk 1, Pretty.quote (Syntax.pretty_typ ctxt' T)])
  end;
# ~~/src/Pure/Syntax/syntax.ML
val string_of_term = Pretty.string_of oo pretty_term;
fun pretty_term ctxt = singleton (uncheck_terms ctxt) #> unparse_term ctxt;
# ~~/src/Pure/Syntax/syntax_phases.ML
fun unparse_term ctxt =
  let
    val thy = Proof_Context.theory_of ctxt;
    val syn = Proof_Context.syn_of ctxt;
  in
    unparse_t (term_to_ast (is_some o Syntax.lookup_const syn))
      (Printer.pretty_term_ast (not (Pure_Thy.old_appl_syntax thy)))
      (Markup.language_term false) ctxt
  end;
\<close> ML \<open>
(* ^^^^^^^^^^^ term_to_ast: is local, so we copy the minium from ...
   ~~/src/Pure/Syntax/syntax_phases.ML: ...*)
fun simple_ast_of ctxt =
  let
    val tune_var = if Config.get ctxt show_question_marks then I else unprefix "?";
    fun ast_of (Const (c, _)) = Ast.Constant c
      | ast_of (Free (x, _)) = Ast.Variable x
      | ast_of (Var (xi, _)) = Ast.Variable (tune_var (Term.string_of_vname xi))
      | ast_of (t as _ $ _) =
          let val (f, args) = strip_comb t
          in Ast.mk_appl (ast_of f) (map ast_of args) end
      | ast_of (Bound i) = Ast.Appl [Ast.Constant "_loose", Ast.Variable ("B." ^ string_of_int i)]
      | ast_of (Abs _) = raise Fail "simple_ast_of: Abs";
  in ast_of end;
\<close> ML \<open>
val isaast = simple_ast_of ctxt termpoly : Ast.ast
\<close> 

subsubsection \<open>List.thy: syntax, translations\<close>
text \<open>
syntax
  \<comment> \<open>list Enumeration\<close>
  "_list" :: "args => 'a list"    ("[(_)]")

translations
  "[x, xs]" == "x#[xs]"
  "[x]" == "x#[]"
\<close>
datatype 'a xlist =
    XNil    ("{|| ||}")
  | XCons 'a  "'a xlist"    (infixr "@#" 65)

syntax
  -- {* list Enumeration *}
  "_xlist" :: "args => 'a xlist"    ("{|| (_) ||}")

translations
  "{||x, xs||}" == "x @# {||xs||}"
  "{||x||}" == "x @# {|| ||}"

term "{|| ||}"
term "{||1,2,3||}"

ML \<open>
\<close> ML \<open>
\<close>
subsubsection \<open>HOL.thy: syntax, translations, print_translation\<close>
ML \<open>
@{term "THE xxx. xxx > 0"}(*= ("HOL.The", "('a \<Rightarrow> bool) \<Rightarrow> 'a") $ Abs ("xxx", "'a", ...*);
(*@{term "THE xxx"}          ...Inner syntax error: unexpected end of input\<here> Failed to parse term*)
\<close> text \<open>
HOL.thy:
syntax "_The" :: "[pttrn, bool] \<Rightarrow> 'a"  ("(3THE _./ _)" [0, 10] 10)
translations "THE x. P" \<rightleftharpoons> "CONST The (\<lambda>x. P)"
print_translation \<open>
  [(@{const_syntax The}, fn _ => fn [Abs abs] =>
      let val (x, t) = Syntax_Trans.atomic_abs_tr' abs
      in Syntax.const @{syntax_const "_The"} $ x $ t end)]
\<close>  \<comment> \<open>To avoid eta-contraction of body\<close>
\<close> ML \<open>
Abs
\<close> 
subsubsection \<open>Product_Type.thy: \\\NO: split...\\\\<close>
text \<open>
\<close> 
ML \<open>
\<close> ML \<open>
\<close>

subsection \<open>Milestone 5: concrete syntax in Isabelle\<close>
text \<open>
  Isabelle's code encapsulates functionality as much as possible.
  So, when inspecting Isabelle's mechanisms, which concern asts and translation to strings,
  some functions appear not accessible without breaking encapsulation.
\<close> ML \<open>
(* this is public: *)
Ast.normalize;
(* = fn: Proof.context -> (string -> (Ast.ast * Ast.ast) list) -> Ast.ast -> Ast.ast*);
Printer.pretty_term_ast (* = fn:
   bool ->
     Proof.context ->
       Printer.prtabs ->
         (string -> Proof.context -> Ast.ast list -> Ast.ast) ->
           (string -> Ast.ast list -> Pretty.T option) ->
             (string -> Markup.T list * string) -> Ast.ast -> Pretty.T list *);
Pretty.string_of (* = fn: Pretty.T -> string *);
\<close> text \<open>
  But the arguments of Printer.pretty_term_ast are comprehensive such that we look into the code, 
  where this function is called (it is called only once in all Isabelle code).
  The call is in ~~/src/Pure/Syntax/syntax_phases.ML by:

unparse_term (*  = fn: Proof.context -> term -> Pretty.T *);

  Below the respective code's encapsulation is broken up. 
\<close>

subsubsection \<open>Isabelle's code for translation from ast to strings\<close>
ML \<open>
open Syntax_Phases
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun mark_atoms is_syntax_const ctxt tm =
  let
    val {structs, fixes} = Syntax_Trans.get_idents ctxt;
    val show_structs = Config.get ctxt show_structs;

    fun mark ((t as Const (c, _)) $ u) =
          if member (op =) Pure_Thy.token_markers c
          then t $ u else mark t $ mark u
      | mark (t $ u) = mark t $ mark u
      | mark (Abs (x, T, t)) = Abs (x, T, mark t)
      | mark (t as Const (c, T)) =
          if is_syntax_const c then t
          else Const (Lexicon.mark_const c, T)
      | mark (t as Free (x, T)) =
          let val i = find_index (fn s => s = x) structs + 1 in
            if i = 0 andalso member (op =) fixes x then
              Const (Lexicon.mark_fixed x, T)
            else if i = 1 andalso not show_structs then
              Syntax.const "_struct" $ Syntax.const "_indexdefault"
            else Syntax.const "_free" $ t
          end
      | mark (t as Var (xi, T)) =
          if xi = Syntax_Ext.dddot_indexname then Const ("_DDDOT", T)
          else Syntax.const "_var" $ t
      | mark a = a;
  in mark tm end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun prune_types tm =
  let
    fun regard t t' seen =
      if Type_Annotation.is_omitted (Type_Annotation.fastype_of [] t) then (t, seen)
      else if member (op aconv) seen t then (t', seen)
      else (t, t :: seen);

    fun prune (t as Const _, seen) = (t, seen)
      | prune (t as Free (x, T), seen) = regard t (Free (x, Type_Annotation.ignore_type T)) seen
      | prune (t as Var (xi, T), seen) = regard t (Var (xi, Type_Annotation.ignore_type T)) seen
      | prune (t as Bound _, seen) = (t, seen)
      | prune (Abs (x, T, t), seen) =
          let val (t', seen') = prune (t, seen);
          in (Abs (x, T, t'), seen') end
      | prune (t1 $ t2, seen) =
          let
            val (t1', seen') = prune (t1, seen);
            val (t2', seen'') = prune (t2, seen');
          in (t1' $ t2', seen'') end;
  in #1 (prune (tm, [])) end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun mark_aprop tm =
  let
    fun aprop t = Syntax.const "_aprop" $ t;

    fun is_prop Ts t =
      Type_Annotation.clean (Type_Annotation.fastype_of Ts t) = propT
        handle TERM _ => false;

    fun is_term (Const ("Pure.term", _) $ _) = true
      | is_term _ = false;

    fun mark _ (t as Const _) = t
      | mark Ts (t as Const ("_bound", _) $ u) = if is_prop Ts u then aprop t else t
      | mark Ts (t as Free _) = if is_prop Ts t then aprop t else t
      | mark Ts (t as Var _) = if is_prop Ts t then aprop t else t
      | mark Ts (t as Bound _) = if is_prop Ts t then aprop t else t
      | mark Ts (Abs (x, T, t)) = Abs (x, T, mark (T :: Ts) t)
      | mark Ts (t as t1 $ (t2 as Const ("Pure.type", Type ("itself", [T])))) =
          if is_prop Ts t andalso not (is_term t) then Const ("_type_prop", T) $ mark Ts t1
          else mark Ts t1 $ mark Ts t2
      | mark Ts (t as t1 $ t2) =
          (if is_Const (Term.head_of t) orelse not (is_prop Ts t) then I else aprop)
            (mark Ts t1 $ mark Ts t2);
  in mark [] tm end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun ast_of_termT ctxt trf tm =
  let
    val ctxt' = Config.put show_sorts false ctxt;
    fun ast_of (t as Const ("_tfree", _) $ Free _) = simple_ast_of ctxt t
      | ast_of (t as Const ("_tvar", _) $ Var _) = simple_ast_of ctxt t
      | ast_of (Const (a, _)) = trans a []
      | ast_of (t as _ $ _) =
          (case strip_comb t of
            (Const (a, _), args) => trans a args
          | (f, args) => Ast.Appl (map ast_of (f :: args)))
      | ast_of t = simple_ast_of ctxt t
    and trans a args = ast_of (trf a ctxt' dummyT args)
      handle Match => Ast.mk_appl (Ast.Constant a) (map ast_of args);
  in ast_of tm end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun term_to_ast is_syntax_const ctxt trf tm =
  let
    val show_types = Config.get ctxt show_types orelse Config.get ctxt show_sorts;
    val show_markup = Config.get ctxt show_markup;

    fun ast_of tm =
      (case strip_comb tm of
        (t as Abs _, ts) => Ast.mk_appl (ast_of (Syntax_Trans.abs_tr' ctxt t)) (map ast_of ts)
      | ((c as Const ("_free", _)), Free (x, T) :: ts) =>
          Ast.mk_appl (constrain (c $ Syntax.free x) T) (map ast_of ts)
      | ((c as Const ("_var", _)), Var (xi, T) :: ts) =>
          Ast.mk_appl (constrain (c $ Syntax.var xi) T) (map ast_of ts)
      | ((c as Const ("_bound", B)), Free (x, T) :: ts) =>
          let
            val X =
              if show_markup andalso not show_types orelse B <> dummyT then T
              else dummyT;
          in Ast.mk_appl (constrain (c $ Syntax.free x) X) (map ast_of ts) end
      | (Const ("_idtdummy", T), ts) =>
          Ast.mk_appl (constrain (Syntax.const "_idtdummy") T) (map ast_of ts)
      | (Const (c, T), ts) => trans c (Type_Annotation.smash T) ts
      | (t, ts) => Ast.mk_appl (simple_ast_of ctxt t) (map ast_of ts))

    and trans a T args = ast_of (trf a ctxt T args)
      handle Match => Ast.mk_appl (Ast.Constant a) (map ast_of args)

    and constrain t T0 =
      let
        val T =
          if show_markup andalso not show_types
          then Type_Annotation.clean T0
          else Type_Annotation.smash T0;
      in
        if (show_types orelse show_markup) andalso T <> dummyT then
          Ast.Appl [Ast.Constant "_constrain", simple_ast_of ctxt t,
            ast_of_termT ctxt trf (term_of_typ ctxt T)]
        else simple_ast_of ctxt t
      end;
  in
    tm
    |> mark_aprop
    |> show_types ? prune_types
    |> Variable.revert_bounds ctxt
    |> mark_atoms is_syntax_const ctxt
    |> ast_of
  end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
val typing_elem = YXML.output_markup_elem Markup.typing;
val sorting_elem = YXML.output_markup_elem Markup.sorting;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun var_or_skolem s =
  (case Lexicon.read_variable s of
    SOME (x, i) =>
      (case try Name.dest_skolem x of
        NONE => (Markup.var, s)
      | SOME x' => (Markup.skolem, Term.string_of_vname (x', i)))
  | NONE => (Markup.var, s));
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun free_or_skolem ctxt x =
  let
    val m =
      if Variable.is_fixed ctxt x orelse Syntax.is_pretty_global ctxt
      then Markup.fixed x else Markup.intensify;
  in
    if Name.is_skolem x
    then ([m, Markup.skolem], Variable.revert_fixed ctxt x)
    else ([m, Markup.free], x)
  end;
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun markup_class ctxt c =
  [Name_Space.markup (Type.class_space (Proof_Context.tsig_of ctxt)) c];

fun markup_type ctxt c =
  [Name_Space.markup (Type.type_space (Proof_Context.tsig_of ctxt)) c];

fun markup_const ctxt c =
  [Name_Space.markup (Consts.space_of (Proof_Context.consts_of ctxt)) c];
\<close> ML \<open>(* copy from ~~/src/Pure/Syntax/syntax_phases.ML *)
fun unparse_t t_to_ast prt_t markup ctxt t =
  let
    val show_markup = Config.get ctxt show_markup;
    val show_sorts = Config.get ctxt show_sorts;
    val show_types = Config.get ctxt show_types orelse show_sorts;

    val syn = Proof_Context.syn_of ctxt;
    val prtabs = Syntax.prtabs syn;
    val trf = Syntax.print_ast_translation syn;

    fun markup_extern c =
      (case Syntax.lookup_const syn c of
        SOME "" => ([], c)
      | SOME b => markup_extern b
      | NONE => c |> Lexicon.unmark
         {case_class = fn x => (markup_class ctxt x, Proof_Context.extern_class ctxt x),
          case_type = fn x => (markup_type ctxt x, Proof_Context.extern_type ctxt x),
          case_const = fn x => (markup_const ctxt x, Proof_Context.extern_const ctxt x),
          case_fixed = fn x => free_or_skolem ctxt x,
          case_default = fn x => ([], x)});

    fun token_trans "_tfree" x = SOME (Pretty.mark_str (Markup.tfree, x))
      | token_trans "_tvar" x = SOME (Pretty.mark_str (Markup.tvar, x))
      | token_trans "_free" x = SOME (Pretty.marks_str (free_or_skolem ctxt x))
      | token_trans "_bound" x = SOME (Pretty.mark_str (Markup.bound, x))
      | token_trans "_loose" x = SOME (Pretty.mark_str (Markup.bad, x))
      | token_trans "_var" x = SOME (Pretty.mark_str (var_or_skolem x))
      | token_trans "_numeral" x = SOME (Pretty.mark_str (Markup.numeral, x))
      | token_trans "_inner_string" x = SOME (Pretty.mark_str (Markup.inner_string, x))
      | token_trans _ _ = NONE;

    fun markup_trans a [Ast.Variable x] = token_trans a x
      | markup_trans "_constrain" [t, ty] = constrain_trans t ty
      | markup_trans "_idtyp" [t, ty] = constrain_trans t ty
      | markup_trans "_ofsort" [ty, s] = ofsort_trans ty s
      | markup_trans _ _ = NONE

    and constrain_trans t ty =
      if show_markup andalso not show_types then
        let
          val ((bg1, bg2), en) = typing_elem;
          val bg = bg1 ^ Pretty.symbolic_output (pretty_typ_ast Markup.empty ty) ^ bg2;
        in SOME (Pretty.make_block (bg, en) false 0 [pretty_ast Markup.empty t]) end
      else NONE

    and ofsort_trans ty s =
      if show_markup andalso not show_sorts then
        let
          val ((bg1, bg2), en) = sorting_elem;
          val bg = bg1 ^ Pretty.symbolic_output (pretty_typ_ast Markup.empty s) ^ bg2;
        in SOME (Pretty.make_block (bg, en) false 0 [pretty_typ_ast Markup.empty ty]) end
      else NONE

    and pretty_typ_ast m ast = ast
      |> Printer.pretty_typ_ast ctxt prtabs trf markup_trans markup_extern
      |> Pretty.markup m

    and pretty_ast m ast = ast
      |> prt_t ctxt prtabs trf markup_trans markup_extern
      |> Pretty.markup m;
  in
    t_to_ast ctxt (Syntax.print_translation syn) t
    |> Ast.normalize ctxt (Syntax.print_rules syn)
    |> pretty_ast markup
  end;
\<close> ML \<open>
fun unparse_term ctxt =
  let
    val thy = Proof_Context.theory_of ctxt;
    val syn = Proof_Context.syn_of ctxt;
  in
    unparse_t (term_to_ast (is_some o Syntax.lookup_const syn))
      (Printer.pretty_term_ast (not (Pure_Thy.old_appl_syntax thy)))
      (Markup.language_term false) ctxt
  end;
\<close> ML \<open>
unparse_t (* = fn:
   (Proof.context -> (string -> Proof.context -> typ -> term list -> term) -> 'a -> Ast.ast) ->
     (Proof.context ->
        Printer.prtabs ->
          (string -> Proof.context -> Ast.ast list -> Ast.ast) ->
            (string -> Ast.ast list -> Pretty.T option) ->
              (string -> Markup.T list * string) -> Ast.ast -> Pretty.T list)
       -> Markup.T -> Proof.context -> 'a -> Pretty.T*);
term_to_ast (* = fn:
   (string -> bool) ->
     Proof.context -> (string -> Proof.context -> typ -> term list -> term) -> term -> Ast.ast *);
Syntax.lookup_const (* = fn: Syntax.syntax -> string -> string option *);
Printer.pretty_term_ast (* = fn:
   bool ->
     Proof.context ->
       Printer.prtabs ->
         (string -> Proof.context -> Ast.ast list -> Ast.ast) ->
           (string -> Ast.ast list -> Pretty.T option) ->
             (string -> Markup.T list * string) -> Ast.ast -> Pretty.T list *);
Markup.language_term (* = fn: bool -> Markup.T *);
\<close> ML \<open>
 (**);
unparse_term (* = fn: Proof.context -> term -> Pretty.T *);
\<close> ML \<open>
 (**);
\<close>

subsubsection \<open>Trials with transformations on asts\<close>
ML \<open>
unparse_t (* is the *only* caller of Ast.normalize; it is decomposed as follows *);
fun my_ast_to_ast ctxt syn ast = Ast.normalize ctxt (Syntax.print_rules syn) ast
\<close> ML \<open>
val ctxt = Proof_Context.init_global @{theory} (* = <context>: Proof.context *);
val syn = Proof_Context.syn_of ctxt (* = Syntax.syntax*);
\<close> ML \<open>
astpoly (*= Ast.ast*);
\<close> ML \<open>
(* and now we want to see the effect of transformations (triggered by ctxt | syn ?) *)
   my_ast_to_ast ctxt syn astpoly;
(*^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
  WE DON'T SEE ANY CHANGE !?!*)
\<close> ML \<open>
(*vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
  SO WHERE DOES THE CONVERSION FROM Bit0/1 TO numbers COME FROM ? *)
termpoly |> unparse_term ctxt |> Pretty.string_of |> writeln;
(*Poly ([{[(4, Monom [{[(1, 5), (2, 1)]}]), (6, Monom [{[(3, 1)]}])]}], sort_poly) 
val it = (): unit*)
\<close> text \<open>
~~/src/HOL/Num.thy: typed_print_translation [(@ {const_syntax numeral}, num_tr')]
                    i.e. triggered by       "\ <^const>Num.numeral_class.numeral"
  SEEMS TO PRODUCE THE CONVERSION:
  the translation, however, is applied to terms, i.e. apparently *after* ast rewriting...
\<close> ML \<open>
(* ~~/src/HOL/Num.thy
   typed_print_translation... *)
(* ERROR Undefined constant: "Bit0"\<here>    vvvvvvvvv    vv  avoided herewith *)
  let 
    fun dest_num (Const (@{const_syntax numeral(*Bit0*)}, _) $ n) = 2 * dest_num n
      | dest_num (Const (@{const_syntax numeral(*Bit1*)}, _) $ n) = 2 * dest_num n + 1
      | dest_num (Const (@{const_syntax numeral(*One*)}, _)) = 1;
    fun num_tr' ctxt T [n] =
      let
        val k = dest_num n;
        val t' =
          Syntax.const @{syntax_const "_Numeral"} $
            Syntax.free (string_of_int k);
      in
        (case T of
          Type (@{type_name fun}, [_, T']) =>
            if Printer.type_emphasis ctxt T' then
              Syntax.const @{syntax_const "_constrain"} $ t' $
                Syntax_Phases.term_of_typ ctxt T'
            else t'
        | _ => if T = dummyT then t' else raise Match)
      end;
  in
   [(@{const_syntax numeral}, num_tr')]
  end
\<close> ML \<open>
@{const_syntax numeral} (* = "\<^const>Num.numeral_class.numeral"*);
\<close> ML \<open>
(* ...WE SEE: THIS RULE IS NOT IN THE SYNTAX USED BY Ast.normalize *)
(Syntax.print_rules syn) "\<^const>Num.numeral_class.numeral" = [];
\<close> ML \<open>
(*GOON: conversion from Bit0/1 to numbers as ast rewriting ?!?*)
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>
                                              
subsubsection \<open>xxx\<close>
text \<open>xxx
\<close> ML \<open>
\<close> ML \<open>
\<close>

subsection \<open>Milestone 6: integrate pretty printing into polynomial package\<close>
ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>
subsection \<open>Milestone 7: cleanup and optimise\<close>
ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>
section \<open>xxx\<close>
ML \<open>
\<close> ML \<open>
\<close> ML \<open>
\<close>

end