theory Try_Poly_Demo
imports "../poly-WORK/Poly_Demo"
begin

section {* Joint work Computer Algebra --- Isabelle *}
section {* Towards a Polynomial Package in Isabelle *}
section {* Demo: the proof-of-concept prototype *}
subsection {* Isabelle/jEdit: towards a prover IDE *}
subsection {* An example algorithm and a proof *}
subsection {* Automated code generation *}

subsubsection {* Recursive addition: export_code *}

definition p013_negative :: "int mpoly" (* -2 * x\<^sup>1  +  -3 * z\<^sup>4*)
where "p013_negative = 
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 1) -2 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 2 4) -3"
value p013_negative --{* 2 * x\<^sup>1  +  3 * z\<^sup>4  = "Rec 
 (Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 0, Coeff 0, Coeff 0, Coeff - 3]))])), 
    Coeff - 2])))"*}

definition try_plus :: "int mpoly \<Rightarrow> int mpoly \<Rightarrow> int mpoly"
where "try_plus p q = p + q"

value "try_plus p013 p013_negative" (* is this NOT normalised ? *)
--{*= "Rec (Powers (Dense (Abs_nat_mapping_dense [
            Powers (Dense (Abs_nat_mapping_dense [
            Powers (Dense (Abs_nat_mapping_dense []))]))])))"*}

(* =========================== univariate   examples =========================== *)
definition up_dense :: "int mpoly"
where "up_dense =  (* 1 + 2.x + 3.x^2 *)
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 0) 1 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 1) 2 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 2) 3"
value up_dense --{*= "Rec (Powers (Dense (Abs_nat_mapping_dense [Coeff 1, Coeff 2, Coeff 3])))"*}

definition uq_dense :: "int mpoly"
where "uq_dense =  (* 4 + 5.x *)
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 0) 4 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 1) 5"
value uq_dense --{*= "Rec (Powers (Dense (Abs_nat_mapping_dense [Coeff 4, Coeff 5])))"*}

definition up_sparse :: "int mpoly"
where "up_sparse =  (* 1 + 2.x + 3.x^2 *)
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 0) 1 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 1) 2 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 2) 3"
value up_sparse --{*= "Rec (Powers (Sparse (Abs_nat_mapping_sparse 
  [(0, Coeff 1), (Suc 0, Coeff 2), (Suc (Suc 0), Coeff 3)])))"*}

definition uq_sparse :: "int mpoly"
where "uq_sparse =  (* 4 + 5.x *)
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 0) 4 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 1) 5"
value uq_sparse --{*= "Rec (Powers (Sparse (Abs_nat_mapping_sparse 
  [(0, Coeff 4), (Suc 0, Coeff 5)])))"*}

(* =========================== multivariate examples =========================== *)
--{* with monom' mixed-variate polynomials, x.y.z, are not possible*}
definition pdense :: "int mpoly"
where "pdense =  (* x^3 + 2.y + 3.x^2 *)
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 3) 1 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 1 1) 2 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 2 2) 3"
value pdense --{*= "Rec (
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 0, Coeff 3])), 
    Coeff 2])), 
  Coeff 0, Coeff 0, Coeff 1])))"*}

definition psparse :: "int mpoly"
where "psparse = (* x^3 + 2.y + 3.x^2 *)
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 3) 1 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 1 1) 2 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 2 2) 3"
value psparse --{*= "Rec (
  Powers (Sparse (Abs_nat_mapping_sparse [
    (0, Powers (Sparse (Abs_nat_mapping_sparse [
      (0, Powers (Sparse (Abs_nat_mapping_sparse [(Suc (Suc (Suc (Suc 0))), Coeff 3)]))), 
    (Suc 0, Coeff 2)]))),
  (Suc (Suc (Suc 0)), Coeff 1)])))"*}

definition qdense :: "int mpoly"
where "qdense = (* 5.x + 6.z*)
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 1) 5 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 2 1) 6"
value qdense --{*= "Rec (
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Powers (Dense (Abs_nat_mapping_dense [Coeff 0, Coeff 6]))])), 
  Coeff 5])))"*}

definition qsparse :: "int mpoly"
where "qsparse = (* 5.x + 6.z*)
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 0 1) 5 +
  monom' (REC (\<lambda>_. SPARSE)) (Nat_Mapping.single' DENSE 2 1) 6"
value qsparse --{*= "Rec (
  Powers (Sparse (Abs_nat_mapping_sparse [
    (0, Powers (Sparse (Abs_nat_mapping_sparse [
      (0, Powers (Sparse (Abs_nat_mapping_sparse [(Suc 0, Coeff 6)])))]))), 
  (Suc 0, Coeff 5)])))"*}

(*
export_code try_plus up_dense uq_dense up_sparse uq_sparse in SML module_name Try_Plus
*)
export_code try_plus up_dense uq_dense up_sparse uq_sparse in SML module_name Try_Plus
ML {*
val cnt = Unsynchronized.ref 0;
(*************************************************************************
structure Try_Plus : sig
  type num
  type int
  type nat
  type 'a nat_mapping
  type implb
  type 'a mpoly
  type impl
  val try_plus : int mpoly -> int mpoly -> int mpoly
  val up_dense : int mpoly
  val uq_dense : int mpoly
  val up_sparse : int mpoly
  val uq_sparse : int mpoly
end = struct
*************************************************************************)

type 'a uminus = {uminus : 'a -> 'a};
val uminus = #uminus : 'a uminus -> 'a -> 'a;

fun uminus_funa B_ a x = uminus B_ (a x);

fun uminus_fun B_ = {uminus = uminus_funa B_} : ('a -> 'b) uminus;

datatype num = One | Bit0 of num | Bit1 of num;

fun equal_num (Bit0 numa) (Bit1 num) = false
  | equal_num (Bit1 numa) (Bit0 num) = false
  | equal_num One (Bit1 num) = false
  | equal_num (Bit1 num) One = false
  | equal_num One (Bit0 num) = false
  | equal_num (Bit0 num) One = false
  | equal_num (Bit1 numa) (Bit1 num) = equal_num numa num
  | equal_num (Bit0 numa) (Bit0 num) = equal_num numa num
  | equal_num One One = true;

datatype int = Zero_int | Pos of num | Neg of num;

fun equal_inta (Neg k) (Neg l) = equal_num k l
  | equal_inta (Neg k) (Pos l) = false
  | equal_inta (Neg k) Zero_int = false
  | equal_inta (Pos k) (Neg l) = false
  | equal_inta (Pos k) (Pos l) = equal_num k l
  | equal_inta (Pos k) Zero_int = false
  | equal_inta Zero_int (Neg l) = false
  | equal_inta Zero_int (Pos l) = false
  | equal_inta Zero_int Zero_int = true;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

val equal_int = {equal = equal_inta} : int equal;

val one_inta : int = Pos One;

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_int = {one = one_inta} : int one;

fun uminus_inta (Neg m) = Pos m
  | uminus_inta (Pos m) = Neg m
  | uminus_inta Zero_int = Zero_int;

fun plus_num (Bit1 m) (Bit1 n) = Bit0 (plus_num (plus_num m n) One)
  | plus_num (Bit1 m) (Bit0 n) = Bit1 (plus_num m n)
  | plus_num (Bit1 m) One = Bit0 (plus_num m One)
  | plus_num (Bit0 m) (Bit1 n) = Bit1 (plus_num m n)
  | plus_num (Bit0 m) (Bit0 n) = Bit0 (plus_num m n)
  | plus_num (Bit0 m) One = Bit1 m
  | plus_num One (Bit1 n) = Bit0 (plus_num n One)
  | plus_num One (Bit0 n) = Bit1 n
  | plus_num One One = Bit0 One;

fun times_num (Bit1 m) (Bit1 n) =
  Bit1 (plus_num (plus_num m n) (Bit0 (times_num m n)))
  | times_num (Bit1 m) (Bit0 n) = Bit0 (times_num (Bit1 m) n)
  | times_num (Bit0 m) (Bit1 n) = Bit0 (times_num m (Bit1 n))
  | times_num (Bit0 m) (Bit0 n) = Bit0 (Bit0 (times_num m n))
  | times_num One n = n
  | times_num m One = m;

fun times_inta (Neg m) (Neg n) = Pos (times_num m n)
  | times_inta (Neg m) (Pos n) = Neg (times_num m n)
  | times_inta (Pos m) (Neg n) = Neg (times_num m n)
  | times_inta (Pos m) (Pos n) = Pos (times_num m n)
  | times_inta Zero_int l = Zero_int
  | times_inta k Zero_int = Zero_int;

fun bitM One = One
  | bitM (Bit0 n) = Bit1 (bitM n)
  | bitM (Bit1 n) = Bit1 (Bit0 n);

fun dup (Neg n) = Neg (Bit0 n)
  | dup (Pos n) = Pos (Bit0 n)
  | dup Zero_int = Zero_int;

fun minus_inta (Neg m) (Neg n) = sub n m
  | minus_inta (Neg m) (Pos n) = Neg (plus_num m n)
  | minus_inta (Pos m) (Neg n) = Pos (plus_num m n)
  | minus_inta (Pos m) (Pos n) = sub m n
  | minus_inta Zero_int l = uminus_inta l
  | minus_inta k Zero_int = k
and sub (Bit0 m) (Bit1 n) = minus_inta (dup (sub m n)) (Pos One)
  | sub (Bit1 m) (Bit0 n) = plus_inta (dup (sub m n)) (Pos One)
  | sub (Bit1 m) (Bit1 n) = dup (sub m n)
  | sub (Bit0 m) (Bit0 n) = dup (sub m n)
  | sub One (Bit1 n) = Neg (Bit0 n)
  | sub One (Bit0 n) = Neg (bitM n)
  | sub (Bit1 m) One = Pos (Bit0 m)
  | sub (Bit0 m) One = Pos (bitM m)
  | sub One One = Zero_int
and plus_inta (Neg m) (Neg n) = Neg (plus_num m n)
  | plus_inta (Neg m) (Pos n) = sub n m
  | plus_inta (Pos m) (Neg n) = sub m n
  | plus_inta (Pos m) (Pos n) = Pos (plus_num m n)
  | plus_inta Zero_int l = l
  | plus_inta k Zero_int = k;

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a minus = {minus : 'a -> 'a -> 'a};
val minus = #minus : 'a minus -> 'a -> 'a -> 'a;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

type 'a plus = {plus : 'a -> 'a -> 'a};
val plus = #plus : 'a plus -> 'a -> 'a -> 'a;

type 'a semigroup_add = {plus_semigroup_add : 'a plus};
val plus_semigroup_add = #plus_semigroup_add : 'a semigroup_add -> 'a plus;

type 'a cancel_semigroup_add =
  {semigroup_add_cancel_semigroup_add : 'a semigroup_add};
val semigroup_add_cancel_semigroup_add = #semigroup_add_cancel_semigroup_add :
  'a cancel_semigroup_add -> 'a semigroup_add;

type 'a ab_semigroup_add = {semigroup_add_ab_semigroup_add : 'a semigroup_add};
val semigroup_add_ab_semigroup_add = #semigroup_add_ab_semigroup_add :
  'a ab_semigroup_add -> 'a semigroup_add;

type 'a cancel_ab_semigroup_add =
  {ab_semigroup_add_cancel_ab_semigroup_add : 'a ab_semigroup_add,
    cancel_semigroup_add_cancel_ab_semigroup_add : 'a cancel_semigroup_add};
val ab_semigroup_add_cancel_ab_semigroup_add =
  #ab_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a ab_semigroup_add;
val cancel_semigroup_add_cancel_ab_semigroup_add =
  #cancel_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a cancel_semigroup_add;

type 'a monoid_add =
  {semigroup_add_monoid_add : 'a semigroup_add, zero_monoid_add : 'a zero};
val semigroup_add_monoid_add = #semigroup_add_monoid_add :
  'a monoid_add -> 'a semigroup_add;
val zero_monoid_add = #zero_monoid_add : 'a monoid_add -> 'a zero;

type 'a comm_monoid_add =
  {ab_semigroup_add_comm_monoid_add : 'a ab_semigroup_add,
    monoid_add_comm_monoid_add : 'a monoid_add};
val ab_semigroup_add_comm_monoid_add = #ab_semigroup_add_comm_monoid_add :
  'a comm_monoid_add -> 'a ab_semigroup_add;
val monoid_add_comm_monoid_add = #monoid_add_comm_monoid_add :
  'a comm_monoid_add -> 'a monoid_add;

type 'a cancel_comm_monoid_add =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add : 'a cancel_ab_semigroup_add,
    comm_monoid_add_cancel_comm_monoid_add : 'a comm_monoid_add};
val cancel_ab_semigroup_add_cancel_comm_monoid_add =
  #cancel_ab_semigroup_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a cancel_ab_semigroup_add;
val comm_monoid_add_cancel_comm_monoid_add =
  #comm_monoid_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a comm_monoid_add;

type 'a mult_zero = {times_mult_zero : 'a times, zero_mult_zero : 'a zero};
val times_mult_zero = #times_mult_zero : 'a mult_zero -> 'a times;
val zero_mult_zero = #zero_mult_zero : 'a mult_zero -> 'a zero;

type 'a semigroup_mult = {times_semigroup_mult : 'a times};
val times_semigroup_mult = #times_semigroup_mult :
  'a semigroup_mult -> 'a times;

type 'a semiring =
  {ab_semigroup_add_semiring : 'a ab_semigroup_add,
    semigroup_mult_semiring : 'a semigroup_mult};
val ab_semigroup_add_semiring = #ab_semigroup_add_semiring :
  'a semiring -> 'a ab_semigroup_add;
val semigroup_mult_semiring = #semigroup_mult_semiring :
  'a semiring -> 'a semigroup_mult;

type 'a semiring_0 =
  {comm_monoid_add_semiring_0 : 'a comm_monoid_add,
    mult_zero_semiring_0 : 'a mult_zero, semiring_semiring_0 : 'a semiring};
val comm_monoid_add_semiring_0 = #comm_monoid_add_semiring_0 :
  'a semiring_0 -> 'a comm_monoid_add;
val mult_zero_semiring_0 = #mult_zero_semiring_0 :
  'a semiring_0 -> 'a mult_zero;
val semiring_semiring_0 = #semiring_semiring_0 : 'a semiring_0 -> 'a semiring;

type 'a semiring_0_cancel =
  {cancel_comm_monoid_add_semiring_0_cancel : 'a cancel_comm_monoid_add,
    semiring_0_semiring_0_cancel : 'a semiring_0};
val cancel_comm_monoid_add_semiring_0_cancel =
  #cancel_comm_monoid_add_semiring_0_cancel :
  'a semiring_0_cancel -> 'a cancel_comm_monoid_add;
val semiring_0_semiring_0_cancel = #semiring_0_semiring_0_cancel :
  'a semiring_0_cancel -> 'a semiring_0;

type 'a group_add =
  {cancel_semigroup_add_group_add : 'a cancel_semigroup_add,
    minus_group_add : 'a minus, monoid_add_group_add : 'a monoid_add,
    uminus_group_add : 'a uminus};
val cancel_semigroup_add_group_add = #cancel_semigroup_add_group_add :
  'a group_add -> 'a cancel_semigroup_add;
val minus_group_add = #minus_group_add : 'a group_add -> 'a minus;
val monoid_add_group_add = #monoid_add_group_add :
  'a group_add -> 'a monoid_add;
val uminus_group_add = #uminus_group_add : 'a group_add -> 'a uminus;

type 'a ab_group_add =
  {cancel_comm_monoid_add_ab_group_add : 'a cancel_comm_monoid_add,
    group_add_ab_group_add : 'a group_add};
val cancel_comm_monoid_add_ab_group_add = #cancel_comm_monoid_add_ab_group_add :
  'a ab_group_add -> 'a cancel_comm_monoid_add;
val group_add_ab_group_add = #group_add_ab_group_add :
  'a ab_group_add -> 'a group_add;

type 'a ring =
  {ab_group_add_ring : 'a ab_group_add,
    semiring_0_cancel_ring : 'a semiring_0_cancel};
val ab_group_add_ring = #ab_group_add_ring : 'a ring -> 'a ab_group_add;
val semiring_0_cancel_ring = #semiring_0_cancel_ring :
  'a ring -> 'a semiring_0_cancel;

val plus_int = {plus = plus_inta} : int plus;

val semigroup_add_int = {plus_semigroup_add = plus_int} : int semigroup_add;

val cancel_semigroup_add_int =
  {semigroup_add_cancel_semigroup_add = semigroup_add_int} :
  int cancel_semigroup_add;

val ab_semigroup_add_int = {semigroup_add_ab_semigroup_add = semigroup_add_int}
  : int ab_semigroup_add;

val cancel_ab_semigroup_add_int =
  {ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_int,
    cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_int}
  : int cancel_ab_semigroup_add;

val zero_int = {zero = Zero_int} : int zero;

val monoid_add_int =
  {semigroup_add_monoid_add = semigroup_add_int, zero_monoid_add = zero_int} :
  int monoid_add;

val comm_monoid_add_int =
  {ab_semigroup_add_comm_monoid_add = ab_semigroup_add_int,
    monoid_add_comm_monoid_add = monoid_add_int}
  : int comm_monoid_add;

val cancel_comm_monoid_add_int =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add = cancel_ab_semigroup_add_int,
    comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_int}
  : int cancel_comm_monoid_add;

val times_int = {times = times_inta} : int times;

val mult_zero_int = {times_mult_zero = times_int, zero_mult_zero = zero_int} :
  int mult_zero;

val semigroup_mult_int = {times_semigroup_mult = times_int} :
  int semigroup_mult;

val semiring_int =
  {ab_semigroup_add_semiring = ab_semigroup_add_int,
    semigroup_mult_semiring = semigroup_mult_int}
  : int semiring;

val semiring_0_int =
  {comm_monoid_add_semiring_0 = comm_monoid_add_int,
    mult_zero_semiring_0 = mult_zero_int, semiring_semiring_0 = semiring_int}
  : int semiring_0;

val semiring_0_cancel_int =
  {cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_int,
    semiring_0_semiring_0_cancel = semiring_0_int}
  : int semiring_0_cancel;

val uminus_int = {uminus = uminus_inta} : int uminus;

val minus_int = {minus = minus_inta} : int minus;

val group_add_int =
  {cancel_semigroup_add_group_add = cancel_semigroup_add_int,
    minus_group_add = minus_int, monoid_add_group_add = monoid_add_int,
    uminus_group_add = uminus_int}
  : int group_add;

val ab_group_add_int =
  {cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_int,
    group_add_ab_group_add = group_add_int}
  : int ab_group_add;

val ring_int =
  {ab_group_add_ring = ab_group_add_int,
    semiring_0_cancel_ring = semiring_0_cancel_int}
  : int ring;

datatype nat = Zero_nat | Suc of nat;

fun equal_nata Zero_nat (Suc x2) = false
  | equal_nata (Suc x2) Zero_nat = false
  | equal_nata (Suc x2) (Suc y2) = equal_nata x2 y2
  | equal_nata Zero_nat Zero_nat = true;

val equal_nat = {equal = equal_nata} : nat equal;

val zero_nat = {zero = Zero_nat} : nat zero;

fun less_eq_nat (Suc m) n = less_nat m n
  | less_eq_nat Zero_nat n = true
and less_nat m (Suc n) = less_eq_nat m n
  | less_nat n Zero_nat = false;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

val preorder_nat = {ord_preorder = ord_nat} : nat preorder;

val order_nat = {preorder_order = preorder_nat} : nat order;

type 'a linpreorder = {preorder_linpreorder : 'a preorder};
val preorder_linpreorder = #preorder_linpreorder :
  'a linpreorder -> 'a preorder;

type 'a linorder =
  {linpreorder_linorder : 'a linpreorder, order_linorder : 'a order};
val linpreorder_linorder = #linpreorder_linorder :
  'a linorder -> 'a linpreorder;
val order_linorder = #order_linorder : 'a linorder -> 'a order;

val linpreorder_nat = {preorder_linpreorder = preorder_nat} : nat linpreorder;

val linorder_nat =
  {linpreorder_linorder = linpreorder_nat, order_linorder = order_nat} :
  nat linorder;

fun compare_nat (Suc n) (Suc m) = compare_nat n m
  | compare_nat (Suc n) Zero_nat = General.GREATER
  | compare_nat Zero_nat (Suc m) = General.LESS
  | compare_nat Zero_nat Zero_nat = General.EQUAL;

type 'a linpreorder_compare =
  {linpreorder_linpreorder_compare : 'a linpreorder,
    compare : 'a -> 'a -> General.order};
val linpreorder_linpreorder_compare = #linpreorder_linpreorder_compare :
  'a linpreorder_compare -> 'a linpreorder;
val compare = #compare : 'a linpreorder_compare -> 'a -> 'a -> General.order;

type 'a linorder_compare =
  {linpreorder_compare_linorder_compare : 'a linpreorder_compare,
    linorder_linorder_compare : 'a linorder};
val linpreorder_compare_linorder_compare = #linpreorder_compare_linorder_compare
  : 'a linorder_compare -> 'a linpreorder_compare;
val linorder_linorder_compare = #linorder_linorder_compare :
  'a linorder_compare -> 'a linorder;

val linpreorder_compare_nat =
  {linpreorder_linpreorder_compare = linpreorder_nat, compare = compare_nat} :
  nat linpreorder_compare;

val linorder_compare_nat =
  {linpreorder_compare_linorder_compare = linpreorder_compare_nat,
    linorder_linorder_compare = linorder_nat}
  : nat linorder_compare;

fun eq A_ a b = equal A_ a b;

fun equal_proda A_ B_ (x1, x2) (y1, y2) = eq A_ x1 y1 andalso eq B_ x2 y2;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

fun less_eq_prod A_ B_ (x1, y1) (x2, y2) =
  less A_ x1 x2 orelse less_eq A_ x1 x2 andalso less_eq B_ y1 y2;

fun less_prod A_ B_ (x1, y1) (x2, y2) =
  less A_ x1 x2 orelse less_eq A_ x1 x2 andalso less B_ y1 y2;

fun ord_prod A_ B_ = {less_eq = less_eq_prod A_ B_, less = less_prod A_ B_} :
  ('a * 'b) ord;

fun preorder_prod A_ B_ =
  {ord_preorder = ord_prod (ord_preorder A_) (ord_preorder B_)} :
  ('a * 'b) preorder;

fun order_prod A_ B_ =
  {preorder_order = preorder_prod (preorder_order A_) (preorder_order B_)} :
  ('a * 'b) order;

fun uminus_comparisona General.LESS = General.GREATER
  | uminus_comparisona General.EQUAL = General.EQUAL
  | uminus_comparisona General.GREATER = General.LESS;

val uminus_comparison = {uminus = uminus_comparisona} : General.order uminus;

fun less_eq_list (A1_, A2_) (x :: xs) (y :: ys) =
  less ((ord_preorder o preorder_order) A2_) x y orelse
    eq A1_ x y andalso less_eq_list (A1_, A2_) xs ys
  | less_eq_list (A1_, A2_) [] xs = true
  | less_eq_list (A1_, A2_) (x :: xs) [] = false;

datatype 'a nat_mapping_sparse = Abs_nat_mapping_sparse of (nat * 'a) list;

fun rep_nat_mapping_sparse A_ (Abs_nat_mapping_sparse x) = x;

fun less_eq_nat_mapping_sparse (A1_, A2_, A3_) xa xc =
  less_eq_list
    (equal_prod equal_nat A2_, order_prod order_nat (order_linorder A3_))
    (rep_nat_mapping_sparse A1_ xa) (rep_nat_mapping_sparse A1_ xc);

datatype 'a nat_mapping_dense = Abs_nat_mapping_dense of 'a list;

fun rep_nat_mapping_dense A_ (Abs_nat_mapping_dense x) = 
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#rep_nat_mapping_dense: A_ (Abs_nat_mapping_dense (" ^ PolyML.makestring x ^ ")");
  x
);

fun less_eq_nat_mapping_dense (A1_, A2_, A3_) xa xc =
  less_eq_list (A2_, order_linorder A3_) (rep_nat_mapping_dense A1_ xa)
    (rep_nat_mapping_dense A1_ xc);

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

datatype 'a nat_mapping = Dense of 'a nat_mapping_dense |
  Sparse of 'a nat_mapping_sparse;

datatype char = Char of nibble * nibble;

fun less_eq_nat_mapping (A1_, A2_, A3_) (Sparse x) (Sparse y) =
  less_eq_nat_mapping_sparse (A1_, A2_, A3_) x y
  | less_eq_nat_mapping (A1_, A2_, A3_) (Dense x) (Dense y) =
    less_eq_nat_mapping_dense (A1_, A2_, A3_) x y
  | less_eq_nat_mapping (A1_, A2_, A3_) x y =
    (raise Fail
      "order comparison between different representations unsupported")
      (fn _ => less_eq_nat_mapping (A1_, A2_, A3_) x y);

fun less_list (A1_, A2_) (x :: xs) (y :: ys) =
  less ((ord_preorder o preorder_order) A2_) x y orelse
    eq A1_ x y andalso less_list (A1_, A2_) xs ys
  | less_list (A1_, A2_) [] (x :: xs) = true
  | less_list (A1_, A2_) xs [] = false;

fun less_nat_mapping_sparse (A1_, A2_, A3_) xa xc =
  less_list
    (equal_prod equal_nat A2_, order_prod order_nat (order_linorder A3_))
    (rep_nat_mapping_sparse A1_ xa) (rep_nat_mapping_sparse A1_ xc);

fun less_nat_mapping_dense (A1_, A2_, A3_) xa xc =
  less_list (A2_, order_linorder A3_) (rep_nat_mapping_dense A1_ xa)
    (rep_nat_mapping_dense A1_ xc);

fun less_nat_mapping (A1_, A2_, A3_) (Sparse x) (Sparse y) =
  less_nat_mapping_sparse (A1_, A2_, A3_) x y
  | less_nat_mapping (A1_, A2_, A3_) (Dense x) (Dense y) =
    less_nat_mapping_dense (A1_, A2_, A3_) x y
  | less_nat_mapping (A1_, A2_, A3_) x y =
    (raise Fail
      "strict order comparison between different representations unsupported")
      (fn _ => less_nat_mapping (A1_, A2_, A3_) x y);

fun ord_nat_mapping (A1_, A2_, A3_) =
  {less_eq = less_eq_nat_mapping (A1_, A2_, A3_),
    less = less_nat_mapping (A1_, A2_, A3_)}
  : 'a nat_mapping ord;

fun equal_list A_ [] (x21 :: x22) = false
  | equal_list A_ (x21 :: x22) [] = false
  | equal_list A_ (x21 :: x22) (y21 :: y22) =
    eq A_ x21 y21 andalso equal_list A_ x22 y22
  | equal_list A_ [] [] = true;

fun equal_nat_mapping_sparse (A1_, A2_) xa xc =
  equal_list (equal_prod equal_nat A2_) (rep_nat_mapping_sparse A1_ xa)
    (rep_nat_mapping_sparse A1_ xc);

fun equal_nat_mapping_dense (A1_, A2_) xa xc =
  equal_list A2_ (rep_nat_mapping_dense A1_ xa) (rep_nat_mapping_dense A1_ xc);

fun equal_nat_mapping (A1_, A2_) (Sparse x) (Sparse y) =
  equal_nat_mapping_sparse (A1_, A2_) x y
  | equal_nat_mapping (A1_, A2_) (Dense x) (Dense y) =
    equal_nat_mapping_dense (A1_, A2_) x y
  | equal_nat_mapping (A1_, A2_) x y =
    (raise Fail
      "equality comparison between different representations unsupported")
      (fn _ => equal_nat_mapping (A1_, A2_) x y);

datatype 'a poly_rec = Coeff of 'a | Powers of 'a poly_rec nat_mapping;

fun zero_poly_reca A_ = Coeff (zero A_);

fun zero_poly_rec A_ = {zero = zero_poly_reca A_} : 'a poly_rec zero;

fun equal_poly_rec (A1_, A2_) = {equal = equal_poly_reca (A1_, A2_)} :
  'a poly_rec equal
and equal_poly_reca (A1_, A2_) (Coeff x1) (Powers x2) = false
  | equal_poly_reca (A1_, A2_) (Powers x2) (Coeff x1) = false
  | equal_poly_reca (A1_, A2_) (Powers x2) (Powers y2) =
    equal_nat_mapping (zero_poly_rec A1_, equal_poly_rec (A1_, A2_)) x2 y2
  | equal_poly_reca (A1_, A2_) (Coeff x1) (Coeff y1) = eq A2_ x1 y1;

fun comparator_of_le A_ x y =
  (if less_eq A_ x y
    then (if less_eq A_ y x then General.EQUAL else General.LESS)
    else General.GREATER);

fun filter_map2 f ((k, v) :: xs) =
  (case f k v of NONE => filter_map2 f xs
    | SOME w => (k, w) :: filter_map2 f xs)
  | filter_map2 f [] = [];

fun oa_zip_with cmp f g h ((ka, va) :: xs) ((k, v) :: ys) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#oa_zip_with 1 f g h : ((" ^ 
PolyML.makestring ka ^ ", " ^ PolyML.makestring va ^ ") :: (" ^ PolyML.makestring xs ^ "))  ((" ^ 
PolyML.makestring k ^ ", " ^ PolyML.makestring v ^ ") :: (" ^ PolyML.makestring ys ^ "))");
  (case cmp ka k
    of General.LESS =>
      (case f ka va of NONE => oa_zip_with cmp f g h xs ((k, v) :: ys)
        | SOME w => (ka, w) :: oa_zip_with cmp f g h xs ((k, v) :: ys))
    | General.EQUAL =>
      (case h ka va v of NONE => oa_zip_with cmp f g h xs ys
        | SOME w => (ka, w) :: oa_zip_with cmp f g h xs ys)
    | General.GREATER =>
      (case g k v of NONE => oa_zip_with cmp f g h ((ka, va) :: xs) ys
        | SOME w => (k, w) :: oa_zip_with cmp f g h ((ka, va) :: xs) ys))
)
  | oa_zip_with cmp f g h xs [] = 
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#oa_zip_with 2 f g h : (" ^ 
PolyML.makestring xs ^ ") []");
    filter_map2 f xs
)
  | oa_zip_with cmp f g h [] ys = 
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#oa_zip_with 3 f g h : [] (" ^ 
PolyML.makestring ys ^ ")");
    filter_map2 g ys
);

fun zip_with_sparse_raw A_ B_ C_ (D1_, D2_) f xs ys =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_sparse_raw: f (" ^ 
PolyML.makestring xs ^ ")  (" ^ PolyML.makestring ys ^ ")");
  oa_zip_with
    (comparator_of_le ((ord_preorder o preorder_order o order_linorder) A_))
    (fn k => fn v =>
      let
        val z = f k v (zero C_);
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    (fn k => fn v =>
      let
        val z = f k (zero B_) v;
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    (fn k => fn v => fn va =>
      let
        val z = f k v va;
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    xs ys
);

fun zip_with_sparse B_ C_ (A1_, A2_) xc xd xe =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_sparse xc: (" ^ PolyML.makestring xd ^ 
")  (" ^ PolyML.makestring xe ^ ")");
  Abs_nat_mapping_sparse
    (zip_with_sparse_raw linorder_nat B_ C_ (A1_, A2_) xc
      (rep_nat_mapping_sparse B_ xd) (rep_nat_mapping_sparse C_ xe))
);

fun apply0 (A1_, A2_) B_ f x = 
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#apply0: f (" ^ PolyML.makestring x ^ ") \<longrightarrow> (" ^
PolyML.makestring (if eq A2_ x (zero A1_) then zero B_ else f x) ^ ")");
  (if eq A2_ x (zero A1_) then zero B_ else f x)
);

fun null [] = true
  | null (x :: xs) = false;

fun strip_map (A1_, A2_) (B1_, B2_) f n [] = 
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#strip_map 1: (" ^ PolyML.makestring n ^ ") \<longrightarrow> []");
    []
)
  | strip_map (A1_, A2_) (B1_, B2_) f n (x :: xs) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#strip_map 2: (" ^ PolyML.makestring n ^ 
")  ((" ^ PolyML.makestring x ^ ") :: (" ^ PolyML.makestring xs ^ "))");
    let
      val ys = strip_map (A1_, A2_) (B1_, B2_) f (Suc n) xs;
      val y = apply0 (A1_, A2_) B1_ (f n) x;
    in
      (if null ys andalso eq B2_ y (zero B1_) then [] else y :: ys)
    end
);

fun apply0_2 (A1_, A2_) (B1_, B2_) C_ f x y =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#apply0_2: f (" ^ PolyML.makestring x ^ 
")  (" ^ PolyML.makestring y ^ ") \<longrightarrow> (" ^ 
PolyML.makestring (if eq A2_ x (zero A1_) andalso eq B2_ y (zero B1_) then zero C_ else f x y) ^ ")");
  (if eq A2_ x (zero A1_) andalso eq B2_ y (zero B1_) then zero C_ else f x y)
);

fun zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n [] ys =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_dense_raw 1: (" ^ PolyML.makestring n ^ 
")  []  (" ^ PolyML.makestring ys ^ ")");
  strip_map (B1_, B2_) (C1_, C2_) (fn na => f na (zero A1_)) n ys
)
  | zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n xs [] =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_dense_raw 2: (" ^ PolyML.makestring n ^ 
")  (" ^ PolyML.makestring xs ^ ")");
    strip_map (A1_, A2_) (C1_, C2_) (fn na => fn x => f na x (zero B1_)) n xs
)
  | zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n (x :: xs) (y :: ys) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_dense_raw 3: (" ^ PolyML.makestring n ^ 
")  ((" ^ PolyML.makestring x ^ ") :: (" ^ PolyML.makestring xs ^ "))  ((" ^ 
PolyML.makestring y ^ ") :: (" ^ PolyML.makestring ys ^ "))");
      let
        val zs =
          zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f (Suc n) xs ys;
        val z = apply0_2 (A1_, A2_) (B1_, B2_) C1_ (f n) x y;
      in
        (if null zs andalso eq C2_ z (zero C1_) then [] else z :: zs)
      end
);

fun zip_with_dense (B1_, B2_) (C1_, C2_) (A1_, A2_) xc xd xe =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#zip_with_dense xc: (" ^ PolyML.makestring xd ^ 
")  (" ^ PolyML.makestring xe ^ ")");
  Abs_nat_mapping_dense
    (zip_with_dense_raw (B1_, B2_) (C1_, C2_) (A1_, A2_) xc Zero_nat
      (rep_nat_mapping_dense B1_ xd) (rep_nat_mapping_dense C1_ xe))
);

fun plus_nat_mapping (A1_, A2_) (Sparse xs) (Sparse ys) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_nat_mapping 1: (Sparse (" ^ PolyML.makestring xs ^ 
"))  (Sparse (" ^ PolyML.makestring ys ^ "))");
  Sparse
    (zip_with_sparse (zero_monoid_add A1_) (zero_monoid_add A1_)
      (zero_monoid_add A1_, A2_)
      (fn _ => plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_)) xs
      ys)
)
  | plus_nat_mapping (A1_, A2_) (Dense xs) (Dense ys) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_nat_mapping 2: (Dense (" ^ PolyML.makestring xs ^ 
"))  (Dense (" ^ PolyML.makestring ys ^ "))");
    Dense (zip_with_dense (zero_monoid_add A1_, A2_) (zero_monoid_add A1_, A2_)
            (zero_monoid_add A1_, A2_)
            (fn _ => plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_))
            xs ys)
);

fun single_sparse (A1_, A2_) xb xc =
  Abs_nat_mapping_sparse (if eq A2_ xc (zero A1_) then [] else [(xb, xc)]);

fun replicate Zero_nat x = []
  | replicate (Suc n) x = x :: replicate n x;

fun single_dense (A1_, A2_) xb xc =
  Abs_nat_mapping_dense
    (if eq A2_ xc (zero A1_) then [] else replicate xb (zero A1_) @ [xc]);

datatype implb = DENSE | SPARSE;

fun single (A1_, A2_) SPARSE k v = Sparse (single_sparse (A1_, A2_) k v)
  | single (A1_, A2_) DENSE k v = Dense (single_dense (A1_, A2_) k v);

fun implT A_ (Sparse s) = SPARSE
  | implT A_ (Dense d) = DENSE;

fun plus_poly_rec (A1_, A2_) = {plus = plus_poly_reca (A1_, A2_)} :
  'a poly_rec plus
and semigroup_add_poly_rec (A1_, A2_) =
  {plus_semigroup_add = plus_poly_rec (A1_, A2_)} : 'a poly_rec semigroup_add
and monoid_add_poly_rec (A1_, A2_) =
  {semigroup_add_monoid_add = semigroup_add_poly_rec (A1_, A2_),
    zero_monoid_add = zero_poly_rec (zero_monoid_add A1_)}
  : 'a poly_rec monoid_add
and plus_poly_reca (A1_, A2_) (Powers ps) (Coeff y) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_poly_rec 4: (Powers (" ^ PolyML.makestring ps ^ 
"))  (Coeff (" ^ PolyML.makestring y ^ "))");
  Powers
    (plus_nat_mapping
      (monoid_add_poly_rec (A1_, A2_),
        equal_poly_rec (zero_monoid_add A1_, A2_))
      ps (single
           (zero_poly_rec (zero_monoid_add A1_),
             equal_poly_rec (zero_monoid_add A1_, A2_))
           (implT (zero_poly_rec (zero_monoid_add A1_)) ps) Zero_nat (Coeff y)))
)
  | plus_poly_reca (A1_, A2_) (Coeff x) (Powers qs) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_poly_rec 3: (Coeff (" ^ PolyML.makestring x ^ 
"))  (Powers (" ^ PolyML.makestring qs ^ "))");
    Powers
      (plus_nat_mapping
        (monoid_add_poly_rec (A1_, A2_),
          equal_poly_rec (zero_monoid_add A1_, A2_))
        (single
          (zero_poly_rec (zero_monoid_add A1_),
            equal_poly_rec (zero_monoid_add A1_, A2_))
          (implT (zero_poly_rec (zero_monoid_add A1_)) qs) Zero_nat (Coeff x))
        qs)
)
  | plus_poly_reca (A1_, A2_) (Powers ps) (Powers qs) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_poly_rec 2: (Powers (" ^ PolyML.makestring ps ^ 
"))  (Powers (" ^ PolyML.makestring qs ^ "))");
    Powers
      (plus_nat_mapping
        (monoid_add_poly_rec (A1_, A2_),
          equal_poly_rec (zero_monoid_add A1_, A2_))
        ps qs)
)
  | plus_poly_reca (A1_, A2_) (Coeff x) (Coeff y) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_poly_rec 1: (Coeff (" ^ PolyML.makestring x ^ 
"))  (Coeff (" ^ PolyML.makestring y ^ "))");
    Coeff (plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_) x y)
);

datatype monom_order = Rlex_mo | Lex_mo;

datatype 'a poly_distr =
  Poly_distr_of_raw of ((nat nat_mapping * 'a) list * monom_order);

datatype 'a mpoly = Distr of 'a poly_distr | Rec of 'a poly_rec;

datatype impla = RLEX | LEX;

datatype impl = REC of (nat -> implb) | DISTR of impla;

datatype ('b, 'a) ignore_arg = Abs_ignore_arg of ('b -> 'a);

fun id x = (fn xa => xa) x;

fun foldr f [] = id
  | foldr f (x :: xs) = f x o foldr f xs;

fun plus_nat (Suc m) n = plus_nat m (Suc n)
  | plus_nat Zero_nat n = n;

val one_nat : nat = Suc Zero_nat;

fun nat_of_num (Bit1 n) = let
                            val m = nat_of_num n;
                          in
                            Suc (plus_nat m m)
                          end
  | nat_of_num (Bit0 n) = let
                            val m = nat_of_num n;
                          in
                            plus_nat m m
                          end
  | nat_of_num One = one_nat;

fun monom_distr_impl (A1_, A2_) mo m c =
  Poly_distr_of_raw (if eq A2_ c (zero A1_) then ([], mo) else ([(m, c)], mo));

fun monom_distr (A1_, A2_) mo m c = Distr (monom_distr_impl (A1_, A2_) mo m c);

fun monom_distr_ignore (A1_, A2_) = Abs_ignore_arg (monom_distr (A1_, A2_));

fun apply_ignore (Abs_ignore_arg x) = x;

fun switch_monom_order f RLEX = apply_ignore f Rlex_mo
  | switch_monom_order f LEX = apply_ignore f Lex_mo;

fun monom_distra (A1_, A2_) =
  switch_monom_order (monom_distr_ignore (A1_, A2_));

fun minus_nat (Suc m) (Suc n) = minus_nat m n
  | minus_nat Zero_nat n = Zero_nat
  | minus_nat m Zero_nat = m;

fun shift_rec (A1_, A2_) (Coeff x) n = Coeff x
  | shift_rec (A1_, A2_) (Powers ps) Zero_nat = Powers ps
  | shift_rec (A1_, A2_) (Powers ps) (Suc n) =
    Powers
      (single (zero_poly_rec A1_, equal_poly_rec (A1_, A2_))
        (implT (zero_poly_rec A1_) ps) Zero_nat
        (shift_rec (A1_, A2_) (Powers ps) n));

fun foldr_sparse A_ x xc =
  foldr (fn (a, b) => x a b) (rep_nat_mapping_sparse A_ xc);

fun foldr_dense_raw (A1_, A2_) f n [] b = b
  | foldr_dense_raw (A1_, A2_) f n (x :: xs) b =
    (if eq A2_ x (zero A1_) then id else f n x)
      (foldr_dense_raw (A1_, A2_) f (Suc n) xs b);

fun foldr_dense (A1_, A2_) x xc =
  foldr_dense_raw (A1_, A2_) x Zero_nat (rep_nat_mapping_dense A1_ xc);

fun foldra (A1_, A2_) f (Sparse xs) z = foldr_sparse A1_ f xs z
  | foldra (A1_, A2_) f (Dense xs) z = foldr_dense (A1_, A2_) f xs z;

fun monom_rec (A1_, A2_) impl m c =
  (if eq A2_ c (zero A1_) then Coeff (zero A1_)
    else let
           val (a, b) =
             foldra (zero_nat, equal_nat)
               (fn v => fn n => fn (p, va) =>
                 (Powers
                    (single (zero_poly_rec A1_, equal_poly_rec (A1_, A2_))
                      (impl n) n (shift_rec (A1_, A2_) p (minus_nat va v))),
                   v))
               m (Coeff c, Zero_nat);
         in
           shift_rec (A1_, A2_) a b
         end);

fun monom (A1_, A2_, A3_) (DISTR mo) m c =
  monom_distra
    ((zero_mult_zero o mult_zero_semiring_0 o semiring_0_semiring_0_cancel o
       semiring_0_cancel_ring)
       A3_,
      A2_)
    mo m c
  | monom (A1_, A2_, A3_) (REC impl) m c =
    Rec (monom_rec
          ((zero_mult_zero o mult_zero_semiring_0 o
             semiring_0_semiring_0_cancel o semiring_0_cancel_ring)
             A3_,
            A2_)
          impl m c);

fun fst (x1, x2) = x1;

fun compare_vimage f compare x y = compare (f x) (f y);

fun raw_of_poly_distr A_ (Poly_distr_of_raw x) = x;

fun equal_monom_order Rlex_mo Rlex_mo = true
  | equal_monom_order Rlex_mo Lex_mo = false
  | equal_monom_order Lex_mo Rlex_mo = false
  | equal_monom_order Lex_mo Lex_mo = true;

fun insort_by compare x [] = [x]
  | insort_by compare x (y :: ys) =
    (case compare x y of General.LESS => x :: y :: ys | General.EQUAL => y :: ys
      | General.GREATER => y :: insort_by compare x ys);

fun sort_by compare xs = foldr (insort_by compare) xs [];

fun compare_nat_mapping (A1_, A2_, A3_) =
  comparator_of_le (ord_nat_mapping (A2_, A3_, linorder_linorder_compare A1_));

fun monom_compare Rlex_mo =
  uminus_funa (uminus_fun uminus_comparison)
    (compare_nat_mapping (linorder_compare_nat, zero_nat, equal_nat))
  | monom_compare Lex_mo =
    compare_nat_mapping (linorder_compare_nat, zero_nat, equal_nat);

fun plus_distr_raw (A1_, A2_, A3_) (p1, cmp1) (p2, cmp2) =
  (if equal_monom_order cmp1 cmp2
    then (oa_zip_with (monom_compare cmp1) (fn _ => SOME) (fn _ => SOME)
            (fn _ => fn x => fn y =>
              let
                val z = plus A1_ x y;
              in
                (if eq A3_ z (zero A2_) then NONE else SOME z)
              end)
            p1 p2,
           cmp1)
    else let
           val cmp = monom_compare cmp1;
           val p2a = sort_by (compare_vimage fst cmp) p2;
         in
           (oa_zip_with cmp (fn _ => SOME) (fn _ => SOME)
              (fn _ => fn x => fn y =>
                let
                  val z = plus A1_ x y;
                in
                  (if eq A3_ z (zero A2_) then NONE else SOME z)
                end)
              p1 p2a,
             cmp1)
         end);

fun plus_poly_distr (A1_, A2_) p q =
  Poly_distr_of_raw
    (plus_distr_raw
      ((plus_semigroup_add o semigroup_add_monoid_add) A1_, zero_monoid_add A1_,
        A2_)
      (raw_of_poly_distr (zero_monoid_add A1_) p)
      (raw_of_poly_distr (zero_monoid_add A1_) q));

fun plus_mpoly (A1_, A2_, A3_) (Rec p) (Rec q) =
(cnt := ! cnt + 1;
writeln ("#" ^ PolyML.makestring cnt ^ "#plus_mpoly 1: (Rec (" ^ PolyML.makestring p ^ 
"))  (Rec (" ^ PolyML.makestring q ^ "))");
  Rec (plus_poly_reca
        ((monoid_add_group_add o group_add_ab_group_add o ab_group_add_ring)
           A3_,
          A2_)
        p q)
)
  | plus_mpoly (A1_, A2_, A3_) (Distr p) (Distr q) =
    Distr (plus_poly_distr
            ((monoid_add_group_add o group_add_ab_group_add o ab_group_add_ring)
               A3_,
              A2_)
            p q);

fun try_plus p q = plus_mpoly (one_int, equal_int, ring_int) p q;

val up_dense : int mpoly =
  plus_mpoly (one_int, equal_int, ring_int)
    (plus_mpoly (one_int, equal_int, ring_int)
      (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
        (single (zero_nat, equal_nat) DENSE Zero_nat Zero_nat) (Pos One))
      (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
        (single (zero_nat, equal_nat) DENSE Zero_nat one_nat) (Pos (Bit0 One))))
    (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat (nat_of_num (Bit0 One)))
      (Pos (Bit1 One)));

val uq_dense : int mpoly =
  plus_mpoly (one_int, equal_int, ring_int)
    (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat Zero_nat)
      (Pos (Bit0 (Bit0 One))))
    (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat one_nat)
      (Pos (Bit1 (Bit0 One))));

val up_sparse : int mpoly =
  plus_mpoly (one_int, equal_int, ring_int)
    (plus_mpoly (one_int, equal_int, ring_int)
      (monom (one_int, equal_int, ring_int) (REC (fn _ => SPARSE))
        (single (zero_nat, equal_nat) DENSE Zero_nat Zero_nat) (Pos One))
      (monom (one_int, equal_int, ring_int) (REC (fn _ => SPARSE))
        (single (zero_nat, equal_nat) DENSE Zero_nat one_nat) (Pos (Bit0 One))))
    (monom (one_int, equal_int, ring_int) (REC (fn _ => SPARSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat (nat_of_num (Bit0 One)))
      (Pos (Bit1 One)));

val uq_sparse : int mpoly =
  plus_mpoly (one_int, equal_int, ring_int)
    (monom (one_int, equal_int, ring_int) (REC (fn _ => SPARSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat Zero_nat)
      (Pos (Bit0 (Bit0 One))))
    (monom (one_int, equal_int, ring_int) (REC (fn _ => SPARSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat one_nat)
      (Pos (Bit1 (Bit0 One))));

(*end; struct Try_Plus*)

*} ML {*
(*open Try_Plus*)
*} ML {*
*} 

subsubsection {* Rec dense addition: writeln *}
(*declare[[ML_print_depth = 999]]   ... doens'nt remove the "?" in output ?!?*)
ML {*
cnt := 0;
try_plus up_dense uq_dense
(* = Rec (Powers (Dense (Abs_nat_mapping_dense 
  [Coeff (Pos (Bit1 (Bit0 One))), Coeff (Pos (Bit1 (Bit1 One))), Coeff (Pos (Bit1 One))])))
  revert \<rightarrow>:      1     0 1 \<rightarrow> 101 = 5           1     1 1 \<rightarrow> 111 = 7            1 1 \<rightarrow> 11 = 3*)
*} ML {*
*} text {*
#ref 1 #plus_mpoly 1: (Rec (Powers (Dense (Abs_nat_mapping_dense [Coeff ?, Coeff ?, Coeff ?]))))  
                      (Rec (Powers (Dense (Abs_nat_mapping_dense [Coeff ?, Coeff ?])))) 
#ref 2 #plus_poly_rec 2: (Powers (Dense (Abs_nat_mapping_dense [Coeff ?, Coeff ?, Coeff ?])))  
                         (Powers (Dense (Abs_nat_mapping_dense [Coeff ?, Coeff ?]))) 
        _____________=
#ref 3#  plus_nat_mapping 2: (Dense (Abs_nat_mapping_dense [?, ?, ?]))  (Dense (Abs_nat_mapping_dense [?, ?])) 
#ref 4#    zip_with_dense xc: (Abs_nat_mapping_dense [?, ?, ?])  (Abs_nat_mapping_dense [?, ?]) 
#ref 5#      rep_nat_mapping_dense: A_ (Abs_nat_mapping_dense ([1, 2, 3]) \<longrightarrow> [1, 2, 3]
#ref 6#      rep_nat_mapping_dense: A_ (Abs_nat_mapping_dense ([4, 5]) \<longrightarrow> [4, 5]
           
        _  ____________________=
#ref 7#        zip_with_dense_raw 3a: (Zero_nat)  ((1) :: ([2, 3]))  ((4) :: ([5])) 
#ref 8#        zip_with_dense_raw 3b: (Suc Zero_nat)  ((2) :: ([3]))  ((5) :: ([])) 
#ref 9#        zip_with_dense_raw 2: (Suc (Suc Zero_nat))  ([?]) 
#ref 10#         strip_map 2: (Suc (Suc Zero_nat))  ((?) :: ([])) 
#ref 11#         strip_map 1: (Suc (Suc (Suc Zero_nat))) \<longrightarrow> [] 
#ref 12#apply0: f (?) \<longrightarrow> (?) 
#ref 13#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 14#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 15#apply0_2: f (?)  (?) \<longrightarrow> (?) 
#ref 16#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 17#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 18#apply0_2: f (?)  (?) \<longrightarrow> (?) 
#ref 19#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 20#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
val it = (): unit
*} value "strip_map" --{* :: "(nat \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'b list"*}
ML {*strip_map (*: ....   (nat -> 'a -> 'b) -> nat -> 'a list -> 'b list*)*}
value "zip_with_dense_raw" --{* :: "(nat \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'b list \<Rightarrow> 'c list"*}
ML {*zip_with_dense_raw (*: ....    (nat -> 'a -> 'b -> 'c) -> nat -> 'a list -> 'b list -> 'c list*)*} 
value "apply0"   --{* :: "('a       \<Rightarrow> 'b) \<Rightarrow> 'a       \<Rightarrow> 'b"*}
value "apply0_2" --{* :: "('a \<Rightarrow> 'b \<Rightarrow> 'c) \<Rightarrow> 'a \<Rightarrow> 'b \<Rightarrow> 'c"*}
value "xxx" --{**}
ML {*
*} ML {*
(* --------------------- step into plus_mpoly (1 + 2.x + 3.x^2) (4 + 5.x) ---------------------
NOT YET ADAPTED TO Florian dropping zip_with cf poly/8b5abcdc250c .. 2f29aeb64c06 

"1~~~~~ fun plus_mpoly 1, args:"; val ((A1_, A2_, A3_), (Rec p), (Rec q)) = 
  ((one_int, equal_int, ring_int), up_dense, uq_dense);
(*p = Powers (Dense (Abs_nat_mapping_dense [Coeff (Pos ...))])): int poly_rec*)

"2~~~~~ fun plus_poly_rec 2, args:"; val ((A1_, A2_: int equal(*NEEDED BELOW*)), (Powers ps), (Powers qs)) = 
  (((monoid_add_group_add o group_add_ab_group_add o ab_group_add_ring) A3_, A2_), p, q);
(*ps = Dense (Abs_nat_mapping_dense [Coeff (Pos ...))]): int poly_rec nat_mapping*)

"3~~~~~ fun zip_with, args:"; val ((A1_, A2_), (B1_, B2_), (C1_, C2_), 
  f: nat -> int poly_rec -> int poly_rec -> int poly_rec(*NEEDED BELOW*), (Dense xs), (Dense ys)) = 
  ((zero_poly_rec (zero_monoid_add A1_),
          equal_poly_rec (zero_monoid_add A1_, A2_)),
        (zero_poly_rec (zero_monoid_add A1_),
          equal_poly_rec (zero_monoid_add A1_, A2_)(* GOON this is the culprit *) ),
        (zero_poly_rec (zero_monoid_add A1_),
          equal_poly_rec (zero_monoid_add A1_, A2_)),
        (fn _ => plus_poly_rec (A1_, A2_)), ps, qs);
(*xs = Abs_nat_mapping_dense [Coeff (Pos ...))]: int poly_rec nat_mapping_dense*)

"4~~~~~ fun zip_with_dense, args:"; val ((B1_, B2_), (C1_, C2_), (A1_, A2_), xc, xd, xe) = 
  ((A1_, A2_), (B1_, B2_), (C1_, C2_), f, xs, ys);
(*xd = Abs_nat_mapping_dense [Coeff (Pos ...))]: int poly_rec nat_mapping_dense*)

"7~~~~~ fun zip_with_dense_raw 3a, args:"; val ((A1_', A2_'), (B1_', B2_'), (C1_', C2_'), 
  f, n', (x' :: xs'), (y' :: ys')) = 
  ((B1_, B2_), (C1_, C2_), (A1_, A2_), xc, Zero_nat,
      (rep_nat_mapping_dense B1_ xd), (rep_nat_mapping_dense C1_ xe));
(xd : int poly_rec nat_mapping_dense) = 
  Abs_nat_mapping_dense [Coeff (Pos One), Coeff (Pos (Bit0 One)), Coeff (Pos (Bit1 One))];
(rep_nat_mapping_dense B1_ xd : int poly_rec list) = 
  [Coeff (Pos One), Coeff (Pos (Bit0 One)), Coeff (Pos (Bit1 One))];
(*xs = [Coeff (Pos (Bit0 One)), Coeff (Pos (Bit1 One))]: int poly_rec list*)

        val zs' =
          zip_with_dense_raw (A1_', A2_') (B1_', B2_') (C1_', C2_') f (Suc n') xs' ys';

"8~~~~~ fun zip_with_dense_raw 3b, args:"; val ((A1_'', A2_''), (B1_'', B2_''), (C1_'', C2_''), 
  f, n'', (x'' :: xs''), (y'' :: ys'')) = 
  ((A1_', A2_'), (B1_', B2_'), (C1_', C2_'), f, (Suc n'), xs', ys');

        val zs'' =
          zip_with_dense_raw (A1_'', A2_'') (B1_'', B2_'') (C1_'', C2_'') f (Suc n'') xs'' ys'';

"9~~~~~ fun zip_with_dense_raw 2 args:"; val ((A1_''', A2_'''), (B1_''', B2_'''), (C1_''', C2_'''), f, n''', xs''', []) = 
  ((A1_'', A2_''), (B1_'', B2_''), (C1_'', C2_''), f, (Suc n''), xs'', ys'');

"<9~~~~~ fun zip_with_dense_raw 2, return:"; val return = (*\<longrightarrow> [Coeff (Pos (Bit1 One))]: int poly_rec list*)
  strip_map (A1_''', A2_''') (C1_''', C2_''') (fn na => fn x => f na x (zero B1_''')) n''' xs''';

x'' = Coeff (Pos (Bit0 One));                                                   (* =2 *)
y'' = Coeff (Pos (Bit1 (Bit0 One)));                                            (* =5 *)
        val z'' = apply0_2 (A1_'', A2_'') (B1_'', B2_'') C1_'' (f n'') x'' y''; (* =7 *)

"<8~~~~~ fun zip_with_dense_raw 2, return:"; val return = (*\<longrightarrow> [Coeff (Pos (Bit1 One))]: int poly_rec list*)
  strip_map (A1_'', A2_'') (C1_'', C2_'') (fn na => fn x => f na x (zero B1_'')) n'' xs'';

x' = Coeff (Pos One);                                                           (* =1 *)
y' = Coeff (Pos (Bit0 (Bit0 One)));                                             (* =4 *)
        val z' = apply0_2 (A1_', A2_') (B1_', B2_') C1_' (f n') x' y';          (* =5 *)

cnt := 0;
"<7~~~~~ fun zip_with_dense_raw 2, return:"; val return = (*\<longrightarrow> [Coeff (Pos (Bit1 One))]: int poly_rec list*)
  strip_map (A1_', A2_') (C1_', C2_') (fn na => fn x => f na x (zero B1_')) n' xs';

return = [Coeff (Pos (Bit0 One)), Coeff (Pos (Bit1 One))]         (* [ 2, 3] *)
----------------------- step into plus_mpoly (1 + 2.x + 3.x^2) (4 + 5.x) ---------------------*)
*} text {*
#ref 1#strip_map 2: (Zero_nat)  ((?) :: ([?])) 
#ref 2#strip_map 2: (Suc Zero_nat)  ((?) :: ([])) 
#ref 3#strip_map 1: (Suc (Suc Zero_nat)) 
#ref 4#apply0: f (?) \<longrightarrow> (?) 
#ref 5#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 6#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 7#apply0: f (?) \<longrightarrow> (?) 
#ref 8#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 9#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
*} ML {*
(*
if z :: zs = [Coeff (Pos (Bit1 (Bit0 One))), Coeff (Pos (Bit1 (Bit1 One))), Coeff (Pos (Bit1 One))]
then () else error "plus_mpoly (1 + 2.x + 3.x^2) (4 + 5.x) changed"
*)
*} ML {*
*}

subsubsection {* Re sparse addition: writeln *}
declare[[ML_print_depth = 999]]
ML {*
cnt := 0;
try_plus up_sparse uq_sparse
(* =  Rec (Powers (Sparse (Abs_nat_mapping_sparse
  [(Zero_nat, Coeff (Pos (Bit1 (Bit0 One)))),              101 = 5
   (Suc Zero_nat, Coeff (Pos (Bit1 (Bit1 One)))),          111 = 7
   (Suc (Suc Zero_nat), Coeff (Pos (Bit1 One)))             11 = 3
  ])))                                                           *)
*} ML {*
*} text {*
#ref 1#plus_mpoly 1: (Rec (Powers (Sparse (Abs_nat_mapping_sparse 
    [(Zero_nat, Coeff ?), (Suc Zero_nat, Coeff ?), (Suc (Suc Zero_nat), Coeff ?)]))))  
  (Rec (Powers (Sparse (Abs_nat_mapping_sparse [(Zero_nat, Coeff ?), (Suc Zero_nat, Coeff ?)])))) 
#ref 2#plus_poly_rec 2: (Powers (Sparse (Abs_nat_mapping_sparse 
    [(Zero_nat, Coeff ?), (Suc Zero_nat, Coeff ?), (Suc (Suc Zero_nat), Coeff ?)])))  
  (Powers (Sparse (Abs_nat_mapping_sparse [(Zero_nat, Coeff ?), (Suc Zero_nat, Coeff ?)]))) 

#ref 3#plus_nat_mapping 1: (Sparse (Abs_nat_mapping_sparse 
    [(Zero_nat, ?), (Suc Zero_nat, ?), (Suc (Suc Zero_nat), ?)]))  
  (Sparse (Abs_nat_mapping_sparse [(Zero_nat, ?), (Suc Zero_nat, ?)])) 
#ref 4#zip_with_sparse xc: 
  (Abs_nat_mapping_sparse [(Zero_nat, ?), (Suc Zero_nat, ?), (Suc (Suc Zero_nat), ?)])  
  (Abs_nat_mapping_sparse [(Zero_nat, ?), (Suc Zero_nat, ?)]) 

      =___________________
#ref 5#zip_with_sparse_raw: f ([(?, ?), (?, ?), (?, ?)])  ([(?, ?), (?, ?)]) 

#ref 6#oa_zip_with 1 f g h : ((?, ?) :: ([(?, ?), (?, ?)]))  ((?, ?) :: ([(?, ?)])) 
#ref 7#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 8#oa_zip_with 1 f g h : ((?, ?) :: ([(?, ?)]))  ((?, ?) :: ([])) 
#ref 9#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
#ref 10#oa_zip_with 2 f g h : ([(?, ?)]) [] 
#ref 11#plus_poly_rec 1: (Coeff (?))  (Coeff (?)) 
val it = (): unit
*} text {*
*} ML {*
*}

subsubsection {* shift_rec is NOT inc_power *}
definition p2x3y :: "int mpoly"
where "p2x3y =  (* (R[y])[x]: 3.y + 2.x *)
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 0 1) 2 +
  monom' (REC (\<lambda>_. DENSE)) (Nat_Mapping.single' DENSE 1 1) 3"
value p2x3y --{*"Rec (Powers (Dense (Abs_nat_mapping_dense [
  Powers (Dense (Abs_nat_mapping_dense [
    Coeff 0, Coeff 3])), 
  Coeff 2])))"*}

export_code try_plus shift_rec p2x3y in SML module_name Try_shift_rec
ML {*
(***************************************************************************
structure Try_shift_rec : sig
  type num
  type int
  type 'a equal
  type 'a zero
  type nat
  type 'a nat_mapping
  type 'a poly_rec
  type implb
  type 'a mpoly
  type impl
  val shift_rec : 'a zero * 'a equal -> 'a poly_rec -> nat -> 'a poly_rec
  val p2x3y : int mpoly
  val try_plus : int mpoly -> int mpoly -> int mpoly
end = struct
**************************************************************************)

type 'a uminus = {uminus : 'a -> 'a};
val uminus = #uminus : 'a uminus -> 'a -> 'a;

fun uminus_funa B_ a x = uminus B_ (a x);

fun uminus_fun B_ = {uminus = uminus_funa B_} : ('a -> 'b) uminus;

datatype num = One | Bit0 of num | Bit1 of num;

fun equal_num (Bit0 numa) (Bit1 num) = false
  | equal_num (Bit1 numa) (Bit0 num) = false
  | equal_num One (Bit1 num) = false
  | equal_num (Bit1 num) One = false
  | equal_num One (Bit0 num) = false
  | equal_num (Bit0 num) One = false
  | equal_num (Bit1 numa) (Bit1 num) = equal_num numa num
  | equal_num (Bit0 numa) (Bit0 num) = equal_num numa num
  | equal_num One One = true;

datatype int = Zero_int | Pos of num | Neg of num;

fun equal_inta (Neg k) (Neg l) = equal_num k l
  | equal_inta (Neg k) (Pos l) = false
  | equal_inta (Neg k) Zero_int = false
  | equal_inta (Pos k) (Neg l) = false
  | equal_inta (Pos k) (Pos l) = equal_num k l
  | equal_inta (Pos k) Zero_int = false
  | equal_inta Zero_int (Neg l) = false
  | equal_inta Zero_int (Pos l) = false
  | equal_inta Zero_int Zero_int = true;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

val equal_int = {equal = equal_inta} : int equal;

val one_inta : int = Pos One;

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_int = {one = one_inta} : int one;

fun uminus_inta (Neg m) = Pos m
  | uminus_inta (Pos m) = Neg m
  | uminus_inta Zero_int = Zero_int;

fun plus_num (Bit1 m) (Bit1 n) = Bit0 (plus_num (plus_num m n) One)
  | plus_num (Bit1 m) (Bit0 n) = Bit1 (plus_num m n)
  | plus_num (Bit1 m) One = Bit0 (plus_num m One)
  | plus_num (Bit0 m) (Bit1 n) = Bit1 (plus_num m n)
  | plus_num (Bit0 m) (Bit0 n) = Bit0 (plus_num m n)
  | plus_num (Bit0 m) One = Bit1 m
  | plus_num One (Bit1 n) = Bit0 (plus_num n One)
  | plus_num One (Bit0 n) = Bit1 n
  | plus_num One One = Bit0 One;

fun times_num (Bit1 m) (Bit1 n) =
  Bit1 (plus_num (plus_num m n) (Bit0 (times_num m n)))
  | times_num (Bit1 m) (Bit0 n) = Bit0 (times_num (Bit1 m) n)
  | times_num (Bit0 m) (Bit1 n) = Bit0 (times_num m (Bit1 n))
  | times_num (Bit0 m) (Bit0 n) = Bit0 (Bit0 (times_num m n))
  | times_num One n = n
  | times_num m One = m;

fun times_inta (Neg m) (Neg n) = Pos (times_num m n)
  | times_inta (Neg m) (Pos n) = Neg (times_num m n)
  | times_inta (Pos m) (Neg n) = Neg (times_num m n)
  | times_inta (Pos m) (Pos n) = Pos (times_num m n)
  | times_inta Zero_int l = Zero_int
  | times_inta k Zero_int = Zero_int;

fun bitM One = One
  | bitM (Bit0 n) = Bit1 (bitM n)
  | bitM (Bit1 n) = Bit1 (Bit0 n);

fun dup (Neg n) = Neg (Bit0 n)
  | dup (Pos n) = Pos (Bit0 n)
  | dup Zero_int = Zero_int;

fun minus_inta (Neg m) (Neg n) = sub n m
  | minus_inta (Neg m) (Pos n) = Neg (plus_num m n)
  | minus_inta (Pos m) (Neg n) = Pos (plus_num m n)
  | minus_inta (Pos m) (Pos n) = sub m n
  | minus_inta Zero_int l = uminus_inta l
  | minus_inta k Zero_int = k
and sub (Bit0 m) (Bit1 n) = minus_inta (dup (sub m n)) (Pos One)
  | sub (Bit1 m) (Bit0 n) = plus_inta (dup (sub m n)) (Pos One)
  | sub (Bit1 m) (Bit1 n) = dup (sub m n)
  | sub (Bit0 m) (Bit0 n) = dup (sub m n)
  | sub One (Bit1 n) = Neg (Bit0 n)
  | sub One (Bit0 n) = Neg (bitM n)
  | sub (Bit1 m) One = Pos (Bit0 m)
  | sub (Bit0 m) One = Pos (bitM m)
  | sub One One = Zero_int
and plus_inta (Neg m) (Neg n) = Neg (plus_num m n)
  | plus_inta (Neg m) (Pos n) = sub n m
  | plus_inta (Pos m) (Neg n) = sub m n
  | plus_inta (Pos m) (Pos n) = Pos (plus_num m n)
  | plus_inta Zero_int l = l
  | plus_inta k Zero_int = k;

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a minus = {minus : 'a -> 'a -> 'a};
val minus = #minus : 'a minus -> 'a -> 'a -> 'a;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

type 'a plus = {plus : 'a -> 'a -> 'a};
val plus = #plus : 'a plus -> 'a -> 'a -> 'a;

type 'a semigroup_add = {plus_semigroup_add : 'a plus};
val plus_semigroup_add = #plus_semigroup_add : 'a semigroup_add -> 'a plus;

type 'a cancel_semigroup_add =
  {semigroup_add_cancel_semigroup_add : 'a semigroup_add};
val semigroup_add_cancel_semigroup_add = #semigroup_add_cancel_semigroup_add :
  'a cancel_semigroup_add -> 'a semigroup_add;

type 'a ab_semigroup_add = {semigroup_add_ab_semigroup_add : 'a semigroup_add};
val semigroup_add_ab_semigroup_add = #semigroup_add_ab_semigroup_add :
  'a ab_semigroup_add -> 'a semigroup_add;

type 'a cancel_ab_semigroup_add =
  {ab_semigroup_add_cancel_ab_semigroup_add : 'a ab_semigroup_add,
    cancel_semigroup_add_cancel_ab_semigroup_add : 'a cancel_semigroup_add};
val ab_semigroup_add_cancel_ab_semigroup_add =
  #ab_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a ab_semigroup_add;
val cancel_semigroup_add_cancel_ab_semigroup_add =
  #cancel_semigroup_add_cancel_ab_semigroup_add :
  'a cancel_ab_semigroup_add -> 'a cancel_semigroup_add;

type 'a monoid_add =
  {semigroup_add_monoid_add : 'a semigroup_add, zero_monoid_add : 'a zero};
val semigroup_add_monoid_add = #semigroup_add_monoid_add :
  'a monoid_add -> 'a semigroup_add;
val zero_monoid_add = #zero_monoid_add : 'a monoid_add -> 'a zero;

type 'a comm_monoid_add =
  {ab_semigroup_add_comm_monoid_add : 'a ab_semigroup_add,
    monoid_add_comm_monoid_add : 'a monoid_add};
val ab_semigroup_add_comm_monoid_add = #ab_semigroup_add_comm_monoid_add :
  'a comm_monoid_add -> 'a ab_semigroup_add;
val monoid_add_comm_monoid_add = #monoid_add_comm_monoid_add :
  'a comm_monoid_add -> 'a monoid_add;

type 'a cancel_comm_monoid_add =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add : 'a cancel_ab_semigroup_add,
    comm_monoid_add_cancel_comm_monoid_add : 'a comm_monoid_add};
val cancel_ab_semigroup_add_cancel_comm_monoid_add =
  #cancel_ab_semigroup_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a cancel_ab_semigroup_add;
val comm_monoid_add_cancel_comm_monoid_add =
  #comm_monoid_add_cancel_comm_monoid_add :
  'a cancel_comm_monoid_add -> 'a comm_monoid_add;

type 'a mult_zero = {times_mult_zero : 'a times, zero_mult_zero : 'a zero};
val times_mult_zero = #times_mult_zero : 'a mult_zero -> 'a times;
val zero_mult_zero = #zero_mult_zero : 'a mult_zero -> 'a zero;

type 'a semigroup_mult = {times_semigroup_mult : 'a times};
val times_semigroup_mult = #times_semigroup_mult :
  'a semigroup_mult -> 'a times;

type 'a semiring =
  {ab_semigroup_add_semiring : 'a ab_semigroup_add,
    semigroup_mult_semiring : 'a semigroup_mult};
val ab_semigroup_add_semiring = #ab_semigroup_add_semiring :
  'a semiring -> 'a ab_semigroup_add;
val semigroup_mult_semiring = #semigroup_mult_semiring :
  'a semiring -> 'a semigroup_mult;

type 'a semiring_0 =
  {comm_monoid_add_semiring_0 : 'a comm_monoid_add,
    mult_zero_semiring_0 : 'a mult_zero, semiring_semiring_0 : 'a semiring};
val comm_monoid_add_semiring_0 = #comm_monoid_add_semiring_0 :
  'a semiring_0 -> 'a comm_monoid_add;
val mult_zero_semiring_0 = #mult_zero_semiring_0 :
  'a semiring_0 -> 'a mult_zero;
val semiring_semiring_0 = #semiring_semiring_0 : 'a semiring_0 -> 'a semiring;

type 'a semiring_0_cancel =
  {cancel_comm_monoid_add_semiring_0_cancel : 'a cancel_comm_monoid_add,
    semiring_0_semiring_0_cancel : 'a semiring_0};
val cancel_comm_monoid_add_semiring_0_cancel =
  #cancel_comm_monoid_add_semiring_0_cancel :
  'a semiring_0_cancel -> 'a cancel_comm_monoid_add;
val semiring_0_semiring_0_cancel = #semiring_0_semiring_0_cancel :
  'a semiring_0_cancel -> 'a semiring_0;

type 'a group_add =
  {cancel_semigroup_add_group_add : 'a cancel_semigroup_add,
    minus_group_add : 'a minus, monoid_add_group_add : 'a monoid_add,
    uminus_group_add : 'a uminus};
val cancel_semigroup_add_group_add = #cancel_semigroup_add_group_add :
  'a group_add -> 'a cancel_semigroup_add;
val minus_group_add = #minus_group_add : 'a group_add -> 'a minus;
val monoid_add_group_add = #monoid_add_group_add :
  'a group_add -> 'a monoid_add;
val uminus_group_add = #uminus_group_add : 'a group_add -> 'a uminus;

type 'a ab_group_add =
  {cancel_comm_monoid_add_ab_group_add : 'a cancel_comm_monoid_add,
    group_add_ab_group_add : 'a group_add};
val cancel_comm_monoid_add_ab_group_add = #cancel_comm_monoid_add_ab_group_add :
  'a ab_group_add -> 'a cancel_comm_monoid_add;
val group_add_ab_group_add = #group_add_ab_group_add :
  'a ab_group_add -> 'a group_add;

type 'a ring =
  {ab_group_add_ring : 'a ab_group_add,
    semiring_0_cancel_ring : 'a semiring_0_cancel};
val ab_group_add_ring = #ab_group_add_ring : 'a ring -> 'a ab_group_add;
val semiring_0_cancel_ring = #semiring_0_cancel_ring :
  'a ring -> 'a semiring_0_cancel;

val plus_int = {plus = plus_inta} : int plus;

val semigroup_add_int = {plus_semigroup_add = plus_int} : int semigroup_add;

val cancel_semigroup_add_int =
  {semigroup_add_cancel_semigroup_add = semigroup_add_int} :
  int cancel_semigroup_add;

val ab_semigroup_add_int = {semigroup_add_ab_semigroup_add = semigroup_add_int}
  : int ab_semigroup_add;

val cancel_ab_semigroup_add_int =
  {ab_semigroup_add_cancel_ab_semigroup_add = ab_semigroup_add_int,
    cancel_semigroup_add_cancel_ab_semigroup_add = cancel_semigroup_add_int}
  : int cancel_ab_semigroup_add;

val zero_int = {zero = Zero_int} : int zero;

val monoid_add_int =
  {semigroup_add_monoid_add = semigroup_add_int, zero_monoid_add = zero_int} :
  int monoid_add;

val comm_monoid_add_int =
  {ab_semigroup_add_comm_monoid_add = ab_semigroup_add_int,
    monoid_add_comm_monoid_add = monoid_add_int}
  : int comm_monoid_add;

val cancel_comm_monoid_add_int =
  {cancel_ab_semigroup_add_cancel_comm_monoid_add = cancel_ab_semigroup_add_int,
    comm_monoid_add_cancel_comm_monoid_add = comm_monoid_add_int}
  : int cancel_comm_monoid_add;

val times_int = {times = times_inta} : int times;

val mult_zero_int = {times_mult_zero = times_int, zero_mult_zero = zero_int} :
  int mult_zero;

val semigroup_mult_int = {times_semigroup_mult = times_int} :
  int semigroup_mult;

val semiring_int =
  {ab_semigroup_add_semiring = ab_semigroup_add_int,
    semigroup_mult_semiring = semigroup_mult_int}
  : int semiring;

val semiring_0_int =
  {comm_monoid_add_semiring_0 = comm_monoid_add_int,
    mult_zero_semiring_0 = mult_zero_int, semiring_semiring_0 = semiring_int}
  : int semiring_0;

val semiring_0_cancel_int =
  {cancel_comm_monoid_add_semiring_0_cancel = cancel_comm_monoid_add_int,
    semiring_0_semiring_0_cancel = semiring_0_int}
  : int semiring_0_cancel;

val uminus_int = {uminus = uminus_inta} : int uminus;

val minus_int = {minus = minus_inta} : int minus;

val group_add_int =
  {cancel_semigroup_add_group_add = cancel_semigroup_add_int,
    minus_group_add = minus_int, monoid_add_group_add = monoid_add_int,
    uminus_group_add = uminus_int}
  : int group_add;

val ab_group_add_int =
  {cancel_comm_monoid_add_ab_group_add = cancel_comm_monoid_add_int,
    group_add_ab_group_add = group_add_int}
  : int ab_group_add;

val ring_int =
  {ab_group_add_ring = ab_group_add_int,
    semiring_0_cancel_ring = semiring_0_cancel_int}
  : int ring;

datatype nat = Zero_nat | Suc of nat;

fun equal_nata Zero_nat (Suc x2) = false
  | equal_nata (Suc x2) Zero_nat = false
  | equal_nata (Suc x2) (Suc y2) = equal_nata x2 y2
  | equal_nata Zero_nat Zero_nat = true;

val equal_nat = {equal = equal_nata} : nat equal;

val zero_nat = {zero = Zero_nat} : nat zero;

fun less_eq_nat (Suc m) n = less_nat m n
  | less_eq_nat Zero_nat n = true
and less_nat m (Suc n) = less_eq_nat m n
  | less_nat n Zero_nat = false;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

val preorder_nat = {ord_preorder = ord_nat} : nat preorder;

val order_nat = {preorder_order = preorder_nat} : nat order;

type 'a linpreorder = {preorder_linpreorder : 'a preorder};
val preorder_linpreorder = #preorder_linpreorder :
  'a linpreorder -> 'a preorder;

type 'a linorder =
  {linpreorder_linorder : 'a linpreorder, order_linorder : 'a order};
val linpreorder_linorder = #linpreorder_linorder :
  'a linorder -> 'a linpreorder;
val order_linorder = #order_linorder : 'a linorder -> 'a order;

val linpreorder_nat = {preorder_linpreorder = preorder_nat} : nat linpreorder;

val linorder_nat =
  {linpreorder_linorder = linpreorder_nat, order_linorder = order_nat} :
  nat linorder;

fun compare_nat (Suc n) (Suc m) = compare_nat n m
  | compare_nat (Suc n) Zero_nat = General.GREATER
  | compare_nat Zero_nat (Suc m) = General.LESS
  | compare_nat Zero_nat Zero_nat = General.EQUAL;

type 'a linpreorder_compare =
  {linpreorder_linpreorder_compare : 'a linpreorder,
    compare : 'a -> 'a -> General.order};
val linpreorder_linpreorder_compare = #linpreorder_linpreorder_compare :
  'a linpreorder_compare -> 'a linpreorder;
val compare = #compare : 'a linpreorder_compare -> 'a -> 'a -> General.order;

type 'a linorder_compare =
  {linpreorder_compare_linorder_compare : 'a linpreorder_compare,
    linorder_linorder_compare : 'a linorder};
val linpreorder_compare_linorder_compare = #linpreorder_compare_linorder_compare
  : 'a linorder_compare -> 'a linpreorder_compare;
val linorder_linorder_compare = #linorder_linorder_compare :
  'a linorder_compare -> 'a linorder;

val linpreorder_compare_nat =
  {linpreorder_linpreorder_compare = linpreorder_nat, compare = compare_nat} :
  nat linpreorder_compare;

val linorder_compare_nat =
  {linpreorder_compare_linorder_compare = linpreorder_compare_nat,
    linorder_linorder_compare = linorder_nat}
  : nat linorder_compare;

fun eq A_ a b = equal A_ a b;

fun equal_proda A_ B_ (x1, x2) (y1, y2) = eq A_ x1 y1 andalso eq B_ x2 y2;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

fun less_eq_prod A_ B_ (x1, y1) (x2, y2) =
  less A_ x1 x2 orelse less_eq A_ x1 x2 andalso less_eq B_ y1 y2;

fun less_prod A_ B_ (x1, y1) (x2, y2) =
  less A_ x1 x2 orelse less_eq A_ x1 x2 andalso less B_ y1 y2;

fun ord_prod A_ B_ = {less_eq = less_eq_prod A_ B_, less = less_prod A_ B_} :
  ('a * 'b) ord;

fun preorder_prod A_ B_ =
  {ord_preorder = ord_prod (ord_preorder A_) (ord_preorder B_)} :
  ('a * 'b) preorder;

fun order_prod A_ B_ =
  {preorder_order = preorder_prod (preorder_order A_) (preorder_order B_)} :
  ('a * 'b) order;

fun uminus_comparisona General.LESS = General.GREATER
  | uminus_comparisona General.EQUAL = General.EQUAL
  | uminus_comparisona General.GREATER = General.LESS;

val uminus_comparison = {uminus = uminus_comparisona} : General.order uminus;

fun less_eq_list (A1_, A2_) (x :: xs) (y :: ys) =
  less ((ord_preorder o preorder_order) A2_) x y orelse
    eq A1_ x y andalso less_eq_list (A1_, A2_) xs ys
  | less_eq_list (A1_, A2_) [] xs = true
  | less_eq_list (A1_, A2_) (x :: xs) [] = false;

datatype 'a nat_mapping_sparse = Abs_nat_mapping_sparse of (nat * 'a) list;

fun rep_nat_mapping_sparse A_ (Abs_nat_mapping_sparse x) = x;

fun less_eq_nat_mapping_sparse (A1_, A2_, A3_) xa xc =
  less_eq_list
    (equal_prod equal_nat A2_, order_prod order_nat (order_linorder A3_))
    (rep_nat_mapping_sparse A1_ xa) (rep_nat_mapping_sparse A1_ xc);

datatype 'a nat_mapping_dense = Abs_nat_mapping_dense of 'a list;

fun rep_nat_mapping_dense A_ (Abs_nat_mapping_dense x) = x;

fun less_eq_nat_mapping_dense (A1_, A2_, A3_) xa xc =
  less_eq_list (A2_, order_linorder A3_) (rep_nat_mapping_dense A1_ xa)
    (rep_nat_mapping_dense A1_ xc);

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

datatype 'a nat_mapping = Dense of 'a nat_mapping_dense |
  Sparse of 'a nat_mapping_sparse;

datatype char = Char of nibble * nibble;

fun less_eq_nat_mapping (A1_, A2_, A3_) (Sparse x) (Sparse y) =
  less_eq_nat_mapping_sparse (A1_, A2_, A3_) x y
  | less_eq_nat_mapping (A1_, A2_, A3_) (Dense x) (Dense y) =
    less_eq_nat_mapping_dense (A1_, A2_, A3_) x y
  | less_eq_nat_mapping (A1_, A2_, A3_) x y =
    (raise Fail
      "order comparison between different representations unsupported")
      (fn _ => less_eq_nat_mapping (A1_, A2_, A3_) x y);

fun less_list (A1_, A2_) (x :: xs) (y :: ys) =
  less ((ord_preorder o preorder_order) A2_) x y orelse
    eq A1_ x y andalso less_list (A1_, A2_) xs ys
  | less_list (A1_, A2_) [] (x :: xs) = true
  | less_list (A1_, A2_) xs [] = false;

fun less_nat_mapping_sparse (A1_, A2_, A3_) xa xc =
  less_list
    (equal_prod equal_nat A2_, order_prod order_nat (order_linorder A3_))
    (rep_nat_mapping_sparse A1_ xa) (rep_nat_mapping_sparse A1_ xc);

fun less_nat_mapping_dense (A1_, A2_, A3_) xa xc =
  less_list (A2_, order_linorder A3_) (rep_nat_mapping_dense A1_ xa)
    (rep_nat_mapping_dense A1_ xc);

fun less_nat_mapping (A1_, A2_, A3_) (Sparse x) (Sparse y) =
  less_nat_mapping_sparse (A1_, A2_, A3_) x y
  | less_nat_mapping (A1_, A2_, A3_) (Dense x) (Dense y) =
    less_nat_mapping_dense (A1_, A2_, A3_) x y
  | less_nat_mapping (A1_, A2_, A3_) x y =
    (raise Fail
      "strict order comparison between different representations unsupported")
      (fn _ => less_nat_mapping (A1_, A2_, A3_) x y);

fun ord_nat_mapping (A1_, A2_, A3_) =
  {less_eq = less_eq_nat_mapping (A1_, A2_, A3_),
    less = less_nat_mapping (A1_, A2_, A3_)}
  : 'a nat_mapping ord;

fun equal_list A_ [] (x21 :: x22) = false
  | equal_list A_ (x21 :: x22) [] = false
  | equal_list A_ (x21 :: x22) (y21 :: y22) =
    eq A_ x21 y21 andalso equal_list A_ x22 y22
  | equal_list A_ [] [] = true;

fun equal_nat_mapping_sparse (A1_, A2_) xa xc =
  equal_list (equal_prod equal_nat A2_) (rep_nat_mapping_sparse A1_ xa)
    (rep_nat_mapping_sparse A1_ xc);

fun equal_nat_mapping_dense (A1_, A2_) xa xc =
  equal_list A2_ (rep_nat_mapping_dense A1_ xa) (rep_nat_mapping_dense A1_ xc);

fun equal_nat_mapping (A1_, A2_) (Sparse x) (Sparse y) =
  equal_nat_mapping_sparse (A1_, A2_) x y
  | equal_nat_mapping (A1_, A2_) (Dense x) (Dense y) =
    equal_nat_mapping_dense (A1_, A2_) x y
  | equal_nat_mapping (A1_, A2_) x y =
    (raise Fail
      "equality comparison between different representations unsupported")
      (fn _ => equal_nat_mapping (A1_, A2_) x y);

datatype 'a poly_rec = Coeff of 'a | Powers of 'a poly_rec nat_mapping;

fun zero_poly_reca A_ = Coeff (zero A_);

fun zero_poly_rec A_ = {zero = zero_poly_reca A_} : 'a poly_rec zero;

fun equal_poly_rec (A1_, A2_) = {equal = equal_poly_reca (A1_, A2_)} :
  'a poly_rec equal
and equal_poly_reca (A1_, A2_) (Coeff x1) (Powers x2) = false
  | equal_poly_reca (A1_, A2_) (Powers x2) (Coeff x1) = false
  | equal_poly_reca (A1_, A2_) (Powers x2) (Powers y2) =
    equal_nat_mapping (zero_poly_rec A1_, equal_poly_rec (A1_, A2_)) x2 y2
  | equal_poly_reca (A1_, A2_) (Coeff x1) (Coeff y1) = eq A2_ x1 y1;

fun comparator_of_le A_ x y =
  (if less_eq A_ x y
    then (if less_eq A_ y x then General.EQUAL else General.LESS)
    else General.GREATER);

fun filter_map2 f ((k, v) :: xs) =
  (case f k v of NONE => filter_map2 f xs
    | SOME w => (k, w) :: filter_map2 f xs)
  | filter_map2 f [] = [];

fun oa_zip_with cmp f g h ((ka, va) :: xs) ((k, v) :: ys) =
  (case cmp ka k
    of General.LESS =>
      (case f ka va of NONE => oa_zip_with cmp f g h xs ((k, v) :: ys)
        | SOME w => (ka, w) :: oa_zip_with cmp f g h xs ((k, v) :: ys))
    | General.EQUAL =>
      (case h ka va v of NONE => oa_zip_with cmp f g h xs ys
        | SOME w => (ka, w) :: oa_zip_with cmp f g h xs ys)
    | General.GREATER =>
      (case g k v of NONE => oa_zip_with cmp f g h ((ka, va) :: xs) ys
        | SOME w => (k, w) :: oa_zip_with cmp f g h ((ka, va) :: xs) ys))
  | oa_zip_with cmp f g h xs [] = filter_map2 f xs
  | oa_zip_with cmp f g h [] ys = filter_map2 g ys;

fun zip_with_sparse_raw A_ B_ C_ (D1_, D2_) f xs ys =
  oa_zip_with
    (comparator_of_le ((ord_preorder o preorder_order o order_linorder) A_))
    (fn k => fn v =>
      let
        val z = f k v (zero C_);
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    (fn k => fn v =>
      let
        val z = f k (zero B_) v;
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    (fn k => fn v => fn va =>
      let
        val z = f k v va;
      in
        (if eq D2_ z (zero D1_) then NONE else SOME z)
      end)
    xs ys;

fun zip_with_sparse B_ C_ (A1_, A2_) xc xd xe =
  Abs_nat_mapping_sparse
    (zip_with_sparse_raw linorder_nat B_ C_ (A1_, A2_) xc
      (rep_nat_mapping_sparse B_ xd) (rep_nat_mapping_sparse C_ xe));

fun apply0 (A1_, A2_) B_ f x = (if eq A2_ x (zero A1_) then zero B_ else f x);

fun null [] = true
  | null (x :: xs) = false;

fun strip_map (A1_, A2_) (B1_, B2_) f n [] = []
  | strip_map (A1_, A2_) (B1_, B2_) f n (x :: xs) =
    let
      val ys = strip_map (A1_, A2_) (B1_, B2_) f (Suc n) xs;
      val y = apply0 (A1_, A2_) B1_ (f n) x;
    in
      (if null ys andalso eq B2_ y (zero B1_) then [] else y :: ys)
    end;

fun apply0_2 (A1_, A2_) (B1_, B2_) C_ f x y =
  (if eq A2_ x (zero A1_) andalso eq B2_ y (zero B1_) then zero C_ else f x y);

fun zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n [] ys =
  strip_map (B1_, B2_) (C1_, C2_) (fn na => f na (zero A1_)) n ys
  | zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n xs [] =
    strip_map (A1_, A2_) (C1_, C2_) (fn na => fn x => f na x (zero B1_)) n xs
  | zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f n (x :: xs) (y :: ys)
    = let
        val zs =
          zip_with_dense_raw (A1_, A2_) (B1_, B2_) (C1_, C2_) f (Suc n) xs ys;
        val z = apply0_2 (A1_, A2_) (B1_, B2_) C1_ (f n) x y;
      in
        (if null zs andalso eq C2_ z (zero C1_) then [] else z :: zs)
      end;

fun zip_with_dense (B1_, B2_) (C1_, C2_) (A1_, A2_) xc xd xe =
  Abs_nat_mapping_dense
    (zip_with_dense_raw (B1_, B2_) (C1_, C2_) (A1_, A2_) xc Zero_nat
      (rep_nat_mapping_dense B1_ xd) (rep_nat_mapping_dense C1_ xe));

fun plus_nat_mapping (A1_, A2_) (Sparse xs) (Sparse ys) =
  Sparse
    (zip_with_sparse (zero_monoid_add A1_) (zero_monoid_add A1_)
      (zero_monoid_add A1_, A2_)
      (fn _ => plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_)) xs
      ys)
  | plus_nat_mapping (A1_, A2_) (Dense xs) (Dense ys) =
    Dense (zip_with_dense (zero_monoid_add A1_, A2_) (zero_monoid_add A1_, A2_)
            (zero_monoid_add A1_, A2_)
            (fn _ => plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_))
            xs ys);

fun single_sparse (A1_, A2_) xb xc =
  Abs_nat_mapping_sparse (if eq A2_ xc (zero A1_) then [] else [(xb, xc)]);

fun replicate Zero_nat x = []
  | replicate (Suc n) x = x :: replicate n x;

fun single_dense (A1_, A2_) xb xc =
  Abs_nat_mapping_dense
    (if eq A2_ xc (zero A1_) then [] else replicate xb (zero A1_) @ [xc]);

datatype implb = DENSE | SPARSE;

fun single (A1_, A2_) SPARSE k v = Sparse (single_sparse (A1_, A2_) k v)
  | single (A1_, A2_) DENSE k v = Dense (single_dense (A1_, A2_) k v);

fun implT A_ (Sparse s) = SPARSE
  | implT A_ (Dense d) = DENSE;

fun plus_poly_rec (A1_, A2_) = {plus = plus_poly_reca (A1_, A2_)} :
  'a poly_rec plus
and semigroup_add_poly_rec (A1_, A2_) =
  {plus_semigroup_add = plus_poly_rec (A1_, A2_)} : 'a poly_rec semigroup_add
and monoid_add_poly_rec (A1_, A2_) =
  {semigroup_add_monoid_add = semigroup_add_poly_rec (A1_, A2_),
    zero_monoid_add = zero_poly_rec (zero_monoid_add A1_)}
  : 'a poly_rec monoid_add
and plus_poly_reca (A1_, A2_) (Powers ps) (Coeff y) =
  Powers
    (plus_nat_mapping
      (monoid_add_poly_rec (A1_, A2_),
        equal_poly_rec (zero_monoid_add A1_, A2_))
      ps (single
           (zero_poly_rec (zero_monoid_add A1_),
             equal_poly_rec (zero_monoid_add A1_, A2_))
           (implT (zero_poly_rec (zero_monoid_add A1_)) ps) Zero_nat (Coeff y)))
  | plus_poly_reca (A1_, A2_) (Coeff x) (Powers qs) =
    Powers
      (plus_nat_mapping
        (monoid_add_poly_rec (A1_, A2_),
          equal_poly_rec (zero_monoid_add A1_, A2_))
        (single
          (zero_poly_rec (zero_monoid_add A1_),
            equal_poly_rec (zero_monoid_add A1_, A2_))
          (implT (zero_poly_rec (zero_monoid_add A1_)) qs) Zero_nat (Coeff x))
        qs)
  | plus_poly_reca (A1_, A2_) (Powers ps) (Powers qs) =
    Powers
      (plus_nat_mapping
        (monoid_add_poly_rec (A1_, A2_),
          equal_poly_rec (zero_monoid_add A1_, A2_))
        ps qs)
  | plus_poly_reca (A1_, A2_) (Coeff x) (Coeff y) =
    Coeff (plus ((plus_semigroup_add o semigroup_add_monoid_add) A1_) x y);

datatype monom_order = Rlex_mo | Lex_mo;

datatype 'a poly_distr =
  Poly_distr_of_raw of ((nat nat_mapping * 'a) list * monom_order);

datatype 'a mpoly = Distr of 'a poly_distr | Rec of 'a poly_rec;

datatype impla = RLEX | LEX;

datatype impl = REC of (nat -> implb) | DISTR of impla;

datatype ('b, 'a) ignore_arg = Abs_ignore_arg of ('b -> 'a);

fun id x = (fn xa => xa) x;

fun foldr f [] = id
  | foldr f (x :: xs) = f x o foldr f xs;

fun monom_distr_impl (A1_, A2_) mo m c =
  Poly_distr_of_raw (if eq A2_ c (zero A1_) then ([], mo) else ([(m, c)], mo));

fun monom_distr (A1_, A2_) mo m c = Distr (monom_distr_impl (A1_, A2_) mo m c);

fun monom_distr_ignore (A1_, A2_) = Abs_ignore_arg (monom_distr (A1_, A2_));

fun apply_ignore (Abs_ignore_arg x) = x;

fun switch_monom_order f RLEX = apply_ignore f Rlex_mo
  | switch_monom_order f LEX = apply_ignore f Lex_mo;

fun monom_distra (A1_, A2_) =
  switch_monom_order (monom_distr_ignore (A1_, A2_));

fun minus_nat (Suc m) (Suc n) = minus_nat m n
  | minus_nat Zero_nat n = Zero_nat
  | minus_nat m Zero_nat = m;

fun foldr_sparse A_ x xc =
  foldr (fn (a, b) => x a b) (rep_nat_mapping_sparse A_ xc);

fun foldr_dense_raw (A1_, A2_) f n [] b = b
  | foldr_dense_raw (A1_, A2_) f n (x :: xs) b =
    (if eq A2_ x (zero A1_) then id else f n x)
      (foldr_dense_raw (A1_, A2_) f (Suc n) xs b);

fun foldr_dense (A1_, A2_) x xc =
  foldr_dense_raw (A1_, A2_) x Zero_nat (rep_nat_mapping_dense A1_ xc);

fun foldra (A1_, A2_) f (Sparse xs) z = foldr_sparse A1_ f xs z
  | foldra (A1_, A2_) f (Dense xs) z = foldr_dense (A1_, A2_) f xs z;

fun shift_rec (A1_, A2_) (Coeff x) n = Coeff x
  | shift_rec (A1_, A2_) (Powers ps) Zero_nat = Powers ps
  | shift_rec (A1_, A2_) (Powers ps) (Suc n) =
    Powers
      (single (zero_poly_rec A1_, equal_poly_rec (A1_, A2_))
        (implT (zero_poly_rec A1_) ps) Zero_nat
        (shift_rec (A1_, A2_) (Powers ps) n));

fun monom_rec (A1_, A2_) impl m c =
  (if eq A2_ c (zero A1_) then Coeff (zero A1_)
    else let
           val (a, b) =
             foldra (zero_nat, equal_nat)
               (fn v => fn n => fn (p, va) =>
                 (Powers
                    (single (zero_poly_rec A1_, equal_poly_rec (A1_, A2_))
                      (impl n) n (shift_rec (A1_, A2_) p (minus_nat va v))),
                   v))
               m (Coeff c, Zero_nat);
         in
           shift_rec (A1_, A2_) a b
         end);

fun monom (A1_, A2_, A3_) (DISTR mo) m c =
  monom_distra
    ((zero_mult_zero o mult_zero_semiring_0 o semiring_0_semiring_0_cancel o
       semiring_0_cancel_ring)
       A3_,
      A2_)
    mo m c
  | monom (A1_, A2_, A3_) (REC impl) m c =
    Rec (monom_rec
          ((zero_mult_zero o mult_zero_semiring_0 o
             semiring_0_semiring_0_cancel o semiring_0_cancel_ring)
             A3_,
            A2_)
          impl m c);

fun raw_of_poly_distr A_ (Poly_distr_of_raw x) = x;

fun equal_monom_order Rlex_mo Rlex_mo = true
  | equal_monom_order Rlex_mo Lex_mo = false
  | equal_monom_order Lex_mo Rlex_mo = false
  | equal_monom_order Lex_mo Lex_mo = true;

fun insort_by compare x [] = [x]
  | insort_by compare x (y :: ys) =
    (case compare x y of General.LESS => x :: y :: ys | General.EQUAL => y :: ys
      | General.GREATER => y :: insort_by compare x ys);

fun sort_by compare xs = foldr (insort_by compare) xs [];

fun compare_nat_mapping (A1_, A2_, A3_) =
  comparator_of_le (ord_nat_mapping (A2_, A3_, linorder_linorder_compare A1_));

fun monom_compare Rlex_mo =
  uminus_funa (uminus_fun uminus_comparison)
    (compare_nat_mapping (linorder_compare_nat, zero_nat, equal_nat))
  | monom_compare Lex_mo =
    compare_nat_mapping (linorder_compare_nat, zero_nat, equal_nat);

fun compare_vimage f compare x y = compare (f x) (f y);

fun fst (x1, x2) = x1;

fun plus_distr_raw (A1_, A2_, A3_) (p1, cmp1) (p2, cmp2) =
  (if equal_monom_order cmp1 cmp2
    then (oa_zip_with (monom_compare cmp1) (fn _ => SOME) (fn _ => SOME)
            (fn _ => fn x => fn y =>
              let
                val z = plus A1_ x y;
              in
                (if eq A3_ z (zero A2_) then NONE else SOME z)
              end)
            p1 p2,
           cmp1)
    else let
           val cmp = monom_compare cmp1;
           val p2a = sort_by (compare_vimage fst cmp) p2;
         in
           (oa_zip_with cmp (fn _ => SOME) (fn _ => SOME)
              (fn _ => fn x => fn y =>
                let
                  val z = plus A1_ x y;
                in
                  (if eq A3_ z (zero A2_) then NONE else SOME z)
                end)
              p1 p2a,
             cmp1)
         end);

fun plus_poly_distr (A1_, A2_) p q =
  Poly_distr_of_raw
    (plus_distr_raw
      ((plus_semigroup_add o semigroup_add_monoid_add) A1_, zero_monoid_add A1_,
        A2_)
      (raw_of_poly_distr (zero_monoid_add A1_) p)
      (raw_of_poly_distr (zero_monoid_add A1_) q));

fun plus_mpoly (A1_, A2_, A3_) (Rec p) (Rec q) =
  Rec (plus_poly_reca
        ((monoid_add_group_add o group_add_ab_group_add o ab_group_add_ring)
           A3_,
          A2_)
        p q)
  | plus_mpoly (A1_, A2_, A3_) (Distr p) (Distr q) =
    Distr (plus_poly_distr
            ((monoid_add_group_add o group_add_ab_group_add o ab_group_add_ring)
               A3_,
              A2_)
            p q);

val one_nat : nat = Suc Zero_nat;

val p2x3y : int mpoly =
  plus_mpoly (one_int, equal_int, ring_int)
    (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
      (single (zero_nat, equal_nat) DENSE Zero_nat one_nat) (Pos (Bit0 One)))
    (monom (one_int, equal_int, ring_int) (REC (fn _ => DENSE))
      (single (zero_nat, equal_nat) DENSE one_nat one_nat) (Pos (Bit1 One)));

fun try_plus p q = plus_mpoly (one_int, equal_int, ring_int) p q;

(*end; struct Try_shift_rec*)
*} ML {*
(* find args for shift_rec: *)
zero_int = ({zero = Zero_int}: int zero);
Zero_int = (Zero_int: int);
equal_int;
*} ML {*
shift_rec : 'a zero * 'a equal -> 'a poly_rec -> nat -> 'a poly_rec;
shift_rec (zero_int, equal_int): int poly_rec -> nat -> int poly_rec;
*} ML {*
plus_mpoly : 'a * 'b equal * 'b ring -> 'b mpoly -> 'b mpoly -> 'b mpoly;
plus_mpoly (one_int, equal_int, ring_int) p2x3y p2x3y;
*} ML {*
"~~~~~ fun plus_mpoly, args:"; val ((A1_, A2_, A3_), (Rec p), (Rec q)) = 
  ((one_int, equal_int, ring_int), p2x3y, p2x3y);
*} ML {* (* (R[y])[x]: 3.y + 2.x *)
p = (Powers (Dense (Abs_nat_mapping_dense [
      Powers (Dense (Abs_nat_mapping_dense [
        Coeff Zero_int, Coeff (Pos (Bit1 One))])), (* 0, 11 = 3*)
      Coeff (Pos (Bit0 One))])): int poly_rec)     (* rev 01 = 10 = 2*)
*} ML {* (* no effect             vvvvvvvv *)
shift_rec (zero_int, equal_int) p Zero_nat = 
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [         
      Coeff Zero_int, Coeff (Pos (Bit1 One))])),   (* ...x^0.y^1.3 *)
    Coeff (Pos (Bit0 One))]))                      (* ...x^1.y^0.2 *)
*} ML {* (* (R[y])[x]: 2.y + 3.x --vvvvvvvvvvvv\<longrightarrow> (R[y,x])[z]: 0.z + 3.y + 2.x *)
shift_rec (zero_int, equal_int) p (Suc Zero_nat) = 
  Powers (Dense (Abs_nat_mapping_dense [
    Powers (Dense (Abs_nat_mapping_dense [
      Powers (Dense (Abs_nat_mapping_dense [
        Coeff Zero_int, Coeff (Pos (Bit1 One))])), 
      Coeff (Pos (Bit0 One))]))])) 
*} ML {*
*}
value "xxx" -- {**}

subsubsection {*  *}
subsection {* Isabelle's mechanism ``typedef'' *}
subsection {* Isabelle's mechanism ``instantiation'' *}
section{* Conclusions *}

subsection {**}
ML {*
"~~~~~ fun , args:"; val () = ();
"~~~~~ fun , return:"; val return = ();
*} ML {*
*}
end
